rm(list=ls())
setwd('../Decomposition')

require(xts)
require(tis)
require(lubridate)
require(pROC)
require(xtable)
require(parallel)

# Load from RDS File
dir.data='../Data/'
datasource.name=paste0(dir.data,'Vintage-datasets-May-2019.Rda')
load(datasource.name)

data <- rt.vintages
vintage.dates <- as.yearqtr(as.Date(sapply(rt.vintages, function(x) x[['vintage']])))

build.F <- function(prior) {
  
  F <- matrix(0, nrow = prior$Nobs, ncol = prior$Ns)
  F[1, ] <- c(1, 0, 1)
  
  rownames(F) <- prior$obs.names
  colnames(F) <- prior$state.names
  
  return(F)
}
build.G <- function(prior) {
  
  G <- matrix(0, nrow = prior$Ns, ncol = prior$Ns)
  G[1, ] <- c(2, -1, 0)
  G[2, ] <- c(1, 0, 0)
  G[3, ] <- c(0, 0, 0)
  
  colnames(G) <- rownames(G) <- prior$state.names
  
  return(G)
}
build.H <- function(prior) {
  
  H <- matrix(0, nrow=prior$Ns, ncol=prior$Ns-1)
  rownames(H) <- prior$state.names 
  colnames(H) <- prior$state.names[-2]
  H[1, 1] <- 1
  H[3, 2] <- 1
  
  return(H)
} 

# These functions produce
source('Weight functions.R')
track.contribution <- function(data.vintages, model.vintages, final=FALSE, quasifinal=FALSE, quasireal=FALSE, min.date) {
  
  # Inputs: data.vintages--either rt.vintages or qrt.vintages
  #         model.vintages--the rt- or qrt- model vintages
  #         final -- a flag if you want to decompose the final vintage output gap estimate
  #         min.date -- If you want to subset the data you are decomposing 
  
  # Output: A matrix with columns corresponding to data inputs and the estimated cycle, c
  #         Rows are the quarter being decomposed!
  #         Notice that the sum of the data inputs across rows is that row's cycle!!
  cat("\n ")
  cat("\n ")
  cat("track.contribution is running \n")
  cat("\n ")
  cat("\n ")
  V <- length(model.vintages)
  data.names <- model.vintages[[V]][['prior']][['obs.names']]
  
  taus <- which(vintage.dates > min.date)
  # Eliminate 1996Q1 vintage because BEA did not publish GDP data (government shutdown)
  taus = taus[-122]

  N_draws = dim(model.vintages[[2]]$res$omega)[2]
  temp_draws = array(NA, dim= c(length(taus), length(data.names)+1,N_draws))
  temp_contr_index <- xts(array(NA, c(length(taus), length(data.names)+1)), 
                          order.by=as.yearqtr(as.Date(vintage.dates[taus]) %m+% -months(3)))
  colnames(temp_contr_index) <- c(data.names, 'c')
  
  i <- min(which(sapply(model.vintages, function(x) length(x))!=0)) #Index for storage of functions
  
  if (final==TRUE) {
    cat("\n ")
    cat("Final true")
    cat("\n ")
    
    # For final, decompose the sequence of output gap estimates from the final vintage
    y <- na.omit(data.vintages[[V]][['lgdp']]['1960/',]); colnames(y) <- 'lgdp'
    data$y <- cbind(y)
    prior <- model.vintages[[V]]$prior
    if (prior$T != dim(data$y)[1]) prior$T <- dim(data$y)[1] 
    c <- which(prior$state.names=='c') # Index for cycle
    for_final_fun <- function(i_draws){
  
      param = list()
      param$omega = model.vintages[[V]]$res$omega[,i_draws]
      
      model <- list()
      model$F <- build.F(prior)
      model$G <- build.G(prior)
      model$H <- build.H(prior)
      model$omega <- diag(param$omega)
      
      K <- KFKS(data$y, prior, model)
      
      temp_draws_return = array(NA, dim= c(length(taus), length(data.names)+1))
      for (t in 1:length(taus)) {
        i2 <- which(as.yearqtr(index(data$y)) == index(temp_contr_index[t]))
        w <- w.stT(data$y, prior, model, K, i2)
        W <- apply(w$cont, c(1,2), sum) + matrix(c(w$w0, rep(0, prior$Ns*(prior$Nobs-1))), ncol=prior$Nobs)
        temp_draws_return[t,] <- cbind(W, w$stT)[c,]
      } 
      return(temp_draws_return)
    }
    temp = mclapply(1:N_draws, for_final_fun, mc.cores=detectCores())
    temp_draws = array(as.numeric(unlist(temp)), dim=c(length(taus), length(data.names)+1, N_draws))
    temp_contr <- apply(temp_draws, c(1,2), mean)
    # This is storage for the decomposition
    contr <- xts(array(temp_contr, c(length(taus), length(data.names)+1)), 
                 order.by=as.yearqtr(as.Date(vintage.dates[taus]) %m+% -months(3)))
    colnames(contr) <- c(data.names, 'c')
    
  } else if (quasifinal==TRUE) {
    cat("\n ")
    cat("quasifinal true \n")
    cat("\n ")
    
    for_quasifinal_fun <- function(i_draws){
      temp_draws_return = array(NA, dim= c(length(taus), length(data.names)+1))
      for (t in 1:length(taus)) {
        param = list()
        param$omega = model.vintages[[V]]$res$omega[,i_draws]
        y <- na.omit(data.vintages[[taus[t]]][['lgdp']]['1960/',]); colnames(y) <- 'lgdp'
        data$y <- cbind(y)
        
        prior <- model.vintages[[taus[t]]]$prior
        if (prior$T != dim(data$y)[1]) prior$T <- dim(data$y)[1] 
        c <- which(prior$state.names=='c') # Index for cycle
        
        model <- list()
        model$F <- build.F(prior)
        model$G <- build.G(prior)
        model$H <- build.H(prior)
        model$omega <- diag(param$omega)
        
        K <- KFKS(data$y, prior, model)
        w <- w.stt(data$y, prior, model, K, dim(data.vintages[[taus[t]]][['y']])[1])
        # dim(data.vintages[[taus[t]]][['y']])[1] is the index of the last quarter, which is the quarter we are interested in for the output gap
        
        W <- apply(w$cont, c(1,2), sum) + matrix(c(w$w0, rep(0, prior$Ns*(prior$Nobs-1))), ncol=prior$Nobs)
        temp_draws_return[t,] <- cbind(W, w$stt)[c,]
      }
      return(temp_draws_return)
    }
    temp = mclapply(1:N_draws, for_quasifinal_fun, mc.cores=detectCores())
    temp_draws = array(as.numeric(unlist(temp)), dim=c(length(taus), length(data.names)+1, N_draws))
    temp_contr <- apply(temp_draws, c(1,2), mean)
    # This is storage for the decomposition
    contr <- xts(array(temp_contr, c(length(taus), length(data.names)+1)), 
                 order.by=as.yearqtr(as.Date(vintage.dates[taus]) %m+% -months(3)))
    colnames(contr) <- c(data.names, 'c')

  } else if (quasireal==TRUE) {
    cat("\n ")
    cat("quasireal true \n")
    cat("\n ")
    for_qrt_fun <- function(i_draws){
      temp_draws_return = array(NA, dim= c(length(taus), length(data.names)+1))
      for (t in 1:length(taus)) {
        S = model.vintages[[taus[t]]]$res$S[,i_draws]
        param = list()
        param$omega = model.vintages[[taus[t]]]$res$omega[,i_draws]
        if (length(S)!=0) {
          
          y <- na.omit(data.vintages[[taus[t]]][['lgdp']]['1960/',]); colnames(y) <- 'lgdp'
          data$y <- cbind(y)
          
          prior <- model.vintages[[taus[t]]]$prior
          if (prior$T != dim(data$y)[1]) prior$T <- dim(data$y)[1] 
          c <- which(prior$state.names=='c') # Index for cycle
          
          model <- list()
          model$F <- build.F(prior)
          model$G <- build.G(prior)
          model$H <- build.H(prior)
          model$omega <- diag(param$omega)
          
          K <- KFKS(data$y, prior, model)
          w <- w.stt(data$y, prior, model, K, dim(data.vintages[[taus[t]]][['y']])[1])
          # dim(data.vintages[[taus[t]]][['y']])[1] is the index of the last quarter, which is the quarter we are interested in for the output gap
          
          W <- apply(w$cont, c(1,2), sum) + matrix(c(w$w0, rep(0, prior$Ns*(prior$Nobs-1))), ncol=prior$Nobs)
          temp_draws_return[t,] <- cbind(W, w$stt)[c,]
        }
      }
      return(temp_draws_return)
    }
    temp = mclapply(1:N_draws, for_qrt_fun, mc.cores=detectCores())
    temp_draws = array(as.numeric(unlist(temp)), dim=c(length(taus), length(data.names)+1, N_draws))
    temp_contr <- apply(temp_draws, c(1,2), mean)
    # This is storage for the decomposition
    contr <- xts(array(temp_contr, c(length(taus), length(data.names)+1)), 
                 order.by=as.yearqtr(as.Date(vintage.dates[taus]) %m+% -months(3)))
    colnames(contr) <- c(data.names, 'c')    

  } else {
    cat("\n ")
    cat("real time \n")
    cat("\n ")
    
    for_rt_fun <- function(i_draws){
      temp_draws_return = array(NA, dim= c(length(taus), length(data.names)+1))
      for (t in 1:length(taus)) {
        S = model.vintages[[taus[t]]]$res$S[,i_draws]
        param = list()
        param$omega = model.vintages[[taus[t]]]$res$omega[,i_draws]
        if (length(S)!=0) {
          
          y <- na.omit(data.vintages[[taus[t]]][['lgdp']]['1960/',]); colnames(y) <- 'lgdp'
          data$y <- cbind(y)
          
          prior <- model.vintages[[taus[t]]]$prior
          if (prior$T != dim(data$y)[1]) prior$T <- dim(data$y)[1] 
          c <- which(prior$state.names=='c') # Index for cycle
          
          model <- list()
          model$F <- build.F(prior)
          model$G <- build.G(prior)
          model$H <- build.H(prior)
          model$omega <- diag(param$omega)
          
          K <- KFKS(data$y, prior, model)
          w <- w.stT(data$y, prior, model, K, dim(data.vintages[[taus[t]]][['y']])[1])
          # dim(data.vintages[[taus[t]]][['y']])[1] is the index of the last quarter, which is the quarter we are interested in for the output gap
          
          W <- apply(w$cont, c(1,2), sum) + matrix(c(w$w0, rep(0, prior$Ns*(prior$Nobs-1))), ncol=prior$Nobs)
          temp_draws_return[t,] <- cbind(W, w$stT)[c,]
        }
      }
      return(temp_draws_return)
    }
    temp = mclapply(1:N_draws, for_rt_fun, mc.cores=detectCores())
    temp_draws = array(as.numeric(unlist(temp)), dim=c(length(taus), length(data.names)+1, N_draws))
    temp_contr <- apply(temp_draws, c(1,2), mean)
    # This is storage for the decomposition
    contr <- xts(array(temp_contr, c(length(taus), length(data.names)+1)), 
                 order.by=as.yearqtr(as.Date(vintage.dates[taus]) %m+% -months(3)))
    colnames(contr) <- c(data.names, 'c')
    
  }
  return(contr)
}

tau <- '1965 Q3' # Lower bound for the dates

# HP_SS -------------
rt.hp_ss <- readRDS('../Data/rt-hp_garrat_ss.Rds')
qrt.hp_ss <- readRDS('../Data/qrt-hp_garrat_ss.Rds')
rt.vintages.fa.ar <- readRDS('../Data/rt.vintages.fa.ar.Rds')
qrt.vintages.fa.ar <- readRDS('../Data/qrt.vintages.fa.ar.Rds')

start_time_hp_ss <- Sys.time()
hp_garrat_ss <- list()
hp_garrat_ss$rt <- track.contribution(rt.vintages.fa.ar, rt.hp_ss, final=FALSE, quasifinal=FALSE, quasireal=FALSE, tau)
hp_garrat_ss$qrt <- track.contribution(qrt.vintages.fa.ar, qrt.hp_ss, final=FALSE, quasifinal=FALSE, quasireal=TRUE, tau)
hp_garrat_ss$qf <- track.contribution(qrt.vintages.fa.ar, qrt.hp_ss, final=FALSE, quasifinal=TRUE, quasireal=FALSE, tau)
hp_garrat_ss$final <- track.contribution(rt.vintages.fa.ar, rt.hp_ss, final=TRUE, quasifinal=FALSE, quasireal=FALSE, tau)
end_time_hp_ss <- Sys.time()
print(end_time_hp_ss - start_time_hp_ss)

rm(rt.hp_ss, qrt.hp_ss)

write.table(hp_garrat_ss, "../Decomposition/hp_ss_garratt_decomp.txt", sep="\t") 
