# TABLES AND FIGURES FOR OUTPUT STABILITY PAPER
# 2023 Barbarino-Berge-Stella

# SETTINGS ----

rm(list=ls())

code.loc <- '../Code/'
data.loc <- '../Data/'
figures.loc <- '../Figures/'

setwd(code.loc)

require(xts)
require(tis)
require(lubridate)
require(RColorBrewer)
require(xtable)
require(stargazer)
require(sandwich)


# LOAD DATA ----
gaps <- readRDS(paste0(data.loc, 'All-gap-vintages-Nov-2023.Rds'))

gaps_main_results = gaps[-c(8,9,11)] #remove MW(2020) models and UC(u)

gaps_figure = gaps[-c(1,2,4,6,8,9,10,11,12,13,14)] #remove linear, quad, GLMS, MW, MW(2020) models, UC(GDP), UC(u), UC(GDP/pi) UC(GDP/CU), UC(GDP/u)

gaps_tb_cbo = gaps[-c(1,2,3,4,5,6,7,8,9,10,11,12,13,14,15)] #remove all but TB and CBO

# MORE SETTINGS ----
# Some things for figures 
to.pdf <- function (expr, filename, width=16, height=12, point=10) {
  cat(sprintf("Creating %s\n", filename))
  pdf(filename, width = width/2.54, height = height/2.54, point = point)
  on.exit(dev.off())
  eval.parent(substitute(expr))
}
trim.pdf <- function(filename) {
  system(paste('pdfcrop ', filename, filename)) 
}

# Define palette for all figures to keep colors straight
palette <- c(brewer.pal(name='PiYG', n=length(gaps_figure)-2), 
             'darkcyan', 'gold1') # Replace yellow so it's easier to see, and give Tealbook "teal"
names(palette) <- names(gaps_figure)

palette2 <- c(brewer.pal(name='Paired', n=length(gaps_main_results)-3), 
             'gold1', 'darkcyan', 'maroon') # Replace yellow so it's easier to see, and give Tealbook "teal"
names(palette2) <- names(gaps_main_results)


# FIGURE 1 ------
plot.some.rt <- function() {
  
  t <- as.numeric(index(gaps_figure[['hp']]))
  
  plot(t,gaps_figure[['hp']][,'rt'], type='n', axes=FALSE, xlab='', ylab='',ylim=c(-16,8))
  nberShade()
  abline(h=0)
  for (i in 1:length(names(gaps_figure))) {
    lines(t, gaps_figure[[names(gaps_figure)[i]]][,'rt'], col=palette[i])
  }
  
  axis(1, at=seq(1970,2022,10), labels=seq(1970,2022,10), cex.axis=.8)
  axis(1, at=seq(1965,2022,5), labels=FALSE, tck=-0.02)
  axis(2, cex.axis=.8)

  legend(x=1983, y=-9, bty='n', 
         legend=c('HP', 'Hamilton','KMW', 
                  expression(paste('UC (GDP/', pi, '/', 'u.)')),
                  'Tealbook', 
                  'CBO'),
         col=palette, pch=c(16), pt.cex=1.2,
         cex=.75, ncol=3, text.width=8)
  box()
  title('%', adj=1, font.main=1, cex.main=.8, line=.2)
}
to.pdf(plot.some.rt(), filename=paste0(figures.loc, 'Fig-1.pdf'), height=10)
trim.pdf(filename='../Figures/Fig-1.pdf')

# TABLE 1-3 (Summary stats of revisions) --------
gap.stats <- function(y, first, last) {
  
  if ((!missing(first)) && (!missing(last))) {
    dates <- paste(as.Date(as.yearqtr(first)), as.Date(as.yearqtr(last)), sep='/')
    y <- y[dates]
  }
  y['1995-10-01'] <- NA
  y[which(is.na(y[,'rt'])), ] <- NA # Sets QRT, finals to 0 for CU if RT is missing
  
  N <- sum(!is.na(y))
  if (N>0) {
    mean <- mean(y, na.rm=TRUE)
    sd <- sd(y, na.rm=TRUE)
    min <- min(y, na.rm=TRUE)
    max <- max(y, na.rm=TRUE)
    frac.gz <- mean(I(y)>0, na.rm=TRUE)*100
  } else {
    mean <- NA
    sd <- NA
    min <- NA
    max <- NA
    frac.gz <- NA
  }
  
  output <- list(N=format(round(N),nsmall=0), 
                 mean=format(round(mean,2), nsmall=2), 
                 sd=format(round(sd,2), nsmall=2),
                 min=format(round(min,2), nsmall=2),
                 max=format(round(max,2), nsmall=2),
                 frac.gz=format(round(frac.gz, 0), nsmall=0))
  return(output)
} # Output gap statistics
rev.stats <- function(y.hat, y, first, last) {
  
  if ((!missing(first)) & (!missing(last))) {
    dates <- paste(as.Date(as.yearqtr(first)), as.Date(as.yearqtr(last)), sep='/')
    y <- y[dates]
    y.hat <- y.hat[dates]
    y['1995-10-01'] <- NA # Set to NA for all
    y.hat['1995-10-01'] <- NA # Set to NA for all
    y[which(is.na(y.hat)), ] <- NA # Sets QRT, finals to NA for CU if RT is missing
    
    x <- y - y.hat # Final less real-time
  } else {
    y['1995-10-01'] <- NA # Set to NA for all
    y.hat['1995-10-01'] <- NA # Set to NA for all
    y[which(is.na(y.hat)), ] <- NA # Sets QRT, finals to NA for CU if RT is missing
    
    x <- y - y.hat
  }

  N <- sum(!is.na(x))
  if (N>0) {
    mean <- mean(x, na.rm=TRUE)
    sd <- sd(x, na.rm=TRUE)
    min <- min(x, na.rm=TRUE)
    max <- max(x, na.rm=TRUE)
    rmse <- sqrt(mean(x^2, na.rm=TRUE))
    ns.sd <- sd / sd(y, na.rm=TRUE)
    ns.rmse <- sqrt(mean(x^2, na.rm=TRUE)) / sd(y, na.rm=TRUE)
    corr <- cor(y.hat, y, use='pairwise.complete.obs')
    sign <- mean(sign(y.hat)==sign(y), na.rm=TRUE)
  } else {
    mean <- sd <- min <- max <- rmse <- ns.sd <- ns.rmse <- corr <- sign <- NA
  }
  
  output <- list(N=format(N, nsmall=0),
                 mean=format(round(mean, 2), nsmall=2),
                 sd=format(round(sd, 2), nsmall=2), 
                 rmse=format(round(rmse, 2), nsmall=2), 
                 min=format(round(min, 2), nsmall=2), 
                 max=format(round(max, 2), nsmall=2),
                 ns.sd=format(round(ns.sd, 2), nsmall=2), 
                 ns.rmse=format(round(ns.rmse, 2), nsmall=2),
                 corr=format(round(corr, 2), nsmall=2), 
                 sign=format(round(sign*100), nsmall=0))
  return(output)
} # Revision statistics

# Table 1
t(sapply(gaps_main_results, function(x) gap.stats(x$rt, '1965 Q3', '2022 Q1')))
tab.1 <- t(sapply(gaps_main_results, function(x) gap.stats(x$rt, '1965 Q3', '2022 Q1')))
rownames(tab.1) <- c('Linear', 'Quadratic', 'HP', 'GLMS', 'Hamilton', 'MW', 'KMW',
                 'UC (GDP)', 'UC (GDP/$\\pi$)', 'UC (GDP/CU)', 'UC (GDP/u.)', 'UC (GDP/u./$\\pi$)',
                 'Tealbook', 'CBO')
stargazer(tab.1)

short.tab.1 <- t(sapply(gaps_tb_cbo, function(x) gap.stats(x$rt, '1998 Q1', '2016 Q3')))
rownames(short.tab.1) <- c('Tealbook', 'CBO')
stargazer(short.tab.1)

# Table 2 
t(sapply(gaps_main_results, function(x) rev.stats(x$rt, x$final, '1965 Q3', '2022 Q1')))
tab.2 <- t(sapply(gaps_main_results, function(x) rev.stats(x$rt, x$final, '1965 Q3', '2022 Q1')))
rownames(tab.2) <- c('Linear', 'Quadratic', 'HP', 'GLMS', 'Hamilton', 'MW', 'KMW',
                 'UC (GDP)', 'UC (GDP/$\\pi$)', 'UC (GDP/CU)', 'UC (GDP/u.)', 'UC (GDP/u./$\\pi$)',
                 'Tealbook', 'CBO')
tab.2 = tab.2[, c('N', 'mean', 'sd', 'rmse', 'ns.sd', 'ns.rmse', 'sign')]
stargazer(tab.2)


# Full sample using final 2016
short.tab.2 <- t(sapply(gaps_tb_cbo, function(x) rev.stats(x$rt, x$final.2016, '1965 Q1', '2016 Q3')))
rownames(short.tab.2) <- c('Tealbook', 'CBO')
short.tab.2 = short.tab.2[, c('N', 'mean', 'sd', 'rmse', 'ns.sd', 'ns.rmse', 'sign')]
stargazer(short.tab.2)

# Short sample using final 2016
short.tab.2 <- t(sapply(gaps_tb_cbo, function(x) rev.stats(x$rt, x$final.2016, '1998 Q1', '2016 Q3')))
rownames(short.tab.2) <- c('Tealbook', 'CBO')
short.tab.2=short.tab.2[, c('N', 'mean', 'sd', 'rmse', 'ns.sd', 'ns.rmse', 'sign')]
stargazer(short.tab.2)

# Tab 3
rev.stats_tab3 <- function(y.hat, y, denom, first, last) {
  
  if ((!missing(first)) & (!missing(last))) {
    dates <- paste(as.Date(as.yearqtr(first)), as.Date(as.yearqtr(last)), sep='/')
    y <- y[dates]
    y.hat <- y.hat[dates]
    y['1995-10-01'] <- NA # Set to NA for all
    y.hat['1995-10-01'] <- NA # Set to NA for all
    y[which(is.na(y.hat)), ] <- NA # Sets QRT, finals to NA for CU if RT is missing
    
    x <- y - y.hat # Final less real-time
  } else {
    y['1995-10-01'] <- NA # Set to NA for all
    y.hat['1995-10-01'] <- NA # Set to NA for all
    y[which(is.na(y.hat)), ] <- NA # Sets QRT, finals to NA for CU if RT is missing
    
    x <- y - y.hat
  }
  
  N <- sum(!is.na(x))
  if (N>0) {
    mean <- mean(x, na.rm=TRUE)
    sd <- sd(x, na.rm=TRUE)
    min <- min(x, na.rm=TRUE)
    max <- max(x, na.rm=TRUE)
    rmse <- sqrt(mean(x^2, na.rm=TRUE))
    ns.sd <- sd / denom
    ns.rmse <- sqrt(mean(x^2, na.rm=TRUE)) / denom
    corr <- cor(y.hat, y, use='pairwise.complete.obs')
    sign <- mean(sign(y.hat)==sign(y), na.rm=TRUE)
  } else {
    mean <- sd <- min <- max <- rmse <- ns.sd <- ns.rmse <- corr <- sign <- NA
  }
  
  output <- list(N=format(N, nsmall=0),
                 mean=format(round(mean, 2), nsmall=2),
                 sd=format(round(sd, 2), nsmall=2), 
                 rmse=format(round(rmse, 2), nsmall=2), 
                 min=format(round(min, 2), nsmall=2), 
                 max=format(round(max, 2), nsmall=2),
                 ns.sd=format(round(ns.sd, 2), nsmall=2), 
                 ns.rmse=format(round(ns.rmse, 2), nsmall=2),
                 corr=format(round(corr, 2), nsmall=2), 
                 sign=format(round(sign*100), nsmall=0))
  return(output)
} # Revision statistics

nber <- list()
nber$peak <- c('1960 Q2', '1969 Q4', '1973 Q4', '1980 Q1', '1981 Q3', '1990 Q3', '2001 Q1', '2007 Q4', '2019 Q4')
nber$trough <- c('1961 Q1', '1970 Q4', '1975 Q1', '1980 Q3', '1982 Q4', '1991 Q1', '2001 Q4', '2009 Q2', '2020 Q2')

gaps.peak <- lapply(gaps_main_results, function(x) x[paste(as.Date(as.yearqtr(nber$peak)-3), 
                                                           as.Date(as.yearqtr(nber$peak)), sep='/')])
gaps.trough <- lapply(gaps_main_results, function(x) x[paste(as.Date(as.yearqtr(nber$trough)-3), 
                                                             as.Date(as.yearqtr(nber$trough)), sep='/')])

gaps.peak$line$sd = sd(gaps_main_results$line, na.rm=TRUE)
gaps.peak$quad$sd = sd(gaps_main_results$quad, na.rm=TRUE)
gaps.peak$hp$sd = sd(gaps_main_results$hp, na.rm=TRUE)
gaps.peak$hp.exp$sd = sd(gaps_main_results$hp.exp, na.rm=TRUE)
gaps.peak$ham$sd = sd(gaps_main_results$ham, na.rm=TRUE)
gaps.peak$mw$sd = sd(gaps_main_results$mw, na.rm=TRUE)
gaps.peak$kmw$sd = sd(gaps_main_results$kmw, na.rm=TRUE)
gaps.peak$uc1$sd = sd(gaps_main_results$uc1, na.rm=TRUE)
gaps.peak$uc2.pi$sd = sd(gaps_main_results$uc2.pi, na.rm=TRUE)
gaps.peak$uc2.cu$sd = sd(gaps_main_results$uc2.cu, na.rm=TRUE)
gaps.peak$uc2.u$sd = sd(gaps_main_results$uc2.u, na.rm=TRUE)
gaps.peak$uc3$sd = sd(gaps_main_results$uc3, na.rm=TRUE)
gaps.peak$tb$sd = sd(gaps_main_results$tb, na.rm=TRUE)
gaps.peak$cbo$sd = sd(gaps_main_results$cbo, na.rm=TRUE)

gaps.trough$line$sd = sd(gaps_main_results$line, na.rm=TRUE)
gaps.trough$quad$sd = sd(gaps_main_results$quad, na.rm=TRUE)
gaps.trough$hp$sd = sd(gaps_main_results$hp, na.rm=TRUE)
gaps.trough$hp.exp$sd = sd(gaps_main_results$hp.exp, na.rm=TRUE)
gaps.trough$ham$sd = sd(gaps_main_results$ham, na.rm=TRUE)
gaps.trough$mw$sd = sd(gaps_main_results$mw, na.rm=TRUE)
gaps.trough$kmw$sd = sd(gaps_main_results$kmw, na.rm=TRUE)
gaps.trough$uc1$sd = sd(gaps_main_results$uc1, na.rm=TRUE)
gaps.trough$uc2.pi$sd = sd(gaps_main_results$uc2.pi, na.rm=TRUE)
gaps.trough$uc2.cu$sd = sd(gaps_main_results$uc2.cu, na.rm=TRUE)
gaps.trough$uc2.u$sd = sd(gaps_main_results$uc2.u, na.rm=TRUE)
gaps.trough$uc3$sd = sd(gaps_main_results$uc3, na.rm=TRUE)
gaps.trough$tb$sd = sd(gaps_main_results$tb, na.rm=TRUE)
gaps.trough$cbo$sd = sd(gaps_main_results$cbo, na.rm=TRUE)

peaks <- t(sapply(gaps.peak, function(x) rev.stats_tab3(x$rt, x$final, denom=x$sd[[1]])))
troughs <- t(sapply(gaps.trough, function(x) rev.stats_tab3(x$rt, x$final, denom=x$sd[[1]])))
tab.3 <- cbind(peaks[,c('N','ns.sd','ns.rmse','sign')], 
               troughs[,c('N','ns.sd','ns.rmse','sign')])
rownames(tab.3) <- c('Linear', 'Quadratic', 'HP', 'GLMS', 'Hamilton', 'MW', 'KMW',
                     'UC (GDP)', 'UC (GDP/$\\pi$)', 'UC (GDP/CU)', 'UC (GDP/u.)', 'UC (GDP/u./$\\pi$)',
                     'Tealbook', 'CBO')
stargazer(tab.3)

# Tealbook long sample
peaks <- t(sapply(gaps.peak, function(x) rev.stats_tab3(x$rt, x$final.2016, denom=x$sd[[1]])))
troughs <- t(sapply(gaps.trough, function(x) rev.stats_tab3(x$rt, x$final.2016, denom=x$sd[[1]])))
tab.3.tb <- cbind(peaks[,c('N','ns.sd','ns.rmse','sign')], 
                  troughs[,c('N','ns.sd','ns.rmse','sign')])
rownames(tab.3.tb) <- c('Linear', 'Quadratic', 'HP', 'GLMS', 'Hamilton', 'MW', 'KMW', 
                        'UC (GDP)', 'UC (GDP/$\\pi$)', 'UC (GDP/CU)', 'UC (GDP/u.)', 'UC (GDP/u./$\\pi$)',
                        'Tealbook', 'CBO')
stargazer(tab.3.tb)

# Tealbook and CBO short sample
peaks <- t(sapply(gaps.peak, function(x) rev.stats_tab3(x$rt, x$final.2016, denom=x$sd[[1]], '1998 Q1', '2016 Q3')))
troughs <- t(sapply(gaps.trough, function(x) rev.stats_tab3(x$rt, x$final.2016, denom=x$sd[[1]], '1998 Q1', '2016 Q3')))
tab.3.tb.short <- cbind(peaks[,c('N','ns.sd','ns.rmse','sign')], 
                        troughs[,c('N','ns.sd','ns.rmse','sign')])
rownames(tab.3.tb.short) <- c('Linear', 'Quadratic', 'HP', 'GLMS', 'Hamilton', 'MW', 'KMW',
                              'UC (GDP)', 'UC (GDP/$\\pi$)', 'UC (GDP/CU)', 'UC (GDP/u.)', 'UC (GDP/u./$\\pi$)',
                              'Tealbook', 'CBO')
stargazer(tab.3.tb.short)


# FIGURE 2 --------------------------------------
load('../Data/Vintage-datasets-Nov-2023.Rda') # Produced in wrangle-vintage-data.R

gather.data <- FALSE
if (gather.data) {
  # UC1u 
  uc1u.rt <- readRDS('../Raw model output/rt-uc1u.Rds')
  # UC2 
  uc2.rt <- readRDS('../Raw model output/rt-uc2.u.Rds')
  
  temp = rowMeans(uc1u.rt[[226]]$res$S, dims = 2)
  temp2 = temp[,3]
  temp = rowMeans(uc2.rt[[226]]$res$S, dims = 2)
  uc2u_trend = temp[,3]
  uc2u_uerror = temp[,7]
  data = rt.vintages[[226]]$u[53:300]
  uc2u_cycle = data - uc2u_trend - uc2u_uerror
  uc1u_cycle = uc2u_cycle
  uc1u_cycle["1960-01-01/2021-10-01"] = temp2
  saveRDS(uc1u_cycle, file=paste0(data.loc, 'figure2_uc1.Rds'))
  saveRDS(uc2u_cycle, file=paste0(data.loc, 'figure2_uc2.Rds'))
} else {
  uc1u_cycle = readRDS(paste0(data.loc, 'figure2_uc1.Rds'))
  uc2u_cycle = readRDS(paste0(data.loc, 'figure2_uc2.Rds'))
}




plot.ugaps <- function() {
  
  t <- as.numeric(index(uc1u_cycle))
  
  plot(t, uc2u_cycle, type='n', axes=FALSE, xlab='', ylab='', ylim=c(-2.5, 6))
  nberShade()
  abline(h=0)
  
  points(t, uc1u_cycle, col='navy', pch=16)
  points(t, uc2u_cycle, col='darkorange3', pch=16)
  
  axis(1, at=seq(1960, 2020, 10), cex.axis=.9)
  axis(1, at=seq(1960, 2020, 10), labels=FALSE, tck=-.02)  
  axis(2, at=seq(-2, 6, 2), cex.axis=.9)
  axis(4, at=seq(-2, 6, 2), labels=FALSE)  
  
  legend(1961, 5.9, c('UC1 (u.)', 'UC2 (GDP/u.)'), col=c('navy', 'darkorange3'), pch=16,
         bty='n', ncol=1)
  
  box()
  title('Sequence of real-time unemployment gap estimates', adj=0, line=.2, cex.main=.8)
  title('%', adj=1, font.main=1, line=.2, cex.main=.8)
}

to.pdf(plot.ugaps(), '../Figures/Figure2.pdf', height=10)
trim.pdf('../Figures/Figure2.pdf')



# FIGURE 3 --------------------------------------
decomp.loc <- '../Decomposition/'
vintage.dates <- index(gaps[[1]])
taus <- 1:length(vintage.dates)
V <- length(taus)

# Figure 3a
temp <- read.table(paste0(decomp.loc, 'uc2pi_decomp.txt'))
data.names <- c('lgdp', 'pi', 'c')
i <- which(vintage.dates=='1995 Q4')

uc2.pi <- xts(rbind(temp[1:(i-1), paste0('rt.',data.names)], 
                    rep(NA,4), 
                    temp[i:(V-1), paste0('rt.',data.names)]), 
              order.by=vintage.dates)
colnames(uc2.pi) <- data.names

fig.uc2.pi <- function() {

  t <- as.numeric(index(uc2.pi))
  plot(t, uc2.pi[,'c'], type='n', axes=FALSE, xlab='', ylab='', ylim=c(-10,6))
  
  nberShade()
  abline(h=0)
  for (i in c(1:121, 123:V)) { # Skip 1995 Q4
    polygon(x=c(t[i]-1/10, t[i]-1/10, t[i]+1/10, t[i]+1/10), 
            y=c(0,matrix(uc2.pi[i,1]),matrix(uc2.pi[i,1]),0), col='navy', border=NA)

    if (sign(uc2.pi[i,1])==sign(uc2.pi[i,2])) {
      polygon(x=c(t[i]-1/10, t[i]-1/10, t[i]+1/10, t[i]+1/10), 
            y=c(matrix(uc2.pi[i,1]), matrix(uc2.pi[i,1])+matrix(uc2.pi[i,2]),
                matrix(uc2.pi[i,1])+matrix(uc2.pi[i,2]), matrix(uc2.pi[i,1])), col='chartreuse2', border=NA)    
    } else {
      polygon(x=c(t[i]-1/10, t[i]-1/10, t[i]+1/10, t[i]+1/10), 
              y=c(0, matrix(uc2.pi[i,2]),
                  matrix(uc2.pi[i,2]), 0), col='chartreuse2', border=NA)
    }
  }
  points(t, (uc2.pi[,1]+uc2.pi[,2]), pch=16)

  axis(1, at=seq(1970,2020,10), labels=seq(1970,2020,10), cex.axis=.8)
  axis(1, at=seq(1965,2020,5), labels=FALSE, tck=-0.02)
  axis(2, at=seq(-10,6,2), labels=seq(-10,6,2), cex.axis=.8)
  axis(4, at=seq(-10,6,2), labels=FALSE)
  
  legend(x=2004, y=5.75,
         legend=c('Real GDP', expression(pi), 'Output gap'),
         col=c('navy', 'chartreuse2', 'black'),
         pch=c(15,15,16),
         bty='n', cex=1)
  box()
}
fig.uc2.pi()
to.pdf(fig.uc2.pi(), '../Figures/Fig-3a.pdf')
trim.pdf('../Figures/Fig-3a.pdf')



# Figure 3b
. <- read.table(paste0(decomp.loc, 'uc2cu_decomp.txt'))
temp <- xts(., order.by=as.yearqtr(rownames(.)))[,1:3]
data.names <- c('lgdp', 'cu', 'c')
i <- which(vintage.dates=='1995 Q4')

empty <- xts(, order.by=as.yearqtr(vintage.dates))
uc2.cu <- merge(empty, temp)
colnames(uc2.cu) <- data.names

fig.uc2.cu <- function() {
  
  X <- uc2.cu
  
  t <- as.numeric(index(X))
  plot(t, X[,'c'], type='n', axes=FALSE, xlab='', ylab='', ylim=c(-10,6))
  
  nberShade()
  abline(h=0)
  for (i in which(!is.na(X[,1]))) { # Skip 1995 Q4
    polygon(x=c(t[i]-1/10, t[i]-1/10, t[i]+1/10, t[i]+1/10), 
            y=c(0,matrix(X[i,1]),matrix(X[i,1]),0), col='navy', border=NA)
    
    if (sign(X[i,1])==sign(X[i,2])) {
      polygon(x=c(t[i]-1/10, t[i]-1/10, t[i]+1/10, t[i]+1/10), 
              y=c(matrix(X[i,1]), matrix(X[i,1])+matrix(X[i,2]),
                  matrix(X[i,1])+matrix(X[i,2]), matrix(X[i,1])), col='darkorange3', border=NA)    
    } else {
      polygon(x=c(t[i]-1/10, t[i]-1/10, t[i]+1/10, t[i]+1/10), 
              y=c(0, matrix(X[i,2]),
                  matrix(X[i,2]), 0), col='darkorange3', border=NA)
    }
  }
  points(t, (X[,1]+X[,2]), pch=16)
  
  axis(1, at=seq(1970,2020,10), labels=seq(1970,2020,10), cex.axis=.8)
  axis(1, at=seq(1965,2020,5), labels=FALSE, tck=-0.02)
  axis(2, at=seq(-10,6,2), labels=seq(-10,6,2), cex.axis=.8)
  axis(4, at=seq(-10,6,2), labels=FALSE)
  
  legend(x=2004, y=5.75,
         legend=c('Real GDP', 'Cap utilization', 'Output gap'),
         col=c('navy', 'darkorange3', 'black'),
         pch=c(15,15,16),
         bty='n')
  box()
}
fig.uc2.cu()
to.pdf(fig.uc2.cu(), '../Figures/Fig-3b.pdf')
trim.pdf('../Figures/Fig-3b.pdf')


# Figure 3c
temp <- read.table(paste0(decomp.loc, 'uc2u_decomp.txt'))
data.names <- c('lgdp','u', 'c')
i <- which(vintage.dates=='1995 Q4')

uc2.u <- xts(rbind(temp[1:(i-1), paste0('rt.',data.names)], 
                    rep(NA,4), 
                    temp[i:(V-1), paste0('rt.',data.names)]), 
              order.by=vintage.dates)
colnames(uc2.u) <- data.names

fig.uc2.u <- function() {
  
  X <- uc2.u
  
  t <- as.numeric(index(X))
  plot(t, X[,'c'], type='n', axes=FALSE, xlab='', ylab='', ylim=c(-10,6))
  
  nberShade()
  abline(h=0)
  for (i in which(!is.na(X[,1]))) { # Skip 1995 Q4
    polygon(x=c(t[i]-1/10, t[i]-1/10, t[i]+1/10, t[i]+1/10), 
            y=c(0,matrix(X[i,1]),matrix(X[i,1]),0), col='navy', border=NA)
    
    if (sign(X[i,1])==sign(X[i,2])) {
      polygon(x=c(t[i]-1/10, t[i]-1/10, t[i]+1/10, t[i]+1/10), 
              y=c(matrix(X[i,1]), matrix(X[i,1])+matrix(X[i,2]),
                  matrix(X[i,1])+matrix(X[i,2]), matrix(X[i,1])), col='red', border=NA)    
    } else {
      polygon(x=c(t[i]-1/10, t[i]-1/10, t[i]+1/10, t[i]+1/10), 
              y=c(0, matrix(X[i,2]),
                  matrix(X[i,2]), 0), col='red', border=NA)
    }
  }
  points(t, (X[,1]+X[,2]), pch=16)
  
  axis(1, at=seq(1970,2020,10), labels=seq(1970,2020,10), cex.axis=.8)
  axis(1, at=seq(1965,2020,5), labels=FALSE, tck=-0.02)
  axis(2, at=seq(-10,6,2), labels=seq(-10,6,2), cex.axis=.8)
  axis(4, at=seq(-10,6,2), labels=FALSE)
  
  legend(x=2004, y=5.75, 
         legend=c('Real GDP', 'Unemployment rate', 'Output gap'),
         col=c('navy', 'red', 'black'),
         pch=c(15,15,16),
         bty='n')
  box()
}
fig.uc2.u()
to.pdf(fig.uc2.u(), '../Figures/Fig-3c.pdf')
trim.pdf('../Figures/Fig-3c.pdf')



# Figure 3d
temp <- read.table(paste0(decomp.loc, 'uc3_decomp.txt'))
data.names <- c('lgdp','u', 'pi', 'c')
i <- which(vintage.dates=='1995 Q4')

uc3 <- xts(rbind(temp[1:(i-1), paste0('rt.',data.names)], 
                   rep(NA,4), 
                   temp[i:(V-1), paste0('rt.',data.names)]), 
             order.by=vintage.dates)
colnames(uc3) <- data.names

fig.uc3 <- function() {
  
  X <- uc3
  
  t <- as.numeric(index(X))
  plot(t, X[,'c'], type='n', axes=FALSE, xlab='', ylab='', ylim=c(-10,6))
  
  nberShade()
  abline(h=0)
  for (i in which(!is.na(X[,1]))) { # Skip 1995 Q4
    polygon(x=c(t[i]-1/10, t[i]-1/10, t[i]+1/10, t[i]+1/10), 
            y=c(0,matrix(X[i,1]),matrix(X[i,1]),0), col='navy', border=NA)
    
    if (sign(X[i,1])==sign(X[i,2])) {
      y1 <- matrix(X[i,1])
      y2 <- matrix(X[i,1]+X[i,2])

      polygon(x=c(t[i]-1/10, t[i]-1/10, t[i]+1/10, t[i]+1/10), 
              y=c(y1, y2, y2,y1), col='red', border=NA)    
    } else {
      y <- matrix(X[i,2])
      polygon(x=c(t[i]-1/10, t[i]-1/10, t[i]+1/10, t[i]+1/10), 
              y=c(0, y, y, 0), col='red', border=NA)
    }
    
    if (sign(X[i,1]+X[i,2]) == sign(X[i,3])) {
      y1 <- matrix(X[i,1]+X[i,2])
      y2 <- matrix(X[i,1]+X[i,2]+X[i,3])
      
      polygon(x=c(t[i]-1/10, t[i]-1/10, t[i]+1/10, t[i]+1/10), 
              y=c(y1, y2, y2, y1), col='chartreuse2', border=NA)  
    } else {
      y <- matrix(X[i,3])
      polygon(x=c(t[i]-1/10, t[i]-1/10, t[i]+1/10, t[i]+1/10), 
              y=c(0, y, y, 0), col='chartreuse2', border=NA)
    }
  }
  points(t, (X[,1]+X[,2]+X[,3]), pch=16)
  
  axis(1, at=seq(1970,2020,10), labels=seq(1970,2020,10), cex.axis=.8)
  axis(1, at=seq(1965,2020,5), labels=FALSE, tck=-0.02)
  axis(2, at=seq(-10,6,2), labels=seq(-10,6,2), cex.axis=.8)
  axis(4, at=seq(-10,6,2), labels=FALSE)
  
  legend(x=2004, y=5.75, 
         legend=c('Real GDP', 'Unemployment rate', expression(pi), 'Output gap'),
         col=c('navy', 'red', 'chartreuse2', 'black'),
         pch=c(15,15,15,16),
         bty='n')
  box()
}
fig.uc3()
to.pdf(fig.uc3(), '../Figures/Fig-3d.pdf')
trim.pdf('../Figures/Fig-3d.pdf')


# TABLE 5 (inflation GDP Deflator 1970-2022)  ------
# First we need to gather all the real-time vintages
scratch.loc <- '../Raw model output/'
gather.vintages <- FALSE
if (gather.vintages) {
  
  # These are useful for here but won't last beyond this section
  vintages <- as.yearqtr(seq(from=as.Date(as.yearqtr('1965 Q4')), to=as.Date(as.yearqtr('2022 Q2')), by='quarter'))
  quarters <- seq(from=as.Date('1960-01-01'), to=as.Date('2022-01-01'), by='quarter')
  V <- length(vintages) # Number of vintages
  T <- length(quarters) # Time-series length
  
  get.mean.uc.gap <- function(model='uc1') {
    
    temp <- readRDS(paste0(scratch.loc, 'rt-', model, '.Rds'))
    
    rt.gap <- xts(array(NA, c(T, V)), order.by=as.yearqtr(quarters))
    colnames(rt.gap) <- paste0(vintages)
    
    c <- which(temp[[1]][['prior']][['state.names']] == 'c') # Picks off correct column of res[['S']]
    
    for (v in 1:V) {
      if ((v==1 && length(temp[[v]])==6) | length(temp[[v]])==2) { # This prevents a crash for 1996Q1 vintage and early CU vintages
        .S <- apply(temp[[v]][['res']][['S']], c(1,2), mean)
        .c <- xts(.S[,c], order.by=seq(from=as.Date(first(quarters)),
                                       length.out=dim(.S)[1],
                                       by='quarter'))
        rt.gap[index(.c),v] <- .c
      }
    }
    
    return(rt.gap)
  }
  
  uni.vints <- readRDS(paste0(data.loc, 'uni-rt-gap-vintages.Rds'))
  gap.vints <- lapply(uni.vints, function(x) 'colnames<-'(x, paste0(vintages))) # Naming columns to vintages as.yearqtr
  gap.vints[['uc1']] <- get.mean.uc.gap('uc1')
  gap.vints[['uc1u']] <- get.mean.uc.gap('uc1u')
  gap.vints[['uc2.cu']] <- get.mean.uc.gap('uc2.cu')
  gap.vints[['uc2.pi']] <- get.mean.uc.gap('uc2.pi')
  gap.vints[['uc2.u']] <- get.mean.uc.gap('uc2.u')
  gap.vints[['uc3']] <- get.mean.uc.gap('uc3')
  
  
  # Get Tealbook vintages
  all.tb.gaps <- xts(read.csv(paste0(data.loc, 'Greenbook_Output_Gap_Aug_2022.csv'))[,-1],
                     order.by=seq(from=as.Date('1975-01-01'), 
                                  to=as.Date('2019-10-01'),
                                  by='quarter'))
  all.tb.dates <- as.Date(substr(colnames(all.tb.gaps), 7,12), '%y%m%d')
  
  
  # Select the Tealbook that is first in the quarter
  i <- as.Date(c(by(all.tb.dates, as.yearqtr(all.tb.dates), min, simplify=TRUE)))
  early.tb.gaps <- all.tb.gaps[,which(all.tb.dates %in% i)]
  
  # Put in same format as others
  gap.vints[['tb']] <- xts(array(NA, c(T,V)), order.by=quarters)
  colnames(gap.vints[['tb']]) <- paste0(vintages)
  
  cols <- which(colnames(gap.vints[['tb']]) %in% paste0(as.yearqtr(as.Date(i))))
  gap.vints[['tb']][index(all.tb.gaps), cols] <- early.tb.gaps
  
  
  
  # Get CBO vintages
  substr.right <- function(x,n) {
    substr(x, nchar(x)-n+1, nchar(x))
  } 
  
  # Vintages of nominal potential GDP
  . <- data.frame(readxl::read_excel(paste0(data.loc, 'All-CBO-nominal-potential.xlsx')))
  npot <- xts(.[,-1], seq(from=as.Date(as.yearqtr('1949 Q1')),
                          to=as.Date(as.yearqtr('2032 Q4')), 
                          by='quarter'))
  colnames(npot) <- paste(substr(colnames(npot), start=1, stop=3), substr.right(colnames(npot), 4))
  
  # Vintages of nominal GDP
  . <- data.frame(readxl::read_excel(paste0(data.loc, 'noutputMvQd.xlsx')))
  ngdp <- xts(.[,-1], seq(from=as.Date(as.yearqtr('1947 Q1')),
                          to=as.Date(as.yearqtr('2022 Q2')), 
                          by='quarter'))
  colnames(ngdp) <- paste0(as.yearmon(seq(from=as.Date('1965-11-01'), to=as.Date('2022-08-01'), by='month')))
  
  # NGDP forecasts from CBO when available
  . <- data.frame(readxl::read_excel(paste0(data.loc, 'All-CBO-GDP-forecast.xlsx')))
  ngdp.hat <- xts(.[,-1], seq(from=as.Date(as.yearqtr('2009 Q1')),
                              to=as.Date(as.yearqtr('2032 Q4')), 
                              by='quarter'))
  colnames(ngdp.hat) <- paste(substr(colnames(ngdp.hat), start=1, stop=3), substr.right(colnames(ngdp.hat), 4))
  
  
  # Create quarterly vintages since Feb 1991
  index <- as.yearqtr(seq(as.Date(as.yearqtr('1947 Q1')),
                          as.Date(as.yearqtr('2032 Q4')), 
                          by='quarter'))
  cbo.vintages <- as.yearmon(seq(from=as.Date('1991-02-01'), to=as.Date('2022-05-01'), by='quarter'))
  cbo.gaps <- xts(array(NA, c(length(index), length(cbo.vintages))), order.by=index)
  colnames(cbo.gaps) <- paste(as.yearmon(cbo.vintages))
  
  for (v in 1:length(cbo.vintages)) {
    
    # Column selectors for pot, ngdp
    c.pot <- max(which(as.yearmon(colnames(npot))<=cbo.vintages[v]))
    c.ngdp <- which(colnames(ngdp) == cbo.vintages[v])
    
    if (cbo.vintages[v] < as.yearmon(colnames(ngdp.hat)[1])) {
      
      X <- cbind(npot[,c.pot], ngdp[,c.ngdp]); colnames(X) <- c('pot','ngdp') # cbind just to be sure dates line up
      cbo.gaps[,v] <- 100*(log(X[,'ngdp'] / X[,'pot']))
      
    } else {
      
      c.ngdp.hat <- max(which(colnames(ngdp.hat) <= as.yearmon(cbo.vintages[v])))  # Column of ngdp.hat
      X <- cbind(npot[,c.pot], ngdp[,c.ngdp], ngdp.hat[,c.ngdp.hat], ngdp.hat[,c.ngdp.hat] / stats::lag(ngdp.hat[,c.ngdp.hat])) 
      colnames(X) <- c('pot','ngdp','ngdp.hat', 'gr')
      
      # Grow ngdp forward
      last <- max(which(!is.na(X[,'ngdp'])))
      for (i in (last+1):dim(X)[1]) {
        X[i,'ngdp'] <- matrix(X[(i-1),'ngdp'])*matrix(X[i,'gr'])
      }
      cbo.gaps[,v] <- 100*(log(X[,'ngdp'] / X[,'pot']))
    }
  }
  colnames(cbo.gaps) <- paste0(as.yearqtr(as.yearmon(colnames(cbo.gaps)))) # to match "gap.vints"
  
  # Now gather into the "usual" format
  gap.vints[['cbo']] <- xts(array(NA, c(T,V)), order.by=quarters)
  colnames(gap.vints[['cbo']]) <- paste0(vintages)
  
  gap.vints[['cbo']][, which(paste0(vintages) %in% colnames(cbo.gaps))] <- cbo.gaps[quarters, ]
  
  
  # Arrange into order shown in paper
  gap.vints <- gap.vints[c('line','quad','hp','hp.exp', 'ham','mw','kmw',
                           'uc1','uc1u','uc2.pi','uc2.cu','uc2.u','uc3', 'tb', 'cbo')] 
  
  rm(list=setdiff(ls(), c('code.loc', 'data.loc', 'figures.loc', 'scratch.loc',
                          'est.start', 'oos.start', 'oos.end', 'gap.vints')))
  
  saveRDS(gap.vints, file=paste0(data.loc, 'gap-vints-for-inflation.Rds'))
  
} else {
  gap.vints <- readRDS(paste0(data.loc, 'gap-vints-for-inflation.Rds'))
}
# Construct matrix of final gaps
qtrs <- index(gap.vints[[1]])
vints <- colnames(gap.vints[[1]])
T <- length(qtrs)
V <- length(vints)
final.gaps <- list()
final.gaps.2016 <- list() # This is used in appendix table a.6 (ends in 2016Q4)
for (i in names(gap.vints)) {
  final.gaps[[i]] <- xts(array(NA, c(T,V)), order.by=as.yearqtr(qtrs))
  colnames(final.gaps[[i]]) <- vints
  
  final.gaps.2016[[i]] <- xts(array(NA, c(T,V)), order.by=as.yearqtr(qtrs))
  colnames(final.gaps.2016[[i]]) <- vints
  
  for (v in 1:V) {
    final.gaps[[i]][,v] <- gap.vints[[i]][,227] # 227 is 2022:Q2 vintage
    final.gaps[[i]][is.na(gap.vints[[i]][,v]), v] <- NA
    
    final.gaps.2016[[i]][,v] <- gap.vints[[i]][,205] # 205 is 2016:Q4 vintage 
    final.gaps.2016[[i]][is.na(gap.vints[[i]][,v]), v] <- NA
    
  }
}

# OOS exercises begin here 
P <- list() # Price level vintages
P[['pcexfe']] <- xts(data.frame(readxl::read_excel(paste0(data.loc, 'Vintages-PCONX.xlsx'))[,-1]),
                     seq(from=as.Date('1947-01-01'), to=as.Date('2022-01-01'), by='quarter'))
P[['pgdp']] <- xts(data.frame(readxl::read_excel(paste0(data.loc, 'Vintages-PGDP.xlsx'))[,-1]),
                   seq(from=as.Date('1947-01-01'), to=as.Date('2022-01-01'), by='quarter'))
# Standardize the column names a bit to make life easier below
colnames(P[['pcexfe']]) <- paste0(as.yearqtr(seq(from=as.Date('1996-01-01'), to=as.Date('2022-04-01'), by='quarter')))
colnames(P[['pgdp']]) <- paste0(as.yearqtr(seq(from=as.Date('1965-10-01'), to=as.Date('2022-04-01'), by='quarter')))
# Construct matrix of final inflation
final.P <- list()
for (i in names(P)) {
  final.P[[i]] <- xts(array(NA, c(T,V)), order.by=as.yearqtr(qtrs))
  colnames(final.P[[i]]) <- vints
}
for (v in 1:V) {
  last.q <- as.Date(as.yearqtr(colnames(final.P[['pcexfe']])[v])) %m+% months(-3)
  final.P[['pcexfe']][paste0('1960-01-01', '/', last.q),v] <- P[['pcexfe']][paste0('1960-01-01', '/', last.q), '2022 Q2']
  final.P[['pgdp']][paste0('1960-01-01', '/', last.q),v] <- P[['pgdp']][paste0('1960-01-01', '/', last.q), '2022 Q2']
}

# Real-time gaps, PGDP 

# Functions 
rt.ar.forecast.inflation <- function(P, hzn, est.start, oos.start, oos.end) {
  
  # Here we produce AR(6) forecasts from an unrestricted AR model
  
  first.vint <- as.yearqtr(oos.start) - (hzn/4)
  last.vint <- as.yearqtr(oos.end) - (hzn/4)
  all.vints <- paste0(as.yearqtr(seq(as.Date(first.vint), as.Date(last.vint), by='quarter')))
  
  # Storage for forecasts
  pi.hat <- xts(array(NA, length(all.vints)), 
                order.by=as.yearqtr(seq(from=as.Date(as.yearqtr(oos.start)), 
                                        to=as.Date(as.yearqtr(oos.end)), 
                                        by='quarter')))
  
  for (v in 1:length(all.vints)) {
    
    if (all.vints[v] == '1996 Q1') {
      pi.hat[v] <- NA # Damned shutdown
      
    } else {
      
      vintage <- paste0(all.vints[v])
      vm1 <- as.yearqtr(as.Date(as.yearqtr(vintage)) %m+% months(-1)) # End of information set
      
      P.v <- P[,vintage]
      
      y <- 400*(log(P.v) - stats::lag(log(P.v), 1)) # Backwards, quarterly
      
      # AR unrestricted model
      ar.mod <- ar(y[paste0(as.Date(as.yearqtr(est.start)), '/', as.Date(vm1))], aic=FALSE, order.max=6)
      y.hat <- predict(ar.mod, n.ahead=hzn+1) # Need "+1" because of data lags
      pi.hat[v] <- mean(y.hat[['pred']])
    }
  }
  
  return(pi.hat)
}
rt.gap.forecast.inflation <- function(gap, P, hzn, est.start, oos.start, oos.end) {
  
  # This function does fully real-time forecasts of inflation,
  #   looping over vintages of P (price level) and the output gap
  
  # The auxilliary model: pi^h_t = alpha + beta_1*pi_{t-1} + beta_2*pi_{t-2} + ...
  #                                beta_6*pi_{t-6} + gamma*gap_t + e_t
  
  # Gap - name of model to be used; in matrix of dimension TxV
  # P - Name of matrix of vintage *price levels*
  # hzn - Forecast horizon
  # est.start - Start of estimation period
  # oos.start - First quarter to be forecasted
  # oos.end - Final quarter to be forecasted
  
  all.obs <- paste0(as.Date(as.yearqtr(est.start)), '/', as.Date(as.yearqtr(oos.end)))
  first.vint <- as.yearqtr(oos.start) - (hzn/4)
  last.vint <- as.yearqtr(oos.end) - (hzn/4)
  all.vints <- paste0(as.yearqtr(seq(as.Date(first.vint), as.Date(last.vint), by='quarter')))
  
  # Storage for forecasts
  pi.hat <- xts(array(NA, length(all.vints)), 
                order.by=as.yearqtr(seq(from=as.Date(as.yearqtr(oos.start)), 
                                        to=as.Date(as.yearqtr(oos.end)), 
                                        by='quarter')))
  
  for (v in 1:length(all.vints)) {
    
    vintage <- paste0(all.vints[v])
    
    P.v <- P[,vintage]
    gap.v <- gap[,vintage]
    
    y <- 400/hzn*(stats::lag(log(P.v), -hzn) - log(P.v)) # Note this leads!!
    x <- 400*(log(P.v) - stats::lag(log(P.v), 1)) # Backwards, quarterly
    
    lag.x <- stats::lag(x, seq(1,6))
    colnames(lag.x) <- paste0('l', seq(1,6), '.pi')
    
    data <- cbind(y, stats::lag(gap.v,1), lag.x)
    colnames(data)[1] <- 'y'
    colnames(data)[2] <- 'l.gap'
    
    # Given the NAs in the data this isn't needed but it's more clear
    last.qtr <- as.Date(as.yearqtr(as.Date(as.yearqtr(vintage)) %m+% months(-(hzn*3+1))))
    
    if (sum(complete.cases(data[paste0(as.Date(as.yearqtr(est.start)), '/', last.qtr)]))>0) {
      reg <- lm(y~., data=data[paste0(as.Date(as.yearqtr(est.start)), '/', last.qtr)]) 
      pi.hat[v] <- as.numeric(predict(reg, newdata=data[as.yearqtr(vintage), -1]))
    } else {
      pi.hat[v] <- NA
    }
  }
  return(pi.hat)
}

rmse <- function(y, yhat) unname(sqrt(mean((y - yhat)^2, na.rm=TRUE)))
dm.test <- function(y, yhat, ynull) {
  # A little function that returns the t-stat and p-value for a DM test
  #   under squared-error loss
  # !! Note that d.loss is defined as "Model - Null Model", 
  #     so the "model" outperforms the null model when the t-stat is < 0
  
  L.null <- (y - ynull)^2
  L <- (y - yhat)^2
  d.L <- L - L.null
  
  i <- which(!is.na(d.L))
  N <- length(i)
  
  reg <- lm(d.L[i]~1)
  d.bar <- reg$coefficients
  sig2 <- unname(NeweyWest(reg))
  
  dm <- unname(d.bar / sqrt(sig2))
  pval <- dnorm(dm)
  
  output <- c(d.bar, dm, pval, N)
  names(output) <- c('d.bar', 'DM', 'p.value', 'N')
  
  return(output)
}
rmse.table <- function(pi.hats, P.vintage, hzn, oos.start, oos.end) {
  
  # Inputs
  #   pi.hats - A list that contains the various model inflation forecasts
  #   P.vintage - The vintage of the price index (levels!) taken as "truth"
  #               Note P.vintage is a Tx1 matrix/vector
  #   hzn - Forecast horizon
  #   oos.start - Start of OOS period (i.e., first period forecasted)
  #   oos.end - End of OOS period (i.e., last period forecasted)
  
  all.obs <- seq(from=as.Date(as.yearqtr(oos.start)),
                 to=as.Date(as.yearqtr(oos.end)),
                 by='quarter')
  T <- length(all.obs)
  
  # First, define pi in the right way
  pi <- 400/hzn*(log(P.vintage) - stats::lag(log(P.vintage), hzn))[all.obs]
  
  # First, calculate RMSEs
  rmses <- sapply(pi.hats, rmse, y=pi)
  numer <- rmses['ar']
  rel.rmses <- rmses/numer
  
  tab <- data.frame(rmse=format(round(rmses, 2), nsmall=2), 
                    rel.rmse=format(round(rel.rmses, 2), nsmall=2), 
                    d.bar = NA,
                    t=NA, 
                    p=NA,
                    N=sapply(pi.hats, function(x) sum(!is.na(x))))
  
  hats.no.ar <- pi.hats[-which(names(pi.hats)=='ar')]
  DM.output <- t(sapply(hats.no.ar, dm.test, y=pi, ynull=pi.hats[['ar']]))
  
  tab[,'d.bar'] <- c(format(round(DM.output[,'d.bar'], 2)), '--')
  tab[,'t'] <- c(format(round(DM.output[,'DM'], 2)), '--')
  tab[,'p'] <- c(format(round(DM.output[,'p.value'], 2)), '--')
  
  return(tab)
}

gap.vints.models = gap.vints[c(-11,-14,-15)] #remove UC(CU), Tealbook, CBO
final.gap.table = final.gaps[c(-11,-14,-15)] #remove UC(CU), Tealbook, CBO

tab <- list()
pgdp.hat.2 <- lapply(gap.vints.models, rt.gap.forecast.inflation, P=P[['pgdp']], hzn=2, est.start='1960 Q2', oos.start='1970 Q1', oos.end='2022 Q1')
pgdp.hat.2[['ar']] <- rt.ar.forecast.inflation(P[['pgdp']], hzn=2, est.start='1960 Q2', oos.start='1970 Q1', oos.end='2022 Q1')
pgdp.hat.4 <- lapply(gap.vints.models, rt.gap.forecast.inflation, P=P[['pgdp']], hzn=4, est.start='1960 Q2', oos.start='1970 Q1', oos.end='2022 Q1')
pgdp.hat.4[['ar']] <- rt.ar.forecast.inflation(P[['pgdp']], hzn=4, est.start='1960 Q2', oos.start='1970 Q1', oos.end='2022 Q1')
pgdp.hat.8 <- lapply(gap.vints.models, rt.gap.forecast.inflation, P=P[['pgdp']], hzn=8, est.start='1960 Q2', oos.start='1970 Q1', oos.end='2022 Q1')
pgdp.hat.8[['ar']] <- rt.ar.forecast.inflation(P[['pgdp']], hzn=8, est.start='1960 Q2', oos.start='1970 Q1', oos.end='2022 Q1')

pgdp.hat.2.final <- lapply(final.gap.table, rt.gap.forecast.inflation, P=final.P[['pgdp']], hzn=2, est.start='1960 Q2', oos.start='1970 Q1', oos.end='2022 Q1')
pgdp.hat.2.final[['ar']] <- rt.ar.forecast.inflation(final.P[['pgdp']], hzn=2, est.start='1960 Q2', oos.start='1970 Q1', oos.end='2022 Q1')
pgdp.hat.4.final <- lapply(final.gap.table, rt.gap.forecast.inflation, P=final.P[['pgdp']], hzn=4, est.start='1960 Q2', oos.start='1970 Q1', oos.end='2022 Q1')
pgdp.hat.4.final[['ar']] <- rt.ar.forecast.inflation(final.P[['pgdp']], hzn=4, est.start='1960 Q2', oos.start='1970 Q1', oos.end='2022 Q1')
pgdp.hat.8.final <- lapply(final.gap.table, rt.gap.forecast.inflation, P=final.P[['pgdp']], hzn=8, est.start='1960 Q2', oos.start='1970 Q1', oos.end='2022 Q1')
pgdp.hat.8.final[['ar']] <- rt.ar.forecast.inflation(final.P[['pgdp']], hzn=8, est.start='1960 Q2', oos.start='1970 Q1', oos.end='2022 Q1')


tab$pgdp.2 <- rmse.table(pgdp.hat.2, P[['pgdp']][,'2022 Q2'], hzn=2, oos.start='1970 Q1', oos.end='2022 Q1') #the forecast is from vintage 1969Q3 to 2021Q3
tab$pgdp.4 <- rmse.table(pgdp.hat.4, P[['pgdp']][,'2022 Q2'], hzn=4, oos.start='1970 Q1', oos.end='2022 Q1') #the forecast is from vintage 1969Q1 to 2021Q1
tab$pgdp.8 <- rmse.table(pgdp.hat.8, P[['pgdp']][,'2022 Q2'], hzn=8, oos.start='1970 Q1', oos.end='2022 Q1') #the forecast is from vintage 1968Q3 to 2020Q3

stargazer(tab$pgdp.2, summary=FALSE)
stargazer(tab$pgdp.4, summary=FALSE)
stargazer(tab$pgdp.8, summary=FALSE)

tab5a = data.frame(tab$pgdp.2$rel.rmse,tab$pgdp.4$rel.rmse,tab$pgdp.8$rel.rmse)
rownames(tab5a) <- c('Linear', 'Quadratic', 'HP', 'GLMS', 'Hamilton', 'MW', 'KMW', 
                     'UC (GDP)', 'UC(u)', 'UC (GDP/$\\pi$)', 'UC (GDP/u.)', 'UC (GDP/u./$\\pi$)','AR')
stargazer(tab5a, summary=FALSE)


tab$pgdp.2.final <- rmse.table(pgdp.hat.2.final, P[['pgdp']][,'2022 Q2'], hzn=2, oos.start='1970 Q1', oos.end='2022 Q1') #the forecast is from vintage 1969Q3 to 2021Q3
tab$pgdp.4.final <- rmse.table(pgdp.hat.4.final, P[['pgdp']][,'2022 Q2'], hzn=4, oos.start='1970 Q1', oos.end='2022 Q1') #the forecast is from vintage 1969Q1 to 2021Q1
tab$pgdp.8.final <- rmse.table(pgdp.hat.8.final, P[['pgdp']][,'2022 Q2'], hzn=8, oos.start='1970 Q1', oos.end='2022 Q1') #the forecast is from vintage 1968Q3 to 2020Q3

stargazer(tab$pgdp.2.final, summary=FALSE)
stargazer(tab$pgdp.4.final, summary=FALSE)
stargazer(tab$pgdp.8.final, summary=FALSE)

tab5b = data.frame(tab$pgdp.2.final$rel.rmse,tab$pgdp.4.final$rel.rmse,tab$pgdp.8.final$rel.rmse)
rownames(tab5b) <- c('Linear', 'Quadratic', 'HP', 'GLMS', 'Hamilton', 'MW', 'KMW', 
                     'UC (GDP)', 'UC(u)', 'UC (GDP/$\\pi$)', 'UC (GDP/u.)', 'UC (GDP/u./$\\pi$)','AR')
stargazer(tab5b, summary=FALSE)




# TABLE 6 (GDP forecasting) -------------------------
rt.ar <- function(P, hzn, est.start, oos.start, oos.end) {
  
  # Here we produce AR(6) forecasts from an unrestricted AR(6) models
  
  first.vint <- as.yearqtr(oos.start) - (hzn/4)
  last.vint <- as.yearqtr(oos.end) - (hzn/4)
  all.vints <- paste0(as.yearqtr(seq(as.Date(first.vint), as.Date(last.vint), by='quarter')))
  
  # Storage for forecasts
  pi.hat <- xts(array(NA, length(all.vints)), 
                order.by=as.yearqtr(seq(from=as.Date(as.yearqtr(oos.start)), 
                                        to=as.Date(as.yearqtr(oos.end)), 
                                        by='quarter')))
  
  for (v in 1:length(all.vints)) {
    
    if (all.vints[v] == '1996 Q1') {
      pi.hat[v] <- NA # Damned shutdown
      
    } else {
      
      vintage <- paste0(all.vints[v])
      vm1 <- as.yearqtr(as.Date(as.yearqtr(vintage)) %m+% months(-1)) # End of information set
      
      P.v <- P[,vintage]
      
      y <- 400*(log(P.v) - stats::lag(log(P.v), 1)) # Backwards, quarterly
      
      # AR unrestricted model
      ar.mod <- ar(y[paste0(as.Date(as.yearqtr(est.start)), '/', as.Date(vm1))], aic=FALSE, order.max=6)
      y.hat <- predict(ar.mod, n.ahead=hzn+1) # Need "+1" because of data lags
      pi.hat[v] <- mean(y.hat[['pred']])
    }
  }
  
  return(pi.hat)
}
rt.gap.gdp <- function(gap, Y, hzn, est.start, oos.start, oos.end) {
  
  # This function does fully real-time forecasts of GDP,
  #   looping over vintages of Y (GDP level) and the output gap
  
  # The auxilliary model: y^h_{t-1} = alpha + beta*gap_{t-1} + e_t
  
  # Gap - name of model to be used; in matrix of dimension TxV
  # P - Name of matrix of vintage *price levels*
  # hzn - Forecast horizon
  # est.start - Start of estimation period
  # oos.start - First quarter to be forecasted
  # oos.end - Final quarter to be forecasted
  
  all.obs <- paste0(as.Date(as.yearqtr(est.start)), '/', as.Date(as.yearqtr(oos.end)))
  first.vint <- as.yearqtr(oos.start) - (hzn/4)
  last.vint <- as.yearqtr(oos.end) - (hzn/4)
  all.vints <- paste0(as.yearqtr(seq(as.Date(first.vint), as.Date(last.vint), by='quarter')))
  
  # Storage for forecasts
  y.hat <- xts(array(NA, length(all.vints)), 
               order.by=as.yearqtr(seq(from=as.Date(as.yearqtr(oos.start)), 
                                       to=as.Date(as.yearqtr(oos.end)), 
                                       by='quarter')))
  
  for (v in 1:length(all.vints)) {
    
    vintage <- paste0(all.vints[v])
    
    Y.v <- Y[,vintage]
    gap.v <- gap[,vintage]
    
    y <- 400/hzn*(stats::lag(log(Y.v), -hzn) - log(Y.v)) # Note this leads!!
    
    data <- cbind(y, stats::lag(gap.v,1))
    colnames(data)[1] <- 'y'
    colnames(data)[2] <- 'l.gap'
    
    # Given the NAs in the data this isn't needed but it's more clear
    last.qtr <- as.Date(as.yearqtr(as.Date(as.yearqtr(vintage)) %m+% months(-(hzn*3+1))))
    
    if (sum(complete.cases(data[paste0(as.Date(as.yearqtr(est.start)), '/', last.qtr)]))>0) {
      reg <- lm(y~., data=data[paste0(as.Date(as.yearqtr(est.start)), '/', last.qtr)]) 
      y.hat[v] <- as.numeric(predict(reg, newdata=data[as.yearqtr(vintage), -1]))
    } else {
      y.hat[v] <- NA
    }
  }
  return(y.hat)
}

Y <- xts(data.frame(readxl::read_excel(paste0(data.loc, 'Vintages-RGDP.xlsx'))[,-1]),
         seq(from=as.Date('1947-01-01'), to=as.Date('2022-01-01'), by='quarter'))
colnames(Y) <- paste0(as.yearqtr(seq(from=as.Date('1965-10-01'), to=as.Date('2022-04-01'), by='quarter')))

final.Y <- xts(array(NA, dim(Y)), 
               order.by=seq(from=as.Date('1947-01-01'), to=as.Date('2022-01-01'), by='quarter'))
colnames(final.Y) <- colnames(Y)
V <- dim(Y)[2]

for (v in 1:V) {
  last.q <- as.Date(as.yearqtr(colnames(final.Y)[v])) %m+% months(-3)
  final.Y[paste0('/', last.q),v] <- Y[paste0('/', last.q), '2022 Q2']
}


gdp.hat <- list()

gdp.hat[['h4']] <- lapply(gap.vints.models, rt.gap.gdp, Y=Y, hzn=4, est.start='1960 Q2', oos.start='1970 Q1', oos.end='2022 Q1')
gdp.hat[['final.h4']] <- lapply(final.gap.table, rt.gap.gdp, Y=final.Y, hzn=4, est.start='1960 Q2', oos.start='1970 Q1', oos.end='2022 Q1')
gdp.hat[['h4']][['ar']] <- rt.ar(Y, hzn=4, est.start='1960 Q2', oos.start='1970 Q1', oos.end='2022 Q1')
gdp.hat[['final.h4']][['ar']] <- rt.ar(final.Y, hzn=4, est.start='1960 Q2', oos.start='1970 Q1', oos.end='2022 Q1')

gdp.hat[['h8']] <- lapply(gap.vints.models, rt.gap.gdp, Y=Y, hzn=8, est.start='1960 Q2', oos.start='1970 Q1', oos.end='2022 Q1')
gdp.hat[['final.h8']] <- lapply(final.gap.table, rt.gap.gdp, Y=final.Y, hzn=8, est.start='1960 Q2', oos.start='1970 Q1', oos.end='2022 Q1')
gdp.hat[['h8']][['ar']] <- rt.ar(Y, hzn=8, est.start='1960 Q2', oos.start='1970 Q1', oos.end='2022 Q1')
gdp.hat[['final.h8']][['ar']] <- rt.ar(final.Y, hzn=8, est.start='1960 Q2', oos.start='1970 Q1', oos.end='2022 Q1')

gdp.hat[['h12']] <- lapply(gap.vints.models, rt.gap.gdp, Y=Y, hzn=12, est.start='1960 Q2', oos.start='1970 Q1', oos.end='2022 Q1')
gdp.hat[['final.h12']] <- lapply(final.gap.table, rt.gap.gdp, Y=final.Y, hzn=12, est.start='1960 Q2', oos.start='1970 Q1', oos.end='2022 Q1')
gdp.hat[['h12']][['ar']] <- rt.ar(Y, hzn=12, est.start='1960 Q2', oos.start='1970 Q1', oos.end='2022 Q1')
gdp.hat[['final.h12']][['ar']] <- rt.ar(final.Y, hzn=12, est.start='1960 Q2', oos.start='1970 Q1', oos.end='2022 Q1')

gdp.hat[['h16']] <- lapply(gap.vints.models, rt.gap.gdp, Y=Y, hzn=16, est.start='1960 Q2', oos.start='1970 Q1', oos.end='2022 Q1')
gdp.hat[['final.h16']] <- lapply(final.gap.table, rt.gap.gdp, Y=final.Y, hzn=16, est.start='1960 Q2', oos.start='1970 Q1', oos.end='2022 Q1')
gdp.hat[['h16']][['ar']] <- rt.ar(Y, hzn=16, est.start='1960 Q2', oos.start='1970 Q1', oos.end='2022 Q1')
gdp.hat[['final.h16']][['ar']] <- rt.ar(final.Y, hzn=16, est.start='1960 Q2', oos.start='1970 Q1', oos.end='2022 Q1')

tab.6 <- list()
tab.6$gdp.4 <- rmse.table(gdp.hat[['h4']], Y[,'2022 Q2'], hzn=4, oos.start='1970 Q1', oos.end='2022 Q1') 
tab.6$gdp.8 <- rmse.table(gdp.hat[['h8']], Y[,'2022 Q2'], hzn=8, oos.start='1970 Q1', oos.end='2022 Q1') 
tab.6$gdp.12 <- rmse.table(gdp.hat[['h12']], Y[,'2022 Q2'], hzn=12, oos.start='1970 Q1', oos.end='2022 Q1') 
tab.6$gdp.16 <- rmse.table(gdp.hat[['h16']], Y[,'2022 Q2'], hzn=16, oos.start='1970 Q1', oos.end='2022 Q1') 

tab.6$final.gdp.4 <- rmse.table(gdp.hat[['final.h4']], Y[,'2022 Q2'], hzn=4, oos.start='1970 Q1', oos.end='2022 Q1') 
tab.6$final.gdp.8 <- rmse.table(gdp.hat[['final.h8']], Y[,'2022 Q2'], hzn=8, oos.start='1970 Q1', oos.end='2022 Q1') 
tab.6$final.gdp.12 <- rmse.table(gdp.hat[['final.h12']], Y[,'2022 Q2'], hzn=12, oos.start='1970 Q1', oos.end='2022 Q1') 
tab.6$final.gdp.16 <- rmse.table(gdp.hat[['final.h16']], Y[,'2022 Q2'], hzn=16, oos.start='1970 Q1', oos.end='2022 Q1') 

tab.6
sapply(tab.6, FUN=function(x) x['ar', 'rmse'])
tab.6.tex <- data.frame(tab.6$gdp.4$rel.rmse, tab.6$gdp.8$rel.rmse, tab.6$gdp.12$rel.rmse, tab.6$gdp.16$rel.rmse, 
                        tab.6$final.gdp.4$rel.rmse, tab.6$final.gdp.8$rel.rmse, tab.6$final.gdp.12$rel.rmse, tab.6$final.gdp.16$rel.rmse)
rownames(tab.6.tex) <- names(gdp.hat[['h4']])
stargazer(tab.6.tex, summary=FALSE)


# TABLE 7 (u forecasting) -------------------------
rt.ar <- function(P, hzn, est.start, oos.start, oos.end) {
  
  # Here we produce AR(6) forecasts from an unrestricted AR(6) models
  
  first.vint <- as.yearqtr(oos.start) - (hzn/4)
  last.vint <- as.yearqtr(oos.end) - (hzn/4)
  all.vints <- paste0(as.yearqtr(seq(as.Date(first.vint), as.Date(last.vint), by='quarter')))
  
  # Storage for forecasts
  pi.hat <- xts(array(NA, length(all.vints)), 
                order.by=as.yearqtr(seq(from=as.Date(as.yearqtr(oos.start)), 
                                        to=as.Date(as.yearqtr(oos.end)), 
                                        by='quarter')))
  
  for (v in 1:length(all.vints)) {
    
    if (all.vints[v] == '1996 Q1') {
      pi.hat[v] <- NA # Damned shutdown
      
    } else {
      
      vintage <- paste0(all.vints[v])
      vm1 <- as.yearqtr(as.Date(as.yearqtr(vintage)) %m+% months(-1)) # End of information set
      
      P.v <- P[,vintage]
      
      y <- 1*(P.v - stats::lag(P.v, 1)) # Backwards, quarterly
      
      # AR unrestricted model
      ar.mod <- ar(y[paste0(as.Date(as.yearqtr(est.start)), '/', as.Date(vm1))], aic=FALSE, order.max=6)
      y.hat <- predict(ar.mod, n.ahead=hzn+1) # Need "+1" because of data lags
      pi.hat[v] <- mean(y.hat[['pred']])
    }
  }
  
  return(pi.hat)
}
rt.gap.u <- function(gap, Y, hzn, est.start, oos.start, oos.end) {
  
  # This function does fully real-time forecasts of GDP,
  #   looping over vintages of Y (GDP level) and the output gap
  
  # The auxilliary model: y^h_{t-1} = alpha + beta*gap_{t-1} + e_t
  
  # Gap - name of model to be used; in matrix of dimension TxV
  # P - Name of matrix of vintage *price levels*
  # hzn - Forecast horizon
  # est.start - Start of estimation period
  # oos.start - First quarter to be forecasted
  # oos.end - Final quarter to be forecasted
  
  all.obs <- paste0(as.Date(as.yearqtr(est.start)), '/', as.Date(as.yearqtr(oos.end)))
  first.vint <- as.yearqtr(oos.start) - (hzn/4)
  last.vint <- as.yearqtr(oos.end) - (hzn/4)
  all.vints <- paste0(as.yearqtr(seq(as.Date(first.vint), as.Date(last.vint), by='quarter')))
  
  # Storage for forecasts
  y.hat <- xts(array(NA, length(all.vints)), 
               order.by=as.yearqtr(seq(from=as.Date(as.yearqtr(oos.start)), 
                                       to=as.Date(as.yearqtr(oos.end)), 
                                       by='quarter')))
  
  for (v in 1:length(all.vints)) {
    
    vintage <- paste0(all.vints[v])
    
    Y.v <- Y[,vintage]
    gap.v <- gap[,vintage]
    
    y <- 1/hzn*(stats::lag(Y.v, -hzn) - Y.v) # Note this leads!!
    
    data <- cbind(y, stats::lag(gap.v,1))
    colnames(data)[1] <- 'y'
    colnames(data)[2] <- 'l.gap'
    
    # Given the NAs in the data this isn't needed but it's more clear
    last.qtr <- as.Date(as.yearqtr(as.Date(as.yearqtr(vintage)) %m+% months(-(hzn*3+1))))
    
    if (sum(complete.cases(data[paste0(as.Date(as.yearqtr(est.start)), '/', last.qtr)]))>0) {
      reg <- lm(y~., data=data[paste0(as.Date(as.yearqtr(est.start)), '/', last.qtr)]) 
      y.hat[v] <- as.numeric(predict(reg, newdata=data[as.yearqtr(vintage), -1]))
    } else {
      y.hat[v] <- NA
    }
  }
  return(y.hat)
}
rmse.table <- function(pi.hats, P.vintage, hzn, oos.start, oos.end) {
  
  # Inputs
  #   pi.hats - A list that contains the various model inflation forecasts
  #   P.vintage - The vintage of the unemployment level taken as "truth"
  #               Note P.vintage is a Tx1 matrix/vector
  #   hzn - Forecast horizon
  #   oos.start - Start of OOS period (i.e., first period forecasted)
  #   oos.end - End of OOS period (i.e., last period forecasted)
  
  all.obs <- seq(from=as.Date(as.yearqtr(oos.start)),
                 to=as.Date(as.yearqtr(oos.end)),
                 by='quarter')
  T <- length(all.obs)
  
  # First, define pi in the right way
  pi <- 1/hzn*(P.vintage - stats::lag(P.vintage, hzn))[all.obs]
  
  # First, calculate RMSEs
  rmses <- sapply(pi.hats, rmse, y=pi)
  numer <- rmses['ar']
  rel.rmses <- rmses/numer
  
  tab <- data.frame(rmse=format(round(rmses, 2), nsmall=2), 
                    rel.rmse=format(round(rel.rmses, 2), nsmall=2), 
                    d.bar = NA,
                    t=NA, 
                    p=NA,
                    N=sapply(pi.hats, function(x) sum(!is.na(x))))
  
  hats.no.ar <- pi.hats[-which(names(pi.hats)=='ar')]
  DM.output <- t(sapply(hats.no.ar, dm.test, y=pi, ynull=pi.hats[['ar']]))
  
  tab[,'d.bar'] <- c(format(round(DM.output[,'d.bar'], 2)), '--')
  tab[,'t'] <- c(format(round(DM.output[,'DM'], 2)), '--')
  tab[,'p'] <- c(format(round(DM.output[,'p.value'], 2)), '--')
  
  return(tab)
}
vintages <- seq(from=as.Date('1965-11-15'), by='quarter', length.out=V)
dates.all <- paste0('/', vintages[v] %m+% -months(3))
udata.m <- xts(data.frame(readxl::read_excel(paste0(data.loc, 'Vintages-unemployment-rate.xlsx'))[,-1]),
         seq(from=as.Date('1947-01-01'), to=as.Date('2022-04-01'), by='month'))
udata = xts(aggregate(udata.m, by=as.yearqtr(index(udata.m)), mean))[dates.all]
colnames(udata) <- paste0(as.yearqtr(seq(from=as.Date('1965-10-01'), to=as.Date('2022-04-01'), by='quarter')))

final.udata <- xts(array(NA, dim(udata)), 
               order.by=seq(from=as.Date('1947-01-01'), to=as.Date('2022-01-01'), by='quarter'))
colnames(final.udata) <- colnames(udata)
V <- dim(udata)[2]

for (v in 1:V) {
  last.q <- as.Date(as.yearqtr(colnames(final.udata)[v])) %m+% months(-3)
  final.udata[paste0('/', last.q),v] <- udata[paste0('/', last.q), '2022 Q2']
}


u.hat <- list()

u.hat[['h4']] <- lapply(gap.vints.models, rt.gap.u, Y=udata, hzn=4, est.start='1960 Q2', oos.start='1970 Q1', oos.end='2022 Q1')
u.hat[['final.h4']] <- lapply(final.gap.table, rt.gap.u, Y=final.udata, hzn=4, est.start='1960 Q2', oos.start='1970 Q1', oos.end='2022 Q1')
u.hat[['h4']][['ar']] <- rt.ar(udata, hzn=4, est.start='1960 Q2', oos.start='1970 Q1', oos.end='2022 Q1')
u.hat[['final.h4']][['ar']] <- rt.ar(final.udata, hzn=4, est.start='1960 Q2', oos.start='1970 Q1', oos.end='2022 Q1')

u.hat[['h8']] <- lapply(gap.vints.models, rt.gap.u, Y=udata, hzn=8, est.start='1960 Q2', oos.start='1970 Q1', oos.end='2022 Q1')
u.hat[['final.h8']] <- lapply(final.gap.table, rt.gap.u, Y=final.udata, hzn=8, est.start='1960 Q2', oos.start='1970 Q1', oos.end='2022 Q1')
u.hat[['h8']][['ar']] <- rt.ar(udata, hzn=8, est.start='1960 Q2', oos.start='1970 Q1', oos.end='2022 Q1')
u.hat[['final.h8']][['ar']] <- rt.ar(final.udata, hzn=8, est.start='1960 Q2', oos.start='1970 Q1', oos.end='2022 Q1')

u.hat[['h12']] <- lapply(gap.vints.models, rt.gap.u, Y=udata, hzn=12, est.start='1960 Q2', oos.start='1970 Q1', oos.end='2022 Q1')
u.hat[['final.h12']] <- lapply(final.gap.table, rt.gap.u, Y=final.udata, hzn=12, est.start='1960 Q2', oos.start='1970 Q1', oos.end='2022 Q1')
u.hat[['h12']][['ar']] <- rt.ar(udata, hzn=12, est.start='1960 Q2', oos.start='1970 Q1', oos.end='2022 Q1')
u.hat[['final.h12']][['ar']] <- rt.ar(final.udata, hzn=12, est.start='1960 Q2', oos.start='1970 Q1', oos.end='2022 Q1')

u.hat[['h16']] <- lapply(gap.vints.models, rt.gap.u, Y=udata, hzn=16, est.start='1960 Q2', oos.start='1970 Q1', oos.end='2022 Q1')
u.hat[['final.h16']] <- lapply(final.gap.table, rt.gap.u, Y=final.udata, hzn=16, est.start='1960 Q2', oos.start='1970 Q1', oos.end='2022 Q1')
u.hat[['h16']][['ar']] <- rt.ar(udata, hzn=16, est.start='1960 Q2', oos.start='1970 Q1', oos.end='2022 Q1')
u.hat[['final.h16']][['ar']] <- rt.ar(final.udata, hzn=16, est.start='1960 Q2', oos.start='1970 Q1', oos.end='2022 Q1')

tab.7 <- list()
tab.7$gdp.4 <- rmse.table(u.hat[['h4']], udata[,'2022 Q2'], hzn=4, oos.start='1970 Q1', oos.end='2022 Q1') 
tab.7$gdp.8 <- rmse.table(u.hat[['h8']], udata[,'2022 Q2'], hzn=8, oos.start='1970 Q1', oos.end='2022 Q1') 
tab.7$gdp.12 <- rmse.table(u.hat[['h12']], udata[,'2022 Q2'], hzn=12, oos.start='1970 Q1', oos.end='2022 Q1') 
tab.7$gdp.16 <- rmse.table(u.hat[['h16']], udata[,'2022 Q2'], hzn=16, oos.start='1970 Q1', oos.end='2022 Q1') 

tab.7$final.gdp.4 <- rmse.table(u.hat[['final.h4']], udata[,'2022 Q2'], hzn=4, oos.start='1970 Q1', oos.end='2022 Q1') 
tab.7$final.gdp.8 <- rmse.table(u.hat[['final.h8']], udata[,'2022 Q2'], hzn=8, oos.start='1970 Q1', oos.end='2022 Q1') 
tab.7$final.gdp.12 <- rmse.table(u.hat[['final.h12']], udata[,'2022 Q2'], hzn=12, oos.start='1970 Q1', oos.end='2022 Q1') 
tab.7$final.gdp.16 <- rmse.table(u.hat[['final.h16']], udata[,'2022 Q2'], hzn=16, oos.start='1970 Q1', oos.end='2022 Q1') 

tab.7
sapply(tab.7, FUN=function(x) x['ar', 'rmse'])
tab.7.tex <- data.frame(tab.7$gdp.4$rel.rmse, tab.7$gdp.8$rel.rmse, tab.7$gdp.12$rel.rmse, tab.7$gdp.16$rel.rmse, 
                        tab.7$final.gdp.4$rel.rmse, tab.7$final.gdp.8$rel.rmse, tab.7$final.gdp.12$rel.rmse, tab.7$final.gdp.16$rel.rmse)
rownames(tab.7.tex) <- names(gdp.hat[['h4']])
stargazer(tab.7.tex, summary=FALSE)




# FIGURE A1 ----
plot.all.rt <- function() {
  
  t <- as.numeric(index(gaps_main_results[[1]]))
  
  plot(t,gaps_main_results[[1]][,'rt'], type='n', axes=FALSE, xlab='', ylab='',ylim=c(-16,8))
  nberShade()
  abline(h=0)
  for (i in 1:length(gaps_main_results)) {
    lines(t, gaps_main_results[[i]][,'rt'], col=palette2[i])
  }
  
  axis(1, at=seq(1970,2022,10), labels=seq(1970,2022,10), cex.axis=.8)
  axis(1, at=seq(1965,2022,5), labels=FALSE, tck=-0.02)
  axis(2, cex.axis=.8)
  
  legend(x=1983, y=-9, bty='n', 
         legend=c('Line', 'Quad', 'HP', 'GLMS', 'Hamilton', 'MW', 
                  'KMW', 
                  expression(paste('UC (GDP)')),
                  expression(paste('UC (GDP/', pi, ')')),
                  expression(paste('UC (GDP/CU)')),
                  expression(paste('UC (GDP/u)')),
                  expression(paste('UC (GDP/', pi, '/', 'u.)')),
                  'Tealbook', 
                  'CBO'),
         col=palette2, pch=c(16), pt.cex=1.2,
         cex=.75, ncol=3, text.width=8)
  box()
  title('%', adj=1, font.main=1, cex.main=.8, line=.2)
}
to.pdf(plot.all.rt(), filename=paste0(figures.loc, 'Fig-A1.pdf'), height=10)
trim.pdf(filename='../Figures/Fig-A1.pdf')
# TABLE A2 (Summary stats split sample) -------------------------------
tab.a2a <- t(sapply(gaps_main_results, function(x) gap.stats(x$rt, '1975 Q1', '1997 Q4')))
tab.a2b <- t(sapply(gaps_main_results, function(x) gap.stats(x$rt, '1998 Q1', '2016 Q3')))
stargazer(cbind(tab.a2a, tab.a2b))

# Table A3 (Summary statistics for revisions, split sample) -----------
tab.a3a <- t(sapply(gaps_main_results, function(x) rev.stats(x$rt, x$final.2016, '1975 Q1', '1997 Q4')))
tab.a3b <- t(sapply(gaps_main_results, function(x) rev.stats(x$rt, x$final.2016, '1998 Q1', '2016 Q3')))

rownames(tab.a3a) <- rownames(tab.a3b) <- c('Linear', 'Quadratic', 'HP', 'GLMS', 'Hamilton', 'MW', 'KMW',
                               'UC (GDP)', 'UC (GDP/$\\pi$)', 'UC (GDP/CU)', 'UC (GDP/u.)', 'UC (GDP/u./$\\pi$)',
                               'Tealbook', 'CBO')
tab.a3a_display = tab.a3a[,c('N', 'ns.sd','ns.rmse','sign')]
tab.a3b_display = tab.a3b[,c('N', 'ns.sd','ns.rmse','sign')]
stargazer(cbind(tab.a3a_display, 
                tab.a3b_display))

# Table A4 (including MW(2020) bivariate and trivariate) -------------------------
gaps_no_ucu = gaps[-c(11)] #remove UC(u)
t(sapply(gaps_no_ucu, function(x) rev.stats(x$rt, x$final, '1965 Q3', '2022 Q1')))
tab.a4 <- t(sapply(gaps_no_ucu, function(x) rev.stats(x$rt, x$final, '1965 Q3', '2022 Q1')))
rownames(tab.a4) <- c('Linear', 'Quadratic', 'HP', 'GLMS', 'Hamilton', 'MW', 'KMW', 'BN (GDP/u.)', 'BN (GDP/u./$\\pi$)',
                     'GDP', 'GDP/$\\pi$', 'GDP/CU', 'GDP/u.', 'GDP/u./$\\pi$',
                     'Tealbook', 'CBO')
tab.a4.display = tab.a4[, c('N', 'mean', 'sd', 'rmse', 'ns.sd', 'ns.rmse', 'sign')]
stargazer(tab.a4.display)

gaps_onlyTB = gaps[c(16)] #only TB
tab.a4.tb.short1 <- t(sapply(gaps_onlyTB, function(x) rev.stats(x$rt, x$final.2016, '1965 Q3', '2016 Q4')))
rownames(tab.a4.tb.short1) <- c('Tealbook')
tab.a4.tb.short1

gaps_onlyCBO = gaps[c(17)] #only CBO
tab.a4.cbo.short1 <- t(sapply(gaps_onlyCBO, function(x) rev.stats(x$rt, x$final, '1990 Q4', '2022 Q1')))
rownames(tab.a4.cbo.short1) <- c('CBO')
tab.a4.cbo.short1

gaps_TB_CBO = gaps[c(16,17)] #only TB
tab.a4.tb_cbo.short1 <- t(sapply(gaps_TB_CBO, function(x) rev.stats(x$rt, x$final.2016, '1998 Q1', '2016 Q3')))
rownames(tab.a4.tb_cbo.short1) <- c('Tealbook','CBO')
tab.a4.tb_cbo.short1



# TABLE A5 (inflation) -------------------------
tab.a5 <- list()
tab.a5$pgdp.2 <- rmse.table(pgdp.hat.2, P[['pgdp']][,'2022 Q2'], hzn=2, oos.start='1975 Q1', oos.end='1997 Q4') 
tab.a5$pgdp.4 <- rmse.table(pgdp.hat.4, P[['pgdp']][,'2022 Q2'], hzn=4, oos.start='1975 Q1', oos.end='1997 Q4') 
tab.a5$pgdp.8 <- rmse.table(pgdp.hat.8, P[['pgdp']][,'2022 Q2'], hzn=8, oos.start='1975 Q1', oos.end='1997 Q4') 
tab.a5$final.pgdp.2 <- rmse.table(pgdp.hat.2.final, P[['pgdp']][,'2022 Q2'], hzn=2, oos.start='1975 Q1', oos.end='1997 Q4') 
tab.a5$final.pgdp.4 <- rmse.table(pgdp.hat.4.final, P[['pgdp']][,'2022 Q2'], hzn=4, oos.start='1975 Q1', oos.end='1997 Q4') 
tab.a5$final.pgdp.8 <- rmse.table(pgdp.hat.8.final, P[['pgdp']][,'2022 Q2'], hzn=8, oos.start='1975 Q1', oos.end='1997 Q4') 

tab.a5
tab.a5.tex <- data.frame(tab.a5$pgdp.2$rel.rmse, tab.a5$pgdp.4$rel.rmse, tab.a5$pgdp.8$rel.rmse,
                         tab.a5$final.pgdp.2$rel.rmse, tab.a5$final.pgdp.4$rel.rmse, tab.a5$final.pgdp.8$rel.rmse)
stargazer(tab.a5.tex, summary=FALSE)

# TABLE A6 (inflation) -------------------------
pgdp.hat <- list()
pgdp.hat[['h2.2016']] <- lapply(gap.vints, rt.gap.forecast.inflation, P=P[['pgdp']], hzn=2, est.start='1975 Q1', oos.start='1998 Q1', oos.end='2016 Q3')
pgdp.hat[['h2.2016']][['ar']] <- rt.ar.forecast.inflation(P[['pgdp']], hzn=2, est.start='1975 Q1', oos.start='1998 Q1', oos.end='2016 Q3')
pgdp.hat[['h4.2016']] <- lapply(gap.vints, rt.gap.forecast.inflation, P=P[['pgdp']], hzn=4, est.start='1975 Q1', oos.start='1998 Q1', oos.end='2016 Q3')
pgdp.hat[['h4.2016']][['ar']] <- rt.ar.forecast.inflation(P[['pgdp']], hzn=4, est.start='1975 Q1', oos.start='1998 Q1', oos.end='2016 Q3')
pgdp.hat[['h8.2016']] <- lapply(gap.vints, rt.gap.forecast.inflation, P=P[['pgdp']], hzn=8, est.start='1975 Q1', oos.start='1998 Q1', oos.end='2016 Q3')
pgdp.hat[['h8.2016']][['ar']] <- rt.ar.forecast.inflation(P[['pgdp']], hzn=8, est.start='1975 Q1', oos.start='1998 Q1', oos.end='2016 Q3')

pgdp.hat[['final.h2.2016']] <- lapply(final.gaps.2016, rt.gap.forecast.inflation, P=final.P[['pgdp']], hzn=2, est.start='1975 Q1', oos.start='1998 Q1', oos.end='2016 Q3')
pgdp.hat[['final.h2.2016']][['ar']] <- rt.ar.forecast.inflation(final.P[['pgdp']], hzn=2, est.start='1975 Q1', oos.start='1998 Q1', oos.end='2016 Q3')
pgdp.hat[['final.h4.2016']] <- lapply(final.gaps.2016, rt.gap.forecast.inflation, P=final.P[['pgdp']], hzn=4, est.start='1975 Q1', oos.start='1998 Q1', oos.end='2016 Q3')
pgdp.hat[['final.h4.2016']][['ar']] <- rt.ar.forecast.inflation(final.P[['pgdp']], hzn=4, est.start='1975 Q1', oos.start='1998 Q1', oos.end='2016 Q3')
pgdp.hat[['final.h8.2016']] <- lapply(final.gaps.2016, rt.gap.forecast.inflation, P=final.P[['pgdp']], hzn=8, est.start='1975 Q1', oos.start='1998 Q1', oos.end='2016 Q3')
pgdp.hat[['final.h8.2016']][['ar']] <- rt.ar.forecast.inflation(final.P[['pgdp']], hzn=8, est.start='1975 Q1', oos.start='1998 Q1', oos.end='2016 Q3')


tab.a6 <- list()
tab.a6$pgdp.2 <- rmse.table(pgdp.hat[['h2.2016']], P[['pgdp']][,'2016 Q4'], hzn=2, oos.start='1998 Q1', oos.end='2016 Q3') 
tab.a6$pgdp.4 <- rmse.table(pgdp.hat[['h4.2016']], P[['pgdp']][,'2016 Q4'], hzn=4, oos.start='1998 Q1', oos.end='2016 Q3') 
tab.a6$pgdp.8 <- rmse.table(pgdp.hat[['h8.2016']], P[['pgdp']][,'2016 Q4'], hzn=8, oos.start='1998 Q1', oos.end='2016 Q3') 
tab.a6$final.pgdp.2 <- rmse.table(pgdp.hat[['final.h2.2016']], P[['pgdp']][,'2016 Q4'], hzn=2, oos.start='1998 Q1', oos.end='2016 Q3') 
tab.a6$final.pgdp.4 <- rmse.table(pgdp.hat[['final.h4.2016']], P[['pgdp']][,'2016 Q4'], hzn=4, oos.start='1998 Q1', oos.end='2016 Q3') 
tab.a6$final.pgdp.8 <- rmse.table(pgdp.hat[['final.h8.2016']], P[['pgdp']][,'2016 Q4'], hzn=8, oos.start='1998 Q1', oos.end='2016 Q3') 

tab.a6
tab.a6.tex <- data.frame(tab.a6$pgdp.2$rel.rmse, tab.a6$pgdp.4$rel.rmse, tab.a6$pgdp.8$rel.rmse,
                         tab.a6$final.pgdp.2$rel.rmse, tab.a6$final.pgdp.4$rel.rmse, tab.a6$final.pgdp.8$rel.rmse)
rownames(tab.a6.tex) <- names(pgdp.hat[['h2.2016']])
stargazer(tab.a6.tex, summary=FALSE)

# TABLE A7 (inflation) -----

rt.uc3 <- readRDS('../Raw model output/rt-uc3.Rds')
rt.uc2pi <- readRDS('../Raw model output/rt-uc2.pi.Rds')

forecast_inflation_outer_uc3 <- function(vint){
  forecast_inflation_inner <- function(draw_num){
    
    draw = list()
    draw$beta =  rt.uc3[[vint]]$res$beta[,draw_num]
    draw$phi =  rt.uc3[[vint]]$res$phi[,draw_num]
    draw$S =  rt.uc3[[vint]]$res$S[,,draw_num]
    
    name_columns = c('outputgap')
    output_gap_forecast = matrix(nrow=8,ncol=1, dimnames = list(1:8,name_columns) )
    output_gap_forecast[1] =  draw$phi[1]*draw$S[dim(draw$S)[1],5]+draw$phi[2]*draw$S[dim(draw$S)[1]-1,5]
    output_gap_forecast[2] =  draw$phi[1]*output_gap_forecast[1]+draw$phi[2]*draw$S[dim(draw$S)[1],5]
    for (i in 3:10){
      output_gap_forecast[i] =  draw$phi[1]*output_gap_forecast[i-1]+draw$phi[2]*output_gap_forecast[i-2]
    }
    
    name_columns = c('pi')
    pi_forecast = matrix(nrow=8,ncol=1, dimnames = list(1:8,name_columns) )
    pi_forecast[1] = draw$S[dim(draw$S)[1],4] + draw$beta[1]*output_gap_forecast[1]+draw$beta[2]*draw$S[dim(draw$S)[1],5]
    for (i in 2:10){
      pi_forecast[i] = draw$S[dim(draw$S)[1],4] + draw$beta[1]*output_gap_forecast[i]+draw$beta[2]*output_gap_forecast[i-1]
    }
    return(pi_forecast)
  }
  if (vint==122){
    # Drop 1996Q1 vintage because of the shutdown
    pi_forecast_mean= c(1,NA,NA,NA,NA,NA,NA,NA,NA,NA)
    pi_forecast_mean[1] = NA
  }else{
    pi_forecast_list <- lapply(1:1000, forecast_inflation_inner)
    pi_forecast_df = do.call(cbind.data.frame, pi_forecast_list)
    pi_forecast_mean = rowMeans(pi_forecast_df)
  }
  return(pi_forecast_mean)
}
forecast_inflation_outer_uc2pi <- function(vint){
  forecast_inflation_inner <- function(draw_num){
    
    draw = list()
    draw$beta =  rt.uc2pi[[vint]]$res$beta[,draw_num]
    draw$phi =  rt.uc2pi[[vint]]$res$phi[,draw_num]
    draw$S =  rt.uc2pi[[vint]]$res$S[,,draw_num]
    
    name_columns = c('outputgap')
    output_gap_forecast = matrix(nrow=8,ncol=1, dimnames = list(1:8,name_columns) )
    output_gap_forecast[1] =  draw$phi[1]*draw$S[dim(draw$S)[1],4]+draw$phi[2]*draw$S[dim(draw$S)[1]-1,4]
    output_gap_forecast[2] =  draw$phi[1]*output_gap_forecast[1]+draw$phi[2]*draw$S[dim(draw$S)[1],4]
    for (i in 3:10){
      output_gap_forecast[i] =  draw$phi[1]*output_gap_forecast[i-1]+draw$phi[2]*output_gap_forecast[i-2]
    }
    
    name_columns = c('pi')
    pi_forecast = matrix(nrow=8,ncol=1, dimnames = list(1:8,name_columns) )
    pi_forecast[1] = draw$S[dim(draw$S)[1],3] + draw$beta[1]*output_gap_forecast[1]+draw$beta[2]*draw$S[dim(draw$S)[1],4]
    for (i in 2:10){
      pi_forecast[i] = draw$S[dim(draw$S)[1],3] + draw$beta[1]*output_gap_forecast[i]+draw$beta[2]*output_gap_forecast[i-1]
    }
    return(pi_forecast)
  }
  if (vint==122){
    # Drop 1996Q1 vintage because of the shutdown
    pi_forecast_mean= c(1,NA,NA,NA,NA,NA,NA,NA,NA,NA)
    pi_forecast_mean[1] = NA
  }else{
    pi_forecast_list <- lapply(1:1000, forecast_inflation_inner)
    pi_forecast_df = do.call(cbind.data.frame, pi_forecast_list)
    pi_forecast_mean = rowMeans(pi_forecast_df)
  }
  return(pi_forecast_mean)
}

pi_forecasts_list <- lapply(1:226, forecast_inflation_outer_uc3)
pi_forecasts_list_uc2pi <- lapply(1:226, forecast_inflation_outer_uc2pi)

pi_forecasts_df = do.call(cbind.data.frame, pi_forecasts_list)
names(pi_forecasts_df) = NULL
pi_forecasts_df_uc2pi = do.call(cbind.data.frame, pi_forecasts_list_uc2pi)
names(pi_forecasts_df_uc2pi) = NULL

pi.hat.2.rt = list()
name_columns = c('rt')
pi.hat.2.rt$uc2pi = matrix(nrow=208,ncol=1, dimnames = list(1:208,name_columns) )
pi.hat.2.rt$uc2pi = (t(pi_forecasts_df_uc2pi[1,16:224])+t(pi_forecasts_df_uc2pi[2,16:224])+t(pi_forecasts_df_uc2pi[2,16:224]))/3
pi.hat.2.rt$uc3 = matrix(nrow=208,ncol=1, dimnames = list(1:208,name_columns) )
pi.hat.2.rt$uc3 = (t(pi_forecasts_df[1,16:224])+t(pi_forecasts_df[2,16:224])+t(pi_forecasts_df[2,16:224]))/3
pi.hat.2.rt$ar <- rt.ar.forecast.inflation(P[['pgdp']], hzn=2, est.start='1960 Q2', oos.start='1970 Q1', oos.end='2022 Q1')

pi.hat.4.rt = list()
name_columns = c('rt')
pi.hat.4.rt$uc2pi = matrix(nrow=208,ncol=1, dimnames = list(1:208,name_columns) )
pi.hat.4.rt$uc2pi = (t(pi_forecasts_df_uc2pi[1,14:222])+t(pi_forecasts_df_uc2pi[2,14:222]) + t(pi_forecasts_df_uc2pi[3,14:222]) + t(pi_forecasts_df_uc2pi[4,14:222]) + t(pi_forecasts_df_uc2pi[5,14:222]))/5
pi.hat.4.rt$uc3 = matrix(nrow=208,ncol=1, dimnames = list(1:208,name_columns) )
pi.hat.4.rt$uc3 = (t(pi_forecasts_df[1,14:222])+t(pi_forecasts_df[2,14:222]) + t(pi_forecasts_df[3,14:222]) + t(pi_forecasts_df[4,14:222]) + t(pi_forecasts_df[5,14:222]))/5
pi.hat.4.rt$ar <- rt.ar.forecast.inflation(P[['pgdp']], hzn=4, est.start='1960 Q2', oos.start='1970 Q1', oos.end='2022 Q1')

pi.hat.8.rt = list()
name_columns = c('rt')
pi.hat.8.rt$uc2pi = matrix(nrow=208,ncol=1, dimnames = list(1:208,name_columns) )
pi.hat.8.rt$uc2pi = (t(pi_forecasts_df_uc2pi[1,12:220])+t(pi_forecasts_df_uc2pi[2,12:220]) + t(pi_forecasts_df_uc2pi[3,12:220]) + t(pi_forecasts_df_uc2pi[4,12:220]) + t(pi_forecasts_df_uc2pi[5,12:220]) + t(pi_forecasts_df_uc2pi[6,12:220]) + t(pi_forecasts_df_uc2pi[7,12:220]) + t(pi_forecasts_df_uc2pi[8,12:220]) + t(pi_forecasts_df_uc2pi[9,12:220]))/9
pi.hat.8.rt$uc3 = matrix(nrow=208,ncol=1, dimnames = list(1:208,name_columns) )
pi.hat.8.rt$uc3 = (t(pi_forecasts_df[1,12:220])+t(pi_forecasts_df[2,12:220]) + t(pi_forecasts_df[3,12:220]) + t(pi_forecasts_df[4,12:220]) + t(pi_forecasts_df[5,12:220]) + t(pi_forecasts_df[6,12:220]) + t(pi_forecasts_df[7,12:220]) + t(pi_forecasts_df[8,12:220]) + t(pi_forecasts_df[9,12:220]))/9
pi.hat.8.rt$ar <- rt.ar.forecast.inflation(P[['pgdp']], hzn=8, est.start='1960 Q2', oos.start='1970 Q1', oos.end='2022 Q1')

tab <- list()
tab$pgdp.2 <- rmse.table(pi.hat.2.rt, P[['pgdp']][,'2022 Q2'], hzn=2, oos.start='1970 Q1', oos.end='2022 Q1') #the forecast is from vintage 1969Q3 to 2021Q3
tab$pgdp.4 <- rmse.table(pi.hat.4.rt, P[['pgdp']][,'2022 Q2'], hzn=4, oos.start='1970 Q1', oos.end='2022 Q1') #the forecast is from vintage 1969Q1 to 2021Q1
tab$pgdp.8 <- rmse.table(pi.hat.8.rt, P[['pgdp']][,'2022 Q2'], hzn=8, oos.start='1970 Q1', oos.end='2022 Q1') #the forecast is from vintage 1968Q3 to 2020Q3

stargazer(tab$pgdp.2, summary=FALSE)
stargazer(tab$pgdp.4, summary=FALSE)
stargazer(tab$pgdp.8, summary=FALSE)

tab.a7 = data.frame(tab$pgdp.2$rel.rmse,tab$pgdp.4$rel.rmse,tab$pgdp.8$rel.rmse)
rownames(tab.a7) <- c( 'UC (GDP/$\\pi$)', 'UC (GDP/u./$\\pi$)','AR')
stargazer(tab.a7, summary=FALSE)

# Table A8 Additional lags of cycle ----------------
rt.gap.ovn <- function(gap, P, hzn, est.start, oos.start, oos.end) {
  
  # This function does fully real-time forecasts of inflation,
  #   looping over vintages of P (price level) and the output gap
  
  # The auxilliary model: pi^h_t = alpha + \sum_i=1^6 (beta_i*pi_{t-i} + \gamma_i*gap_{t-i}) + e_t
  
  # Gap - name of model to be used; in matrix of dimension TxV
  # P - Name of matrix of vintage *price levels*
  # hzn - Forecast horizon
  # est.start - Start of estimation period
  # oos.start - First quarter to be forecasted
  # oos.end - Final quarter to be forecasted
  
  all.obs <- paste0(as.Date(as.yearqtr(est.start)), '/', as.Date(as.yearqtr(oos.end)))
  first.vint <- as.yearqtr(oos.start) - (hzn/4)
  last.vint <- as.yearqtr(oos.end) - (hzn/4)
  all.vints <- paste0(as.yearqtr(seq(as.Date(first.vint), as.Date(last.vint), by='quarter')))
  
  # Storage for forecasts
  pi.hat <- xts(array(NA, length(all.vints)), 
                order.by=as.yearqtr(seq(from=as.Date(as.yearqtr(oos.start)), 
                                        to=as.Date(as.yearqtr(oos.end)), 
                                        by='quarter')))
  
  for (v in 1:length(all.vints)) {
    
    vintage <- paste0(all.vints[v])
    
    P.v <- P[,vintage]
    gap.v <- gap[,vintage]
    
    y <- 400/hzn*(stats::lag(log(P.v), -hzn) - log(P.v)) # Note this leads!!
    x <- 400*(log(P.v) - stats::lag(log(P.v), 1)) # Backwards, quarterly
    
    lag.x <- stats::lag(x, seq(1,6))
    colnames(lag.x) <- paste0('l', seq(1,6), '.pi')
    
    data <- cbind(y, stats::lag(gap.v, 1:6), lag.x)
    colnames(data)[1] <- 'y'
    colnames(data)[2:7] <- paste0('l', 1:6, '.gap')
    
    # Given the NAs in the data this isn't needed but it's more clear
    last.qtr <- as.Date(as.yearqtr(as.Date(as.yearqtr(vintage)) %m+% months(-(hzn*3+1))))
    
    if (sum(complete.cases(data[paste0(as.Date(as.yearqtr(est.start)), '/', last.qtr)]))>0) {
      reg <- lm(y~., data=data[paste0(as.Date(as.yearqtr(est.start)), '/', last.qtr)]) 
      pi.hat[v] <- as.numeric(predict(reg, newdata=data[as.yearqtr(vintage), -1]))
    } else {
      pi.hat[v] <- NA
    }
  }
  return(pi.hat)
} # For Appendix some tables

ovn.hat <- list()
ovn.hat[['h2']] <- lapply(gap.vints.models, rt.gap.ovn, P=P[['pgdp']], hzn=2, est.start='1960 Q2', oos.start='1970 Q1', oos.end='2022 Q1')
ovn.hat[['final.h2']] <- lapply(final.gap.table, rt.gap.ovn, P=final.P[['pgdp']], hzn=2, est.start='1960 Q2', oos.start='1970 Q1', oos.end='2022 Q1')
ovn.hat[['h2']][['ar']] <- rt.ar.forecast.inflation(P[['pgdp']], hzn=2, est.start='1960 Q2', oos.start='1970 Q1', oos.end='2022 Q1')
ovn.hat[['final.h2']][['ar']] <- rt.ar.forecast.inflation(final.P[['pgdp']], hzn=2, est.start='1960 Q2', oos.start='1970 Q1', oos.end='2022 Q1')

ovn.hat[['h4']] <- lapply(gap.vints.models, rt.gap.ovn, P=P[['pgdp']], hzn=4, est.start='1960 Q2', oos.start='1970 Q1', oos.end='2022 Q1')
ovn.hat[['final.h4']] <- lapply(final.gap.table, rt.gap.ovn, P=final.P[['pgdp']], hzn=4, est.start='1960 Q2', oos.start='1970 Q1', oos.end='2022 Q1')
ovn.hat[['h4']][['ar']] <- rt.ar.forecast.inflation(P[['pgdp']], hzn=4, est.start='1960 Q2', oos.start='1970 Q1', oos.end='2022 Q1')
ovn.hat[['final.h4']][['ar']] <- rt.ar.forecast.inflation(final.P[['pgdp']], hzn=4, est.start='1960 Q2', oos.start='1970 Q1', oos.end='2022 Q1')

ovn.hat[['h8']] <- lapply(gap.vints.models, rt.gap.ovn, P=P[['pgdp']], hzn=8, est.start='1960 Q2', oos.start='1970 Q1', oos.end='2022 Q1')
ovn.hat[['final.h8']] <- lapply(final.gap.table, rt.gap.ovn, P=final.P[['pgdp']], hzn=8, est.start='1960 Q2', oos.start='1970 Q1', oos.end='2022 Q1')
ovn.hat[['h8']][['ar']] <- rt.ar.forecast.inflation(P[['pgdp']], hzn=8, est.start='1960 Q2', oos.start='1970 Q1', oos.end='2022 Q1')
ovn.hat[['final.h8']][['ar']] <- rt.ar.forecast.inflation(final.P[['pgdp']], hzn=8, est.start='1960 Q2', oos.start='1970 Q1', oos.end='2022 Q1')


tab.a8 <- list()
tab.a8$ovn.2 <- rmse.table(ovn.hat[['h2']], P[['pgdp']][,'2022 Q2'], hzn=2, oos.start='1970 Q1', oos.end='2022 Q1') 
tab.a8$ovn.4 <- rmse.table(ovn.hat[['h4']], P[['pgdp']][,'2022 Q2'], hzn=4, oos.start='1970 Q1', oos.end='2022 Q1') 
tab.a8$ovn.8 <- rmse.table(ovn.hat[['h8']], P[['pgdp']][,'2022 Q2'], hzn=8, oos.start='1970 Q1', oos.end='2022 Q1') 
tab.a8$final.ovn.2 <- rmse.table(ovn.hat[['final.h2']], P[['pgdp']][,'2022 Q2'], hzn=2, oos.start='1970 Q1', oos.end='2022 Q1') 
tab.a8$final.ovn.4 <- rmse.table(ovn.hat[['final.h4']], P[['pgdp']][,'2022 Q2'], hzn=4, oos.start='1970 Q1', oos.end='2022 Q1') 
tab.a8$final.ovn.8 <- rmse.table(ovn.hat[['final.h8']], P[['pgdp']][,'2022 Q2'], hzn=8, oos.start='1970 Q1', oos.end='2022 Q1') 

tab.a8

tab.a8.tex <- data.frame(tab.a8$ovn.2$rel.rmse, tab.a8$ovn.4$rel.rmse, tab.a8$ovn.8$rel.rmse,
                         tab.a8$final.ovn.2$rel.rmse, tab.a8$final.ovn.4$rel.rmse, tab.a8$final.ovn.8$rel.rmse)
rownames(tab.a8.tex) <- names(ovn.hat[['h2']])
stargazer(tab.a8.tex, summary=FALSE)


# Table A9 Additional lags of cycle, first half ----------------

ovn.hat <- list()
ovn.hat[['h2']] <- lapply(gap.vints.models, rt.gap.ovn, P=P[['pgdp']], hzn=2, est.start='1960 Q2', oos.start='1975 Q1', oos.end='1997 Q4')
ovn.hat[['final.h2']] <- lapply(final.gap.table, rt.gap.ovn, P=final.P[['pgdp']], hzn=2, est.start='1960 Q2', oos.start='1975 Q1', oos.end='1997 Q4')
ovn.hat[['h2']][['ar']] <- rt.ar.forecast.inflation(P[['pgdp']], hzn=2, est.start='1960 Q2', oos.start='1975 Q1', oos.end='1997 Q4')
ovn.hat[['final.h2']][['ar']] <- rt.ar.forecast.inflation(final.P[['pgdp']], hzn=2, est.start='1960 Q2', oos.start='1975 Q1', oos.end='1997 Q4')

ovn.hat[['h4']] <- lapply(gap.vints.models, rt.gap.ovn, P=P[['pgdp']], hzn=4, est.start='1960 Q2', oos.start='1975 Q1', oos.end='1997 Q4')
ovn.hat[['final.h4']] <- lapply(final.gap.table, rt.gap.ovn, P=final.P[['pgdp']], hzn=4, est.start='1960 Q2', oos.start='1975 Q1', oos.end='1997 Q4')
ovn.hat[['h4']][['ar']] <- rt.ar.forecast.inflation(P[['pgdp']], hzn=4, est.start='1960 Q2', oos.start='1975 Q1', oos.end='1997 Q4')
ovn.hat[['final.h4']][['ar']] <- rt.ar.forecast.inflation(final.P[['pgdp']], hzn=4, est.start='1960 Q2', oos.start='1975 Q1', oos.end='1997 Q4')

ovn.hat[['h8']] <- lapply(gap.vints.models, rt.gap.ovn, P=P[['pgdp']], hzn=8, est.start='1960 Q2', oos.start='1975 Q1', oos.end='1997 Q4')
ovn.hat[['final.h8']] <- lapply(final.gap.table, rt.gap.ovn, P=final.P[['pgdp']], hzn=8, est.start='1960 Q2', oos.start='1975 Q1', oos.end='1997 Q4')
ovn.hat[['h8']][['ar']] <- rt.ar.forecast.inflation(P[['pgdp']], hzn=8, est.start='1960 Q2', oos.start='1975 Q1', oos.end='1997 Q4')
ovn.hat[['final.h8']][['ar']] <- rt.ar.forecast.inflation(final.P[['pgdp']], hzn=8, est.start='1960 Q2', oos.start='1975 Q1', oos.end='1997 Q4')


tab.a9 <- list()
tab.a9$ovn.2 <- rmse.table(ovn.hat[['h2']], P[['pgdp']][,'2022 Q2'], hzn=2, oos.start='1975 Q1', oos.end='1997 Q4') 
tab.a9$ovn.4 <- rmse.table(ovn.hat[['h4']], P[['pgdp']][,'2022 Q2'], hzn=4, oos.start='1975 Q1', oos.end='1997 Q4') 
tab.a9$ovn.8 <- rmse.table(ovn.hat[['h8']], P[['pgdp']][,'2022 Q2'], hzn=8, oos.start='1975 Q1', oos.end='1997 Q4') 
tab.a9$final.ovn.2 <- rmse.table(ovn.hat[['final.h2']], P[['pgdp']][,'2022 Q2'], hzn=2, oos.start='1975 Q1', oos.end='1997 Q4') 
tab.a9$final.ovn.4 <- rmse.table(ovn.hat[['final.h4']], P[['pgdp']][,'2022 Q2'], hzn=4, oos.start='1975 Q1', oos.end='1997 Q4') 
tab.a9$final.ovn.8 <- rmse.table(ovn.hat[['final.h8']], P[['pgdp']][,'2022 Q2'], hzn=8, oos.start='1975 Q1', oos.end='1997 Q4') 

tab.a9

tab.a9.tex <- data.frame(tab.a9$ovn.2$rel.rmse, tab.a9$ovn.4$rel.rmse, tab.a9$ovn.8$rel.rmse,
                         tab.a9$final.ovn.2$rel.rmse, tab.a9$final.ovn.4$rel.rmse, tab.a9$final.ovn.8$rel.rmse)
rownames(tab.a9.tex) <- names(ovn.hat[['h2']])
stargazer(tab.a9.tex, summary=FALSE)


# Table A10 Additional lags of cycle, second half ----------------

final.gaps.2016.short = final.gaps.2016[c(-11,-14,-15)] #remove UC(CU), Tealbook, CBO

ovn.hat <- list()
ovn.hat[['h2.2016']] <- lapply(gap.vints.models, rt.gap.ovn, P=P[['pgdp']], hzn=2, est.start='1975 Q1', oos.start='1998 Q1', oos.end='2016 Q3')
ovn.hat[['h2.2016']][['ar']] <- rt.ar.forecast.inflation(P[['pgdp']], hzn=2, est.start='1975 Q1', oos.start='1998 Q1', oos.end='2016 Q3')
ovn.hat[['h4.2016']] <- lapply(gap.vints.models, rt.gap.ovn, P=P[['pgdp']], hzn=4, est.start='1975 Q1', oos.start='1998 Q1', oos.end='2016 Q3')
ovn.hat[['h4.2016']][['ar']] <- rt.ar.forecast.inflation(P[['pgdp']], hzn=4, est.start='1975 Q1', oos.start='1998 Q1', oos.end='2016 Q3')
ovn.hat[['h8.2016']] <- lapply(gap.vints.models, rt.gap.ovn, P=P[['pgdp']], hzn=8, est.start='1975 Q1', oos.start='1998 Q1', oos.end='2016 Q3')
ovn.hat[['h8.2016']][['ar']] <- rt.ar.forecast.inflation(P[['pgdp']], hzn=8, est.start='1975 Q1', oos.start='1998 Q1', oos.end='2016 Q3')

ovn.hat[['final.h2.2016']] <- lapply(final.gaps.2016.short, rt.gap.ovn, P=final.P[['pgdp']], hzn=2, est.start='1975 Q1', oos.start='1998 Q1', oos.end='2016 Q3')
ovn.hat[['final.h2.2016']][['ar']] <- rt.ar.forecast.inflation(final.P[['pgdp']], hzn=2, est.start='1975 Q1', oos.start='1998 Q1', oos.end='2016 Q3')
ovn.hat[['final.h4.2016']] <- lapply(final.gaps.2016.short, rt.gap.ovn, P=final.P[['pgdp']], hzn=4, est.start='1975 Q1', oos.start='1998 Q1', oos.end='2016 Q3')
ovn.hat[['final.h4.2016']][['ar']] <- rt.ar.forecast.inflation(final.P[['pgdp']], hzn=4, est.start='1975 Q1', oos.start='1998 Q1', oos.end='2016 Q3')
ovn.hat[['final.h8.2016']] <- lapply(final.gaps.2016.short, rt.gap.ovn, P=final.P[['pgdp']], hzn=8, est.start='1975 Q1', oos.start='1998 Q1', oos.end='2016 Q3')
ovn.hat[['final.h8.2016']][['ar']] <- rt.ar.forecast.inflation(final.P[['pgdp']], hzn=8, est.start='1975 Q1', oos.start='1998 Q1', oos.end='2016 Q3')


tab.a10 <- list()
tab.a10$pgdp.2 <- rmse.table(ovn.hat[['h2.2016']], P[['pgdp']][,'2016 Q4'], hzn=2, oos.start='1998 Q1', oos.end='2016 Q3') 
tab.a10$pgdp.4 <- rmse.table(ovn.hat[['h4.2016']], P[['pgdp']][,'2016 Q4'], hzn=4, oos.start='1998 Q1', oos.end='2016 Q3') 
tab.a10$pgdp.8 <- rmse.table(ovn.hat[['h8.2016']], P[['pgdp']][,'2016 Q4'], hzn=8, oos.start='1998 Q1', oos.end='2016 Q3') 
tab.a10$final.pgdp.2 <- rmse.table(ovn.hat[['final.h2.2016']], P[['pgdp']][,'2016 Q4'], hzn=2, oos.start='1998 Q1', oos.end='2016 Q3') 
tab.a10$final.pgdp.4 <- rmse.table(ovn.hat[['final.h4.2016']], P[['pgdp']][,'2016 Q4'], hzn=4, oos.start='1998 Q1', oos.end='2016 Q3') 
tab.a10$final.pgdp.8 <- rmse.table(ovn.hat[['final.h8.2016']], P[['pgdp']][,'2016 Q4'], hzn=8, oos.start='1998 Q1', oos.end='2016 Q3') 

tab.a10
tab.a10.tex <- data.frame(tab.a10$pgdp.2$rel.rmse, tab.a10$pgdp.4$rel.rmse, tab.a10$pgdp.8$rel.rmse,
                          tab.a10$final.pgdp.2$rel.rmse, tab.a10$final.pgdp.4$rel.rmse, tab.a10$final.pgdp.8$rel.rmse)
rownames(tab.a10.tex) <- names(ovn.hat[['h2.2016']])
stargazer(tab.a10.tex, summary=FALSE)

# Table A11 Rolling regressions, PGDP, full sample only -------
rt.gap.rolling <- function(gap, P, hzn, window=40, est.start='1960 Q1', oos.end='2022 Q1') {
  
  # This function does fully real-time forecasts of inflation,
  #   looping over vintages of P (price level) and the output gap
  
  # The auxilliary model: pi^h_t = alpha + beta_1*pi_{t-1} + beta_2*pi_{t-2} + ...
  #                                beta_6*pi_{t-6} + gamma*gap_{t-1} + e_t
  
  # Gap - name of model to be used; in matrix of dimension TxV
  # P - Name of matrix of vintage *price levels*
  # hzn - Forecast horizon
  # est.start - Start of estimation period
  # window - Size of window used to estimate PC parameters (in quarters!)
  # oos.end - Final quarter to be forecasted
  
  first.vint <- as.yearqtr(est.start) + 1/4 + window/4 # First vintage used 
  last.vint <- as.yearqtr(oos.end) - hzn/4
  all.vints <- paste0(as.yearqtr(seq(as.Date(first.vint), as.Date(last.vint), by='quarter')))
  
  first.hat <- as.yearqtr(est.start) + (1 + hzn + window)/4 # First quarter forecasted
  last.hat <- as.yearqtr(oos.end)
  
  # Storage for forecasts
  N <- (last.hat - first.hat)*4+1 # Number of forecasts to be made
  pi.hat <- xts(array(NA, N), order.by=as.yearqtr(seq(from=as.Date(as.yearqtr(first.hat)), 
                                                      to=as.Date(as.yearqtr(last.hat)), 
                                                      by='quarter')))
  
  for (v in 1:N) {
    
    vintage <- paste0(all.vints[v])
    
    P.v <- P[,vintage]
    gap.v <- gap[,vintage]
    
    y <- 400/hzn*(stats::lag(log(P.v), -hzn) - log(P.v)) # Note this leads!!
    x <- 400*(log(P.v) - stats::lag(log(P.v), 1)) # Backwards, quarterly
    
    lag.x <- stats::lag(x, seq(1,6))
    colnames(lag.x) <- paste0('l', seq(1,6), '.pi')
    
    data <- cbind(y, stats::lag(gap.v,1), lag.x)
    colnames(data)[1] <- 'y'
    colnames(data)[2] <- 'l.gap'
    
    # Given the NAs in the data this isn't needed but it's more clear
    last.qtr <- as.Date(as.yearqtr(as.Date(as.yearqtr(vintage)) %m+% months(-(hzn*3+1))))
    window <- paste(as.Date(as.yearqtr(vintage)-40/4), last.qtr, sep='/')
    
    if (sum(complete.cases(data[window]))>0) {
      reg <- lm(y~., data=data[window]) 
      pi.hat[v] <- as.numeric(predict(reg, newdata=data[as.yearqtr(vintage), -1]))
    } else {
      pi.hat[v] <- NA
    }
  }
  return(pi.hat)
} 
rt.ar.rolling <- function(P, hzn, window=40, est.start='1960 Q1', oos.end='2022 Q1') {
  
  # Here we produce AR(6) forecasts from an unrestricted AR model
  first.vint <- as.yearqtr(est.start) + 1/4 + window/4 # First vintage used 
  last.vint <- as.yearqtr(oos.end) - hzn/4
  all.vints <- paste0(as.yearqtr(seq(as.Date(first.vint), as.Date(last.vint), by='quarter')))
  
  first.hat <- as.yearqtr(est.start) + (1 + hzn + window)/4 # First quarter forecasted
  last.hat <- as.yearqtr(oos.end)
  
  # Storage for forecasts
  N <- (last.hat - first.hat)*4+1 # Number of forecasts to be made
  pi.hat <- xts(array(NA, N), order.by=as.yearqtr(seq(from=as.Date(as.yearqtr(first.hat)), 
                                                      to=as.Date(as.yearqtr(last.hat)), 
                                                      by='quarter')))
  
  for (v in 1:length(all.vints)) {
    
    if (all.vints[v] == '1996 Q1') {
      pi.hat[v] <- NA # Damned shutdown
      
    } else {
      
      vintage <- paste0(all.vints[v])
      vm1 <- as.yearqtr(as.Date(as.yearqtr(vintage)) %m+% months(-1)) # End of information set
      
      P.v <- P[,vintage]
      
      y <- 400*(log(P.v) - stats::lag(log(P.v), 1)) # Backwards, quarterly
      
      last.qtr <- as.Date(as.yearqtr(as.Date(as.yearqtr(vintage)) %m+% months(-3)))
      window <- paste(as.Date(as.yearqtr(vintage)-40/4), last.qtr, sep='/')
      
      # AR unrestricted model
      ar.mod <- ar(y[window], aic=FALSE, order.max=6)
      y.hat <- predict(ar.mod, n.ahead=hzn+1) # Need "+1" because of data lags
      pi.hat[v] <- mean(y.hat[['pred']])
    }
  }
  
  return(pi.hat)
}

roll.hat <- list()
roll.hat[['h2']] <- lapply(gap.vints.models, rt.gap.rolling, P=P[['pgdp']], hzn=2)
roll.hat[['h2']][['ar']] <- rt.ar.rolling(P[['pgdp']], hzn=2, est.start='1960 Q1', oos.end='2022 Q1')
roll.hat[['h4']] <- lapply(gap.vints.models, rt.gap.rolling, P=P[['pgdp']], hzn=4)
roll.hat[['h4']][['ar']] <- rt.ar.rolling(P[['pgdp']], hzn=4, est.start='1960 Q1', oos.end='2022 Q1')
roll.hat[['h8']] <- lapply(gap.vints.models, rt.gap.rolling, P=P[['pgdp']], hzn=8)
roll.hat[['h8']][['ar']] <- rt.ar.rolling(P[['pgdp']], hzn=8, est.start='1960 Q1', oos.end='2022 Q1')

roll.hat[['final.h2']] <- lapply(final.gap.table, rt.gap.rolling, P=final.P[['pgdp']], hzn=2)
roll.hat[['final.h2']][['ar']] <- rt.ar.rolling(final.P[['pgdp']], hzn=2, est.start='1960 Q1', oos.end='2022 Q1')
roll.hat[['final.h4']] <- lapply(final.gap.table, rt.gap.rolling, P=final.P[['pgdp']], hzn=4)
roll.hat[['final.h4']][['ar']] <- rt.ar.rolling(final.P[['pgdp']], hzn=4, est.start='1960 Q1', oos.end='2022 Q1')
roll.hat[['final.h8']] <- lapply(final.gap.table, rt.gap.rolling, P=final.P[['pgdp']], hzn=8)
roll.hat[['final.h8']][['ar']] <- rt.ar.rolling(final.P[['pgdp']], hzn=8, est.start='1960 Q1', oos.end='2022 Q1')


tab.a11 <- list()
tab.a11$roll.2 <- rmse.table(roll.hat[['h2']], P[['pgdp']][,'2022 Q2'], hzn=2, oos.start='1970 Q1', oos.end='2022 Q1')
tab.a11$roll.4 <- rmse.table(roll.hat[['h4']], P[['pgdp']][,'2022 Q2'], hzn=4, oos.start='1970 Q1', oos.end='2022 Q1')
tab.a11$roll.8 <- rmse.table(roll.hat[['h8']], P[['pgdp']][,'2022 Q2'], hzn=8, oos.start='1970 Q1', oos.end='2022 Q1')
tab.a11$final.roll.2 <- rmse.table(roll.hat[['final.h2']], P[['pgdp']][,'2022 Q2'], hzn=2, oos.start='1970 Q1', oos.end='2022 Q1') 
tab.a11$final.roll.4 <- rmse.table(roll.hat[['final.h4']], P[['pgdp']][,'2022 Q2'], hzn=4, oos.start='1970 Q1', oos.end='2022 Q1') 
tab.a11$final.roll.8 <- rmse.table(roll.hat[['final.h8']], P[['pgdp']][,'2022 Q2'], hzn=8, oos.start='1970 Q1', oos.end='2022 Q1') 

tab.a11

tab.a11.tex <- data.frame(tab.a11$roll.2$rel.rmse, tab.a11$roll.4$rel.rmse, 
                          tab.a11$roll.8$rel.rmse, tab.a11$final.roll.2$rel.rmse, 
                          tab.a11$final.roll.4$rel.rmse, tab.a11$final.roll.8$rel.rmse)
rownames(tab.a11.tex) <- names(roll.hat[['h2']])
stargazer(tab.a11.tex, summary=FALSE)


# Table A12 GDP forecasting, early sample -----------------------------

tab.a12 <- list()
tab.a12$gdp.4 <- rmse.table(gdp.hat[['h4']], Y[,'2022 Q2'], hzn=4, oos.start='1970 Q1', oos.end='2022 Q1') 
tab.a12$gdp.8 <- rmse.table(gdp.hat[['h8']], Y[,'2022 Q2'], hzn=8, oos.start='1970 Q1', oos.end='2022 Q1') 
tab.a12$gdp.12 <- rmse.table(gdp.hat[['h4']], Y[,'2022 Q2'], hzn=12, oos.start='1970 Q1', oos.end='2022 Q1') 
tab.a12$gdp.16 <- rmse.table(gdp.hat[['h8']], Y[,'2022 Q2'], hzn=16, oos.start='1970 Q1', oos.end='2022 Q1') 
tab.a12$final.gdp.4 <- rmse.table(gdp.hat[['final.h4']], Y[,'2022 Q2'], hzn=4, oos.start='1970 Q1', oos.end='2022 Q1') 
tab.a12$final.gdp.8 <- rmse.table(gdp.hat[['final.h8']], Y[,'2022 Q2'], hzn=8, oos.start='1970 Q1', oos.end='2022 Q1') 
tab.a12$final.gdp.12 <- rmse.table(gdp.hat[['final.h4']], Y[,'2022 Q2'], hzn=12, oos.start='1970 Q1', oos.end='2022 Q1') 
tab.a12$final.gdp.16 <- rmse.table(gdp.hat[['final.h8']], Y[,'2022 Q2'], hzn=16, oos.start='1970 Q1', oos.end='2022 Q1') 

tab.a12
sapply(tab.a12, FUN=function(x) x['ar', 'rmse'])
tab.a12.tex <- data.frame(tab.a12$gdp.4$rel.rmse, tab.a12$gdp.8$rel.rmse, tab.a12$gdp.12$rel.rmse, tab.a12$gdp.16$rel.rmse,
                          tab.a12$final.gdp.4$rel.rmse, tab.a12$final.gdp.8$rel.rmse, tab.a12$final.gdp.12$rel.rmse, tab.a12$final.gdp.16$rel.rmse)
rownames(tab.a12.tex) <- names(gdp.hat[['h2']])
stargazer(tab.a12.tex, summary=FALSE)

# Table A13 GDP forecasting, late sample -----------------------------
rt.ar <- function(P, hzn, est.start, oos.start, oos.end) {
  
  # Here we produce AR(6) forecasts from an unrestricted AR(6) models
  
  first.vint <- as.yearqtr(oos.start) - (hzn/4)
  last.vint <- as.yearqtr(oos.end) - (hzn/4)
  all.vints <- paste0(as.yearqtr(seq(as.Date(first.vint), as.Date(last.vint), by='quarter')))
  
  # Storage for forecasts
  pi.hat <- xts(array(NA, length(all.vints)), 
                order.by=as.yearqtr(seq(from=as.Date(as.yearqtr(oos.start)), 
                                        to=as.Date(as.yearqtr(oos.end)), 
                                        by='quarter')))
  
  for (v in 1:length(all.vints)) {
    
    if (all.vints[v] == '1996 Q1') {
      pi.hat[v] <- NA # Damned shutdown
      
    } else {
      
      vintage <- paste0(all.vints[v])
      vm1 <- as.yearqtr(as.Date(as.yearqtr(vintage)) %m+% months(-1)) # End of information set
      
      P.v <- P[,vintage]
      
      y <- 400*(log(P.v) - stats::lag(log(P.v), 1)) # Backwards, quarterly
      
      # AR unrestricted model
      ar.mod <- ar(y[paste0(as.Date(as.yearqtr(est.start)), '/', as.Date(vm1))], aic=FALSE, order.max=6)
      y.hat <- predict(ar.mod, n.ahead=hzn+1) # Need "+1" because of data lags
      pi.hat[v] <- mean(y.hat[['pred']])
    }
  }
  
  return(pi.hat)
}
rt.gap.gdp <- function(gap, Y, hzn, est.start, oos.start, oos.end) {
  
  # This function does fully real-time forecasts of GDP,
  #   looping over vintages of Y (GDP level) and the output gap
  
  # The auxilliary model: y^h_{t-1} = alpha + beta*gap_{t-1} + e_t
  
  # Gap - name of model to be used; in matrix of dimension TxV
  # P - Name of matrix of vintage *price levels*
  # hzn - Forecast horizon
  # est.start - Start of estimation period
  # oos.start - First quarter to be forecasted
  # oos.end - Final quarter to be forecasted
  
  all.obs <- paste0(as.Date(as.yearqtr(est.start)), '/', as.Date(as.yearqtr(oos.end)))
  first.vint <- as.yearqtr(oos.start) - (hzn/4)
  last.vint <- as.yearqtr(oos.end) - (hzn/4)
  all.vints <- paste0(as.yearqtr(seq(as.Date(first.vint), as.Date(last.vint), by='quarter')))
  
  # Storage for forecasts
  y.hat <- xts(array(NA, length(all.vints)), 
               order.by=as.yearqtr(seq(from=as.Date(as.yearqtr(oos.start)), 
                                       to=as.Date(as.yearqtr(oos.end)), 
                                       by='quarter')))
  
  for (v in 1:length(all.vints)) {
    
    vintage <- paste0(all.vints[v])
    
    Y.v <- Y[,vintage]
    gap.v <- gap[,vintage]
    
    y <- 400/hzn*(stats::lag(log(Y.v), -hzn) - log(Y.v)) # Note this leads!!
    
    data <- cbind(y, stats::lag(gap.v,1))
    colnames(data)[1] <- 'y'
    colnames(data)[2] <- 'l.gap'
    
    # Given the NAs in the data this isn't needed but it's more clear
    last.qtr <- as.Date(as.yearqtr(as.Date(as.yearqtr(vintage)) %m+% months(-(hzn*3+1))))
    
    if (sum(complete.cases(data[paste0(as.Date(as.yearqtr(est.start)), '/', last.qtr)]))>0) {
      reg <- lm(y~., data=data[paste0(as.Date(as.yearqtr(est.start)), '/', last.qtr)]) 
      y.hat[v] <- as.numeric(predict(reg, newdata=data[as.yearqtr(vintage), -1]))
    } else {
      y.hat[v] <- NA
    }
  }
  return(y.hat)
}

gdp.hat <- list()
gdp.hat[['h4.2016']] <- lapply(gap.vints, rt.gap.gdp, Y=Y, hzn=4, est.start='1975 Q1', oos.start='1998 Q1', oos.end='2016 Q3')
gdp.hat[['h4.2016']][['ar']] <- rt.ar(Y, hzn=4, est.start='1975 Q1', oos.start='1998 Q1', oos.end='2016 Q3')
gdp.hat[['h8.2016']] <- lapply(gap.vints, rt.gap.gdp, Y=Y, hzn=8, est.start='1975 Q1', oos.start='1998 Q1', oos.end='2016 Q3')
gdp.hat[['h8.2016']][['ar']] <- rt.ar(Y, hzn=8, est.start='1975 Q1', oos.start='1998 Q1', oos.end='2016 Q3')
gdp.hat[['h12.2016']] <- lapply(gap.vints, rt.gap.gdp, Y=Y, hzn=12, est.start='1975 Q1', oos.start='1998 Q1', oos.end='2016 Q3')
gdp.hat[['h12.2016']][['ar']] <- rt.ar(Y, hzn=12, est.start='1975 Q1', oos.start='1998 Q1', oos.end='2016 Q3')
gdp.hat[['h16.2016']] <- lapply(gap.vints, rt.gap.gdp, Y=Y, hzn=16, est.start='1975 Q1', oos.start='1998 Q1', oos.end='2016 Q3')
gdp.hat[['h16.2016']][['ar']] <- rt.ar(Y, hzn=16, est.start='1975 Q1', oos.start='1998 Q1', oos.end='2016 Q3')

gdp.hat[['final.h4.2016']] <- lapply(final.gaps.2016, rt.gap.gdp, Y=final.Y, hzn=4, est.start='1975 Q1', oos.start='1998 Q1', oos.end='2016 Q3')
gdp.hat[['final.h4.2016']][['ar']] <- rt.ar(final.Y, hzn=4, est.start='1975 Q1', oos.start='1998 Q1', oos.end='2016 Q3')
gdp.hat[['final.h8.2016']] <- lapply(final.gaps.2016, rt.gap.gdp, Y=final.Y, hzn=8, est.start='1975 Q1', oos.start='1998 Q1', oos.end='2016 Q3')
gdp.hat[['final.h8.2016']][['ar']] <- rt.ar(final.Y, hzn=8, est.start='1975 Q1', oos.start='1998 Q1', oos.end='2016 Q3')
gdp.hat[['final.h12.2016']] <- lapply(final.gaps.2016, rt.gap.gdp, Y=final.Y, hzn=12, est.start='1975 Q1', oos.start='1998 Q1', oos.end='2016 Q3')
gdp.hat[['final.h12.2016']][['ar']] <- rt.ar(final.Y, hzn=12, est.start='1975 Q1', oos.start='1998 Q1', oos.end='2016 Q3')
gdp.hat[['final.h16.2016']] <- lapply(final.gaps.2016, rt.gap.gdp, Y=final.Y, hzn=16, est.start='1975 Q1', oos.start='1998 Q1', oos.end='2016 Q3')
gdp.hat[['final.h16.2016']][['ar']] <- rt.ar(final.Y, hzn=16, est.start='1975 Q1', oos.start='1998 Q1', oos.end='2016 Q3')


tab.a13 <- list()
tab.a13$gdp.4 <- rmse.table(gdp.hat[['h4.2016']], Y[,'2016 Q4'], hzn=4, oos.start='1998 Q1', oos.end='2016 Q3') 
tab.a13$gdp.8 <- rmse.table(gdp.hat[['h8.2016']], Y[,'2016 Q4'], hzn=8, oos.start='1998 Q1', oos.end='2016 Q3') 
tab.a13$gdp.12 <- rmse.table(gdp.hat[['h12.2016']], Y[,'2016 Q4'], hzn=12, oos.start='1998 Q1', oos.end='2016 Q3') 
tab.a13$gdp.16 <- rmse.table(gdp.hat[['h16.2016']], Y[,'2016 Q4'], hzn=16, oos.start='1998 Q1', oos.end='2016 Q3') 

tab.a13$final.gdp.4 <- rmse.table(gdp.hat[['final.h4.2016']], Y[,'2016 Q4'], hzn=4, oos.start='1998 Q1', oos.end='2016 Q3') 
tab.a13$final.gdp.8 <- rmse.table(gdp.hat[['final.h8.2016']], Y[,'2016 Q4'], hzn=8, oos.start='1998 Q1', oos.end='2016 Q3') 
tab.a13$final.gdp.12 <- rmse.table(gdp.hat[['final.h12.2016']], Y[,'2016 Q4'], hzn=12, oos.start='1998 Q1', oos.end='2016 Q3') 
tab.a13$final.gdp.16 <- rmse.table(gdp.hat[['final.h16.2016']], Y[,'2016 Q4'], hzn=16, oos.start='1998 Q1', oos.end='2016 Q3') 

tab.a13
sapply(tab.a13, FUN=function(x) x['ar', 'rmse'])
tab.a13.tex <- data.frame(tab.a13$gdp.4$rel.rmse, tab.a13$gdp.8$rel.rmse, tab.a13$gdp.12$rel.rmse, tab.a13$gdp.16$rel.rmse,
                          tab.a13$final.gdp.4$rel.rmse, tab.a13$final.gdp.8$rel.rmse, tab.a13$final.gdp.12$rel.rmse, tab.a13$final.gdp.16$rel.rmse)
rownames(tab.a13.tex) <- names(gdp.hat[['h4.2016']])
stargazer(tab.a13.tex, summary=FALSE)

# Table A14 unemployment forecasting, early sample -----------------------------

rt.ar <- function(P, hzn, est.start, oos.start, oos.end) {
  
  # Here we produce AR(6) forecasts from an unrestricted AR(6) models
  
  first.vint <- as.yearqtr(oos.start) - (hzn/4)
  last.vint <- as.yearqtr(oos.end) - (hzn/4)
  all.vints <- paste0(as.yearqtr(seq(as.Date(first.vint), as.Date(last.vint), by='quarter')))
  
  # Storage for forecasts
  pi.hat <- xts(array(NA, length(all.vints)), 
                order.by=as.yearqtr(seq(from=as.Date(as.yearqtr(oos.start)), 
                                        to=as.Date(as.yearqtr(oos.end)), 
                                        by='quarter')))
  
  for (v in 1:length(all.vints)) {
    
    if (all.vints[v] == '1996 Q1') {
      pi.hat[v] <- NA # Damned shutdown
      
    } else {
      
      vintage <- paste0(all.vints[v])
      vm1 <- as.yearqtr(as.Date(as.yearqtr(vintage)) %m+% months(-1)) # End of information set
      
      P.v <- P[,vintage]
      
      y <- 1*(P.v - stats::lag(P.v, 1)) # Backwards, quarterly
      
      # AR unrestricted model
      ar.mod <- ar(y[paste0(as.Date(as.yearqtr(est.start)), '/', as.Date(vm1))], aic=FALSE, order.max=6)
      y.hat <- predict(ar.mod, n.ahead=hzn+1) # Need "+1" because of data lags
      pi.hat[v] <- mean(y.hat[['pred']])
    }
  }
  
  return(pi.hat)
}
rt.gap.u <- function(gap, Y, hzn, est.start, oos.start, oos.end) {
  
  # This function does fully real-time forecasts of GDP,
  #   looping over vintages of Y (GDP level) and the output gap
  
  # The auxilliary model: y^h_{t-1} = alpha + beta*gap_{t-1} + e_t
  
  # Gap - name of model to be used; in matrix of dimension TxV
  # P - Name of matrix of vintage *price levels*
  # hzn - Forecast horizon
  # est.start - Start of estimation period
  # oos.start - First quarter to be forecasted
  # oos.end - Final quarter to be forecasted
  
  all.obs <- paste0(as.Date(as.yearqtr(est.start)), '/', as.Date(as.yearqtr(oos.end)))
  first.vint <- as.yearqtr(oos.start) - (hzn/4)
  last.vint <- as.yearqtr(oos.end) - (hzn/4)
  all.vints <- paste0(as.yearqtr(seq(as.Date(first.vint), as.Date(last.vint), by='quarter')))
  
  # Storage for forecasts
  y.hat <- xts(array(NA, length(all.vints)), 
               order.by=as.yearqtr(seq(from=as.Date(as.yearqtr(oos.start)), 
                                       to=as.Date(as.yearqtr(oos.end)), 
                                       by='quarter')))
  
  for (v in 1:length(all.vints)) {
    
    vintage <- paste0(all.vints[v])
    
    Y.v <- Y[,vintage]
    gap.v <- gap[,vintage]
    
    y <- 1/hzn*(stats::lag(Y.v, -hzn) - Y.v) # Note this leads!!
    
    data <- cbind(y, stats::lag(gap.v,1))
    colnames(data)[1] <- 'y'
    colnames(data)[2] <- 'l.gap'
    
    # Given the NAs in the data this isn't needed but it's more clear
    last.qtr <- as.Date(as.yearqtr(as.Date(as.yearqtr(vintage)) %m+% months(-(hzn*3+1))))
    
    if (sum(complete.cases(data[paste0(as.Date(as.yearqtr(est.start)), '/', last.qtr)]))>0) {
      reg <- lm(y~., data=data[paste0(as.Date(as.yearqtr(est.start)), '/', last.qtr)]) 
      y.hat[v] <- as.numeric(predict(reg, newdata=data[as.yearqtr(vintage), -1]))
    } else {
      y.hat[v] <- NA
    }
  }
  return(y.hat)
}
rmse.table <- function(pi.hats, P.vintage, hzn, oos.start, oos.end) {
  
  # Inputs
  #   pi.hats - A list that contains the various model inflation forecasts
  #   P.vintage - The vintage of the unemployment level taken as "truth"
  #               Note P.vintage is a Tx1 matrix/vector
  #   hzn - Forecast horizon
  #   oos.start - Start of OOS period (i.e., first period forecasted)
  #   oos.end - End of OOS period (i.e., last period forecasted)
  
  all.obs <- seq(from=as.Date(as.yearqtr(oos.start)),
                 to=as.Date(as.yearqtr(oos.end)),
                 by='quarter')
  T <- length(all.obs)
  
  # First, define pi in the right way
  pi <- 1/hzn*(P.vintage - stats::lag(P.vintage, hzn))[all.obs]
  
  # First, calculate RMSEs
  rmses <- sapply(pi.hats, rmse, y=pi)
  numer <- rmses['ar']
  rel.rmses <- rmses/numer
  
  tab <- data.frame(rmse=format(round(rmses, 2), nsmall=2), 
                    rel.rmse=format(round(rel.rmses, 2), nsmall=2), 
                    d.bar = NA,
                    t=NA, 
                    p=NA,
                    N=sapply(pi.hats, function(x) sum(!is.na(x))))
  
  hats.no.ar <- pi.hats[-which(names(pi.hats)=='ar')]
  DM.output <- t(sapply(hats.no.ar, dm.test, y=pi, ynull=pi.hats[['ar']]))
  
  tab[,'d.bar'] <- c(format(round(DM.output[,'d.bar'], 2)), '--')
  tab[,'t'] <- c(format(round(DM.output[,'DM'], 2)), '--')
  tab[,'p'] <- c(format(round(DM.output[,'p.value'], 2)), '--')
  
  return(tab)
}

u.hat <- list()

u.hat[['h4']] <- lapply(gap.vints.models, rt.gap.u, Y=udata, hzn=4, est.start='1960 Q2', oos.start='1975 Q1', oos.end='1997 Q4')
u.hat[['final.h4']] <- lapply(final.gap.table, rt.gap.u, Y=final.udata, hzn=4, est.start='1960 Q2', oos.start='1975 Q1', oos.end='1997 Q4')
u.hat[['h4']][['ar']] <- rt.ar(udata, hzn=4, est.start='1960 Q2', oos.start='1975 Q1', oos.end='1997 Q4')
u.hat[['final.h4']][['ar']] <- rt.ar(final.udata, hzn=4, est.start='1960 Q2', oos.start='1975 Q1', oos.end='1997 Q4')

u.hat[['h8']] <- lapply(gap.vints.models, rt.gap.u, Y=udata, hzn=8, est.start='1960 Q2', oos.start='1975 Q1', oos.end='1997 Q4')
u.hat[['final.h8']] <- lapply(final.gap.table, rt.gap.u, Y=final.udata, hzn=8, est.start='1960 Q2', oos.start='1975 Q1', oos.end='1997 Q4')
u.hat[['h8']][['ar']] <- rt.ar(udata, hzn=8, est.start='1960 Q2', oos.start='1975 Q1', oos.end='1997 Q4')
u.hat[['final.h8']][['ar']] <- rt.ar(final.udata, hzn=8, est.start='1960 Q2', oos.start='1975 Q1', oos.end='1997 Q4')

u.hat[['h12']] <- lapply(gap.vints.models, rt.gap.u, Y=udata, hzn=12, est.start='1960 Q2', oos.start='1975 Q1', oos.end='1997 Q4')
u.hat[['final.h12']] <- lapply(final.gap.table, rt.gap.u, Y=final.udata, hzn=12, est.start='1960 Q2', oos.start='1975 Q1', oos.end='1997 Q4')
u.hat[['h12']][['ar']] <- rt.ar(udata, hzn=12, est.start='1960 Q2', oos.start='1975 Q1', oos.end='1997 Q4')
u.hat[['final.h12']][['ar']] <- rt.ar(final.udata, hzn=12, est.start='1960 Q2', oos.start='1975 Q1', oos.end='1997 Q4')

u.hat[['h16']] <- lapply(gap.vints.models, rt.gap.u, Y=udata, hzn=16, est.start='1960 Q2', oos.start='1975 Q1', oos.end='1997 Q4')
u.hat[['final.h16']] <- lapply(final.gap.table, rt.gap.u, Y=final.udata, hzn=16, est.start='1960 Q2', oos.start='1975 Q1', oos.end='1997 Q4')
u.hat[['h16']][['ar']] <- rt.ar(udata, hzn=16, est.start='1960 Q2', oos.start='1975 Q1', oos.end='1997 Q4')
u.hat[['final.h16']][['ar']] <- rt.ar(final.udata, hzn=16, est.start='1960 Q2', oos.start='1975 Q1', oos.end='1997 Q4')

tab.a14 <- list()
tab.a14$gdp.4 <- rmse.table(u.hat[['h4']], udata[,'2022 Q2'], hzn=4, oos.start='1975 Q1', oos.end='1997 Q4') 
tab.a14$gdp.8 <- rmse.table(u.hat[['h8']], udata[,'2022 Q2'], hzn=8, oos.start='1975 Q1', oos.end='1997 Q4') 
tab.a14$gdp.12 <- rmse.table(u.hat[['h12']], udata[,'2022 Q2'], hzn=12, oos.start='1975 Q1', oos.end='1997 Q4') 
tab.a14$gdp.16 <- rmse.table(u.hat[['h16']], udata[,'2022 Q2'], hzn=16, oos.start='1975 Q1', oos.end='1997 Q4') 

tab.a14$final.gdp.4 <- rmse.table(u.hat[['final.h4']], udata[,'2022 Q2'], hzn=4, oos.start='1975 Q1', oos.end='1997 Q4') 
tab.a14$final.gdp.8 <- rmse.table(u.hat[['final.h8']], udata[,'2022 Q2'], hzn=8, oos.start='1975 Q1', oos.end='1997 Q4') 
tab.a14$final.gdp.12 <- rmse.table(u.hat[['final.h12']], udata[,'2022 Q2'], hzn=12, oos.start='1975 Q1', oos.end='1997 Q4') 
tab.a14$final.gdp.16 <- rmse.table(u.hat[['final.h16']], udata[,'2022 Q2'], hzn=16, oos.start='1975 Q1', oos.end='1997 Q4') 

tab.a14
sapply(tab.a14, FUN=function(x) x['ar', 'rmse'])
tab.a14.tex <- data.frame(tab.a14$gdp.4$rel.rmse, tab.a14$gdp.8$rel.rmse, tab.a14$gdp.12$rel.rmse, tab.a14$gdp.16$rel.rmse, 
                        tab.a14$final.gdp.4$rel.rmse, tab.a14$final.gdp.8$rel.rmse, tab.a14$final.gdp.12$rel.rmse, tab.a14$final.gdp.16$rel.rmse)
rownames(tab.a14.tex) <- names(gdp.hat[['h4']])
stargazer(tab.a14.tex, summary=FALSE)


# Table A15 unemployment forecasting, late sample -----------------------------

rt.ar <- function(P, hzn, est.start, oos.start, oos.end) {
  
  # Here we produce AR(6) forecasts from an unrestricted AR(6) models
  
  first.vint <- as.yearqtr(oos.start) - (hzn/4)
  last.vint <- as.yearqtr(oos.end) - (hzn/4)
  all.vints <- paste0(as.yearqtr(seq(as.Date(first.vint), as.Date(last.vint), by='quarter')))
  
  # Storage for forecasts
  pi.hat <- xts(array(NA, length(all.vints)), 
                order.by=as.yearqtr(seq(from=as.Date(as.yearqtr(oos.start)), 
                                        to=as.Date(as.yearqtr(oos.end)), 
                                        by='quarter')))
  
  for (v in 1:length(all.vints)) {
    
    if (all.vints[v] == '1996 Q1') {
      pi.hat[v] <- NA # Damned shutdown
      
    } else {
      
      vintage <- paste0(all.vints[v])
      vm1 <- as.yearqtr(as.Date(as.yearqtr(vintage)) %m+% months(-1)) # End of information set
      
      P.v <- P[,vintage]
      
      y <- 1*(P.v - stats::lag(P.v, 1)) # Backwards, quarterly
      
      # AR unrestricted model
      ar.mod <- ar(y[paste0(as.Date(as.yearqtr(est.start)), '/', as.Date(vm1))], aic=FALSE, order.max=6)
      y.hat <- predict(ar.mod, n.ahead=hzn+1) # Need "+1" because of data lags
      pi.hat[v] <- mean(y.hat[['pred']])
    }
  }
  
  return(pi.hat)
}
rt.gap.u <- function(gap, Y, hzn, est.start, oos.start, oos.end) {
  
  # This function does fully real-time forecasts of GDP,
  #   looping over vintages of Y (GDP level) and the output gap
  
  # The auxilliary model: y^h_{t-1} = alpha + beta*gap_{t-1} + e_t
  
  # Gap - name of model to be used; in matrix of dimension TxV
  # P - Name of matrix of vintage *price levels*
  # hzn - Forecast horizon
  # est.start - Start of estimation period
  # oos.start - First quarter to be forecasted
  # oos.end - Final quarter to be forecasted
  
  all.obs <- paste0(as.Date(as.yearqtr(est.start)), '/', as.Date(as.yearqtr(oos.end)))
  first.vint <- as.yearqtr(oos.start) - (hzn/4)
  last.vint <- as.yearqtr(oos.end) - (hzn/4)
  all.vints <- paste0(as.yearqtr(seq(as.Date(first.vint), as.Date(last.vint), by='quarter')))
  
  # Storage for forecasts
  y.hat <- xts(array(NA, length(all.vints)), 
               order.by=as.yearqtr(seq(from=as.Date(as.yearqtr(oos.start)), 
                                       to=as.Date(as.yearqtr(oos.end)), 
                                       by='quarter')))
  
  for (v in 1:length(all.vints)) {
    
    vintage <- paste0(all.vints[v])
    
    Y.v <- Y[,vintage]
    gap.v <- gap[,vintage]
    
    y <- 1/hzn*(stats::lag(Y.v, -hzn) - Y.v) # Note this leads!!
    
    data <- cbind(y, stats::lag(gap.v,1))
    colnames(data)[1] <- 'y'
    colnames(data)[2] <- 'l.gap'
    
    # Given the NAs in the data this isn't needed but it's more clear
    last.qtr <- as.Date(as.yearqtr(as.Date(as.yearqtr(vintage)) %m+% months(-(hzn*3+1))))
    
    if (sum(complete.cases(data[paste0(as.Date(as.yearqtr(est.start)), '/', last.qtr)]))>0) {
      reg <- lm(y~., data=data[paste0(as.Date(as.yearqtr(est.start)), '/', last.qtr)]) 
      y.hat[v] <- as.numeric(predict(reg, newdata=data[as.yearqtr(vintage), -1]))
    } else {
      y.hat[v] <- NA
    }
  }
  return(y.hat)
}
rmse.table <- function(pi.hats, P.vintage, hzn, oos.start, oos.end) {
  
  # Inputs
  #   pi.hats - A list that contains the various model inflation forecasts
  #   P.vintage - The vintage of the unemployment level taken as "truth"
  #               Note P.vintage is a Tx1 matrix/vector
  #   hzn - Forecast horizon
  #   oos.start - Start of OOS period (i.e., first period forecasted)
  #   oos.end - End of OOS period (i.e., last period forecasted)
  
  all.obs <- seq(from=as.Date(as.yearqtr(oos.start)),
                 to=as.Date(as.yearqtr(oos.end)),
                 by='quarter')
  T <- length(all.obs)
  
  # First, define pi in the right way
  pi <- 1/hzn*(P.vintage - stats::lag(P.vintage, hzn))[all.obs]
  
  # First, calculate RMSEs
  rmses <- sapply(pi.hats, rmse, y=pi)
  numer <- rmses['ar']
  rel.rmses <- rmses/numer
  
  tab <- data.frame(rmse=format(round(rmses, 2), nsmall=2), 
                    rel.rmse=format(round(rel.rmses, 2), nsmall=2), 
                    d.bar = NA,
                    t=NA, 
                    p=NA,
                    N=sapply(pi.hats, function(x) sum(!is.na(x))))
  
  hats.no.ar <- pi.hats[-which(names(pi.hats)=='ar')]
  DM.output <- t(sapply(hats.no.ar, dm.test, y=pi, ynull=pi.hats[['ar']]))
  
  tab[,'d.bar'] <- c(format(round(DM.output[,'d.bar'], 2)), '--')
  tab[,'t'] <- c(format(round(DM.output[,'DM'], 2)), '--')
  tab[,'p'] <- c(format(round(DM.output[,'p.value'], 2)), '--')
  
  return(tab)
}
final.gap2016.table = final.gaps.2016[c(-11,-14,-15)] #remove UC(CU), Tealbook, CBO

u.hat <- list()

u.hat[['h4']] <- lapply(gap.vints.models, rt.gap.u, Y=udata, hzn=4, est.start='1960 Q2', oos.start='1998 Q1', oos.end='2016 Q3')
u.hat[['final.h4']] <- lapply(final.gap2016.table, rt.gap.u, Y=final.udata, hzn=4, est.start='1960 Q2', oos.start='1998 Q1', oos.end='2016 Q3')
u.hat[['h4']][['ar']] <- rt.ar(udata, hzn=4, est.start='1960 Q2', oos.start='1998 Q1', oos.end='2016 Q3')
u.hat[['final.h4']][['ar']] <- rt.ar(final.udata, hzn=4, est.start='1960 Q2', oos.start='1998 Q1', oos.end='2016 Q3')

u.hat[['h8']] <- lapply(gap.vints.models, rt.gap.u, Y=udata, hzn=8, est.start='1960 Q2', oos.start='1998 Q1', oos.end='2016 Q3')
u.hat[['final.h8']] <- lapply(final.gap2016.table, rt.gap.u, Y=final.udata, hzn=8, est.start='1960 Q2', oos.start='1998 Q1', oos.end='2016 Q3')
u.hat[['h8']][['ar']] <- rt.ar(udata, hzn=8, est.start='1960 Q2', oos.start='1998 Q1', oos.end='2016 Q3')
u.hat[['final.h8']][['ar']] <- rt.ar(final.udata, hzn=8, est.start='1960 Q2', oos.start='1998 Q1', oos.end='2016 Q3')

u.hat[['h12']] <- lapply(gap.vints.models, rt.gap.u, Y=udata, hzn=12, est.start='1960 Q2', oos.start='1998 Q1', oos.end='2016 Q3')
u.hat[['final.h12']] <- lapply(final.gap2016.table, rt.gap.u, Y=final.udata, hzn=12, est.start='1960 Q2', oos.start='1998 Q1', oos.end='2016 Q3')
u.hat[['h12']][['ar']] <- rt.ar(udata, hzn=12, est.start='1960 Q2', oos.start='1998 Q1', oos.end='2016 Q3')
u.hat[['final.h12']][['ar']] <- rt.ar(final.udata, hzn=12, est.start='1960 Q2', oos.start='1998 Q1', oos.end='2016 Q3')

u.hat[['h16']] <- lapply(gap.vints.models, rt.gap.u, Y=udata, hzn=16, est.start='1960 Q2', oos.start='1998 Q1', oos.end='2016 Q3')
u.hat[['final.h16']] <- lapply(final.gap2016.table, rt.gap.u, Y=final.udata, hzn=16, est.start='1960 Q2', oos.start='1998 Q1', oos.end='2016 Q3')
u.hat[['h16']][['ar']] <- rt.ar(udata, hzn=16, est.start='1960 Q2', oos.start='1998 Q1', oos.end='2016 Q3')
u.hat[['final.h16']][['ar']] <- rt.ar(final.udata, hzn=16, est.start='1960 Q2', oos.start='1998 Q1', oos.end='2016 Q3')

tab.a15 <- list()
tab.a15$gdp.4 <- rmse.table(u.hat[['h4']], udata[,'2016 Q4'], hzn=4, oos.start='1998 Q1', oos.end='2016 Q3') 
tab.a15$gdp.8 <- rmse.table(u.hat[['h8']], udata[,'2016 Q4'], hzn=8, oos.start='1998 Q1', oos.end='2016 Q3') 
tab.a15$gdp.12 <- rmse.table(u.hat[['h12']], udata[,'2016 Q4'], hzn=12, oos.start='1998 Q1', oos.end='2016 Q3') 
tab.a15$gdp.16 <- rmse.table(u.hat[['h16']], udata[,'2016 Q4'], hzn=16, oos.start='1998 Q1', oos.end='2016 Q3') 

tab.a15$final.gdp.4 <- rmse.table(u.hat[['final.h4']], udata[,'2016 Q4'], hzn=4, oos.start='1998 Q1', oos.end='2016 Q3') 
tab.a15$final.gdp.8 <- rmse.table(u.hat[['final.h8']], udata[,'2016 Q4'], hzn=8, oos.start='1998 Q1', oos.end='2016 Q3') 
tab.a15$final.gdp.12 <- rmse.table(u.hat[['final.h12']], udata[,'2016 Q4'], hzn=12, oos.start='1998 Q1', oos.end='2016 Q3') 
tab.a15$final.gdp.16 <- rmse.table(u.hat[['final.h16']], udata[,'2016 Q4'], hzn=16, oos.start='1998 Q1', oos.end='2016 Q3') 

tab.a15
sapply(tab.a15, FUN=function(x) x['ar', 'rmse'])
tab.a15.tex <- data.frame(tab.a15$gdp.4$rel.rmse, tab.a15$gdp.8$rel.rmse, tab.a15$gdp.12$rel.rmse, tab.a15$gdp.16$rel.rmse, 
                          tab.a15$final.gdp.4$rel.rmse, tab.a15$final.gdp.8$rel.rmse, tab.a15$final.gdp.12$rel.rmse, tab.a15$final.gdp.16$rel.rmse)
rownames(tab.a15.tex) <- names(gdp.hat[['h4']])
stargazer(tab.a15.tex, summary=FALSE)





# TABLE A16 (comparison against Tealbook) -------
rev.stats <- function(y.hat, y, first, last) {
  
  if ((!missing(first)) & (!missing(last))) {
    dates <- paste(as.Date(as.yearqtr(first)), as.Date(as.yearqtr(last)), sep='/')
    y <- y[dates]
    y.hat <- y.hat[dates]
    y['1995-10-01'] <- NA # Set to NA for all
    y.hat['1995-10-01'] <- NA # Set to NA for all
    y[which(is.na(y.hat)), ] <- NA # Sets QRT, finals to NA for CU if RT is missing
    
    x <- y - y.hat # Final less real-time
  } else {
    y['1995-10-01'] <- NA # Set to NA for all
    y.hat['1995-10-01'] <- NA # Set to NA for all
    y[which(is.na(y.hat)), ] <- NA # Sets QRT, finals to NA for CU if RT is missing
    
    x <- y - y.hat
  }
  
  N <- sum(!is.na(x))
  if (N>0) {
    mean <- mean(x, na.rm=TRUE)
    sd <- sd(x, na.rm=TRUE)
    min <- min(x, na.rm=TRUE)
    max <- max(x, na.rm=TRUE)
    rmse <- sqrt(mean(x^2, na.rm=TRUE))
    ns.sd <- sd / sd(y, na.rm=TRUE)
    ns.rmse <- sqrt(mean(x^2, na.rm=TRUE)) / sd(y, na.rm=TRUE)
    corr <- cor(y.hat, y, use='pairwise.complete.obs')
    sign <- mean(sign(y.hat)==sign(y), na.rm=TRUE)
  } else {
    mean <- sd <- min <- max <- rmse <- ns.sd <- ns.rmse <- corr <- sign <- NA
  }
  
  output <- list(N=format(N, nsmall=0),
                 mean=format(round(mean, 2), nsmall=2),
                 sd=format(round(sd, 2), nsmall=2), 
                 rmse=format(round(rmse, 2), nsmall=2), 
                 min=format(round(min, 2), nsmall=2), 
                 max=format(round(max, 2), nsmall=2),
                 ns.sd=format(round(ns.sd, 2), nsmall=2), 
                 ns.rmse=format(round(ns.rmse, 2), nsmall=2),
                 corr=format(round(corr, 2), nsmall=2), 
                 sign=format(round(sign*100), nsmall=0))
  return(output)
} # Revision statistics
tab.a16 <- t(sapply(gaps_main_results, function(x) rev.stats(x$rt, gaps$tb$rt, '1998 Q1', '2016 Q3')))
rownames(tab.a16) <- c('Linear', 'Quadratic', 'HP', 'GLMS', 'Hamilton', 'MW', 'KMW', 
                     'UC (GDP)', 'UC (GDP/$\\pi$)', 'UC (GDP/CU)', 'UC (GDP/u.)', 'UC (GDP/u./$\\pi$)',
                     'Tealbook', 'CBO')
tab.a16 = tab.a16[,c('N','mean','ns.sd','ns.rmse','sign')]
stargazer(tab.a16)

tab.a16b <- t(sapply(gaps_main_results, function(x) rev.stats(x$rt, gaps$tb$rt, '1975 Q1', '1997 Q4')))
rownames(tab.a16b) <- c('Linear', 'Quadratic', 'HP', 'GLMS', 'Hamilton', 'MW', 'KMW', 
                      'UC (GDP)', 'UC (GDP/$\\pi$)', 'UC (GDP/CU)', 'UC (GDP/u.)', 'UC (GDP/u./$\\pi$)',
                      'Tealbook', 'CBO')
tab.a16b = tab.a16b[,c('N','mean','ns.sd','ns.rmse','sign')]
stargazer(tab.a16b)


# TABLE A17 (comparison against Tealbook) -------
tab.a17 <- t(sapply(gaps_main_results, function(x) rev.stats(x$rt, gaps$tb$rt, '1975 Q1', '2016 Q3')))
rownames(tab.a17) <- c('Linear', 'Quadratic', 'HP', 'GLMS', 'Hamilton', 'MW', 'KMW', 
                     'UC (GDP)', 'UC (GDP/$\\pi$)', 'UC (GDP/CU)', 'UC (GDP/u.)', 'UC (GDP/u./$\\pi$)',
                     'Tealbook', 'CBO')
tab.a17 = tab.a17[,c('N','mean','sd','rmse','ns.sd','ns.rmse','sign')]
stargazer(tab.a17)

# Table A18 (Posterior parameters of UC3, final vintage) ----------
uc3 <- readRDS('../Data/UC3 final vintage.Rds')

param <- list()
param$alpha <- apply(uc3$res$alpha, 1, function(x) quantile(x, c(.05,.15,.5,.85,.95)))
param$beta <- apply(uc3$res$beta, 1, function(x) quantile(x, c(.05,.15,.5,.85,.95)))
param$phi <- apply(uc3$res$phi, 1, function(x) quantile(x, c(.05,.15,.5,.85,.95)))
param$omega <- apply(uc3$res$omega, 1, function(x) quantile(sqrt(x), c(.05,.15,.5,.85,.95)))

lapply(param, function(x) round(x,2))

post.mean <- list()
post.mean$alpha <- apply(uc3$res$alpha, 1, mean)
post.mean$beta <- apply(uc3$res$beta, 1, mean)
post.mean$phi <- apply(uc3$res$phi, 1, mean)
post.mean$omega <- apply(sqrt(uc3$res$omega), 1, mean)

round(cbind(param$alpha[1,], post.mean$alpha, param$alpha[5,]), 2)
round(cbind(param$beta[1,], post.mean$beta, param$beta[5,]), 2)
round(cbind(param$phi[1,], post.mean$phi, param$phi[5,]), 2)
round(cbind(param$omega[1,], post.mean$omega, param$omega[5,]), 2)


# TABLE A19 (no idiosyncratic error) -------
load('../Data/Vintage-datasets-Nov-2023.Rda') # Produced in wrangle-vintage-data.R
vintage.dates <- as.yearqtr(as.Date(sapply(rt.vintages, function(x) x[['vintage']])))

uc1 <- readRDS('../Raw model output/rt-uc1.Rds')
uc1.no.me <- readRDS('../Raw model output/rt-uc1.no.me.Rds')
final <- list()
final[['uc1']] <- uc1[[227]][['res']][['S']]; dimnames(final[['uc1']])[[2]] <- uc1[[227]][['prior']][['state.names']]
final[['uc1.no.me']] <- uc1.no.me[[227]][['res']][['S']]; dimnames(final[['uc1.no.me']])[[2]] <- uc1.no.me[[227]][['prior']][['state.names']]

gather.uc.output <- function(rt, qrt, vintage.dates) {
  
  V <- length(vintage.dates)
  rt.qtrs <- as.Date(vintage.dates) %m+% -months(3)
  c <- which(qrt[[1]][['prior']][['state.names']]=='c')  # An index for the cycle
  
  X <- xts(array(NA, c(V, 4)), 
           order.by=as.yearqtr(rt.qtrs))
  colnames(X) <- c('rt', 'qrt', 'final.2016', 'final')
  
  
  for (v in 1:V) {
    tm1 <- as.yearqtr(as.Date(vintage.dates[v]) %m+% -months(3))
    
    if ((v==1 && length(rt[[v]])==6) | length(rt[[v]])==2) {
      # Fill in real-time vintages
      S.rt <- apply(rt[[v]][['res']][['S']], c(1,2), mean)
      X[tm1, 'rt'] <- last(S.rt)[c]
      
      # Fill in quasi real-time vintages
      S.qrt <- apply(qrt[[v]][['res']][['S']], c(1,2), mean)
      X[tm1, 'qrt'] <- last(S.qrt)[c]
    }
  }
  
  # Fill in 2016Q4 vintage
  V.2016 <- which(vintage.dates == '2016 Q4')
  . <- apply(rt[[V.2016]][['res']][['S']], c(1,2), mean)
  S.v2016 <- xts(., order.by=as.yearqtr(seq(from=as.Date('1960-01-01'), 
                                            length.out=dim(.)[1], 
                                            by='quarter')))['1965-07-01/',]
  X[,'final.2016'] <- c(matrix(S.v2016[,c]), 
                        rep(NA, (V-dim(S.v2016)[1])))
  
  # Fill in 2019Q1 vintage
  . <- apply(rt[[V]][['res']][['S']], c(1,2), mean)
  S.V <- xts(., order.by=as.yearqtr(seq(from=as.Date('1960-01-01'), 
                                        length.out=dim(.)[1], 
                                        by='quarter')))['1965-07-01/',]  
  X[,'final'] <- matrix(S.V[,c])
  return(X)
}
gaps[['uc1.no.me']] <- gather.uc.output(uc1.no.me, uc1.no.me, vintage.dates)
rm('uc1', 'uc1.no.me')
cbind(rev.stats(gaps$uc1[,'rt'], gaps$uc1[,'final'], '1965 Q3', '2022 Q1'),
      rev.stats(gaps$uc1.no.me[,'rt'], gaps$uc1.no.me[,'final'], '1965 Q3', '2022 Q1'))

plot.uc1s.final <- function() {
  
  t <- as.numeric(as.yearqtr(index(S$uc1$mean)))
  c <- which(colnames(S$uc1$mean)=='c')
  
  ylim <- range(c(S$uc1$p95[,c], S$uc1$p05[,c]))
  
  plot(t, S$uc1$mean[,c], type='n', axes=FALSE, xlab='', ylab='', ylim=ylim)
  nberShade()
  
  polygon(c(t, rev(t)), c(matrix(S$uc1$p85[,c]), rev(matrix(S$uc1$p15[,c]))),
          col=adjustcolor('skyblue2', alpha.f=.7), border=NA)
  polygon(c(t, rev(t)), c(matrix(S$uc1$p95[,c]), rev(matrix(S$uc1$p05[,c]))),
          col=adjustcolor('skyblue2', alpha.f=.4), border=NA)
  abline(h=0)
  
  lines(t, S$uc1$mean[,c], col='navy', lw=1.25)
  lines(t, S$uc1.no.me$mean[,c], col='darkorange3', lw=1.25)
  
  axis(1, at=seq(1960, 2020, 10), cex.axis=.9)
  axis(1, at=seq(1960, 2020, 10), labels=FALSE, tck=-.02)  
  axis(2, cex.axis=.9)
  axis(4, labels=FALSE)  
  
  legend('bottomleft', c('UC1', 'UC1 (no GDP m.e.)'), col=c('navy', 'darkorange3'), pch=16,
         bty='n', ncol=1)
  
  box()
  title('Output gap estimates, 2022:Q2 vintage', adj=0, line=.2, cex.main=.8)
  title('%', adj=1, font.main=1, line=.2, cex.main=.8)
}
to.pdf(plot.uc1s.final(), '../Figures/FigureA2.pdf', height=10)
trim.pdf('../Figures/FigureA2.pdf')

# TABLE A20 (lag choice) -------
load('../Data/Vintage-datasets-Nov-2023.Rda') # Produced in wrangle-vintage-data.R
vintage.dates <- as.yearqtr(as.Date(sapply(rt.vintages, function(x) x[['vintage']])))
uc3 <- readRDS('../Raw model output/rt-uc3.Rds')
uc3.v2 <- readRDS('../Raw model output/rt-uc3.v2.Rds')

final <- list()
final[['uc3']] <- uc3[[227]][['res']][['S']]; dimnames(final[['uc3']])[[2]] <- uc3[[227]][['prior']][['state.names']]
final[['uc3.v2']] <- uc3.v2[[227]][['res']][['S']]; dimnames(final[['uc3.v2']])[[2]] <- uc3.v2[[227]][['prior']][['state.names']]
gather.uc.output <- function(rt, qrt, vintage.dates) {
  
  V <- length(vintage.dates)
  rt.qtrs <- as.Date(vintage.dates) %m+% -months(3)
  c <- which(qrt[[1]][['prior']][['state.names']]=='c')  # An index for the cycle
  
  X <- xts(array(NA, c(V, 4)), 
           order.by=as.yearqtr(rt.qtrs))
  colnames(X) <- c('rt', 'qrt', 'final.2016', 'final')
  
  
  for (v in 1:V) {
    tm1 <- as.yearqtr(as.Date(vintage.dates[v]) %m+% -months(3))
    
    if ((v==1 && length(rt[[v]])==6) | length(rt[[v]])==2) {
      # Fill in real-time vintages
      S.rt <- apply(rt[[v]][['res']][['S']], c(1,2), mean)
      X[tm1, 'rt'] <- last(S.rt)[c]
      
      # Fill in quasi real-time vintages
      S.qrt <- apply(qrt[[v]][['res']][['S']], c(1,2), mean)
      X[tm1, 'qrt'] <- last(S.qrt)[c]
    }
  }
  
  # Fill in 2016Q4 vintage
  V.2016 <- which(vintage.dates == '2016 Q4')
  . <- apply(rt[[V.2016]][['res']][['S']], c(1,2), mean)
  S.v2016 <- xts(., order.by=as.yearqtr(seq(from=as.Date('1960-01-01'), 
                                            length.out=dim(.)[1], 
                                            by='quarter')))['1965-07-01/',]
  X[,'final.2016'] <- c(matrix(S.v2016[,c]), 
                        rep(NA, (V-dim(S.v2016)[1])))
  
  # Fill in 2019Q1 vintage
  . <- apply(rt[[V]][['res']][['S']], c(1,2), mean)
  S.V <- xts(., order.by=as.yearqtr(seq(from=as.Date('1960-01-01'), 
                                        length.out=dim(.)[1], 
                                        by='quarter')))['1965-07-01/',]  
  X[,'final'] <- matrix(S.V[,c])
  return(X)
}
gaps[['uc3.v2']] <- gather.uc.output(uc3.v2, uc3.v2, vintage.dates)

cbind(rev.stats(gaps$uc3[,'rt'], gaps$uc3[,'final'], '1965 Q3', '2022 Q1'),
      rev.stats(gaps$uc3.v2[,'rt'], gaps$uc3.v2[,'final'], '1965 Q3', '2022 Q1'))

plot.uc3s.rt <- function() {
  
  t <- as.numeric(index(gaps$uc3))
  
  plot(t, gaps$uc3[,2], type='n', axes=FALSE, xlab='', ylab='', ylim=c(-10, 4))
  nberShade()
  abline(h=0)
  
  points(t, gaps$uc3[,1], col='navy', pch=16)
  points(t, gaps$uc3.v2[,1], col='darkorange3', pch=16)
  
  axis(1, at=seq(1960, 2020, 10), cex.axis=.9)
  axis(1, at=seq(1960, 2020, 10), labels=FALSE, tck=-.02)  
  axis(2, at=seq(-10, 4, 2), cex.axis=.9)
  axis(4, at=seq(-10, 4, 2), labels=FALSE)  
  
  legend(1966, -6, c('UC3', 'UC3 (ext. lags)'), col=c('navy', 'darkorange3'), pch=16,
         bty='n', ncol=1)
  
  box()
  title('Sequence of real-time output gap estimates', adj=0, line=.2, cex.main=.8)
  title('%', adj=1, font.main=1, line=.2, cex.main=.8)
}
to.pdf(plot.uc3s.rt(), '../Figures/FigureA3.pdf', height=10)
trim.pdf('../Figures/FigureA3.pdf')

get.params <- function(output) {
  
  V <- length(vintage.dates)
  
  okun <- array(NA, V)
  pc <- array(NA, V)
  
  for (v in 1:V) {
    okun[v] <- sum(apply(output[[v]][['res']][['alpha']], 1, mean))
    pc[v] <- sum(apply(output[[v]][['res']][['beta']], 1, mean))
  }
  
  params <- xts(cbind(okun, pc), order.by=as.yearqtr(vintage.dates))
  colnames(params) <- c('Okun', 'PC')
  return(params)
}
params <- list()
params$uc3 <- get.params(uc3)
params$v2 <- get.params(uc3.v2)
plot.okun <- function() {
  t <- as.numeric(as.yearqtr(vintage.dates))
  
  plot(t, params$uc3[,1], type='n', axes=FALSE, xlab='', ylab='', ylim=c(-1, 0))
  nberShade()
  abline(h=0)
  
  points(t, params$uc3[,'Okun'], col='navy', pch=16)
  points(t, params$v2[,'Okun'], col='darkorange3', pch=16)
  
  axis(1, at=seq(1960, 2020, 10), cex.axis=.9)
  axis(1, at=seq(1960, 2020, 10), labels=FALSE, tck=-.02)  
  axis(2, at=seq(-1, 0, .25), cex.axis=.9)
  axis(4, at=seq(-1, 0, .25), labels=FALSE)  
  
  legend(1966, -.5, c('UC3', 'UC3 (ext. lags)'), col=c('navy', 'darkorange3'), pch=16,
         bty='n', ncol=1)
  
  box()
  title('Sum of coefficients for unemployment rate', adj=0, line=.2, cex.main=.8)
  title('', adj=1, font.main=1, line=.2, cex.main=.8)
}
plot.pc <- function() {
  t <- as.numeric(as.yearqtr(vintage.dates))
  
  plot(t, params$uc3[,2], type='n', axes=FALSE, xlab='', ylab='', ylim=c(-.25, 1))
  nberShade()
  abline(h=0)
  
  points(t, params$uc3[,'PC'], col='navy', pch=16)
  points(t, params$v2[,'PC'], col='darkorange3', pch=16)
  
  axis(1, at=seq(1960, 2020, 10), cex.axis=.9)
  axis(1, at=seq(1960, 2020, 10), labels=FALSE, tck=-.02)  
  axis(2, at=seq(0, 1, .25), cex.axis=.9)
  axis(4, at=seq(0, 1, .25), labels=FALSE)  
  
  legend(1966, .95, c('UC3', 'UC3 (ext. lags)'), col=c('navy', 'darkorange3'), pch=16,
         bty='n', ncol=1)
  
  box()
  title('Sum of coefficients for inflation', adj=0, line=.2, cex.main=.8)
  title('', adj=1, font.main=1, line=.2, cex.main=.8)
}
to.pdf(plot.okun(), '../Figures/Okun.pdf', width=12)
to.pdf(plot.pc(), '../Figures/PC.pdf', width=12)
trim.pdf('../Figures/Okun.pdf')
trim.pdf('../Figures/PC.pdf')

plot.uc3.scatter <- function(model) {
  
  lim <- c(floor(min(c(gaps[[model]][,'rt'], gaps[[model]][,'final']), na.rm=TRUE)),
           ceiling(max(c(gaps[[model]][,'rt'], gaps[[model]][,'final']), na.rm=TRUE)))
  
  x <- matrix(gaps[[model]][,'rt'])
  y <- matrix(gaps[[model]][,'final'])
  
  if (model=='uc3') {
    col='navy'
    name='UC3'
  }
  if (model=='uc3.v2') {
    col='darkorange3'
    name='UC3 (ext. lags)'
  }
  
  plot(x,y, type='n', axes=FALSE, xlim=lim, ylim=lim,
       xlab='Real-time estimate', ylab='Final vintage estimate')
  points(x,y, pch=16, col=col)
  abline(a=0, b=1, col='black', lty=2, lw=1.5)
  
  axis(1)
  axis(2)  
  
  box()
  title(paste0(name, ' output gap estimates: real-time versus final'), adj=0, line=.2, cex.main=.8)
}
to.pdf(plot.uc3.scatter('uc3'), '../Figures/uc3_scatter.pdf', width=12)
to.pdf(plot.uc3.scatter('uc3.v2'), '../Figures/uc3_ext_scatter.pdf', width=12)
trim.pdf('../Figures/uc3_scatter.pdf')
trim.pdf('../Figures/uc3_ext_scatter.pdf')

rm('uc3', 'uc3.v2')
