
# This code produces real-time output gap estimates from 
#   a suite of UC models
# Travis Berge
# 12 August 2020

## Model
#    y_t  = y*_t + C_t + e_{y,t}
#
#    y*_t   = y*_{t-1} + eps_{y*,t}
#    C_t    = phi_1*C_{t-1} + phi_2*C_{t-2} + eps_{C,t}
#    e_it   = rho_i*e_{i,t-1} + eps_{i,t}
#
# Let: 
#    y_t = (y_t)'
#    s_t = (y*_t, C_t, C_{t-1}, e_1t)'
#
#    y_t = F*s_t
#    s_t = G*s_{t-1} + eps_t; eps_t ~ N(0,\Omega), omega diagonal

require(xts)
require(MASS)
require(mvtnorm)
require(mFilter)
require(lubridate)
require(parallel)

rm(list=ls())

set.seed(20551)
local.dir <- '../Code/'
setwd(local.dir)

load('../Data/Vintage-datasets-Nov-2023.Rda')

# Functions that define each model --------------
uc1u <- function(v, type, mcmc) {
  
  # Inputs: type="rt" or "qrt
  #         v -- the vintage of "rt.vintages" or "qrt.vintages"
  #         mcmc -- The list mcmc that defines 

  # Data -------------------------  
  vintages <- get(paste0(type, '.vintages'))
  inputs <- c('u') 
  keep <- which(colnames(vintages[[v]][['y']]) %in% inputs)
  
  data <- list()
  data$vintage <- vintages[[v]][['vintage']]
  data$y <- vintages[[v]][['y']][,keep] # This is the data
  data$u <- na.omit(vintages[[v]][['u']]['/1959']) # This is the pre-sample
  

  # Priors ---------------- 
  prior <- list()
  prior$T <- dim(data$y)[1]
  prior$Nobs <- dim(data$y)[2]
  
  prior$obs.names <- inputs
  prior$state.names <- c('y*', 'tau', 'c', 'l.c', 'e.u')
  prior$Ns <- length(prior$state.names)
  prior$phi.mu <- c(3/2, -2/3)
  prior$p <- length(prior$phi.mu)
  prior$phi.V <- c(2,2)
  
  # Prior for omega
  utrend <- hpfilter(data$u, 1600)$trend
  ucycle <- hpfilter(data$u, 1600)$cycle
  e <- embed(ucycle, prior$p+1) %*% c(1, -prior$phi.mu)
  prior$omega.mean <- c(as.numeric(var(diff(utrend))), # y^*
                        as.numeric(var(diff(diff(utrend)))), # tau*
                        as.numeric(var(e)), # c
                        1) #e.u
  prior$omega.nu <- c(prior$T/20,
                      prior$T/20,
                      1,
                      1)
  prior$omega.S <- prior$omega.nu * prior$omega.mean

  # Initial state vector  
  prior$S0 <- c(as.numeric(last(data$u)),
                as.numeric(mean(diff(data$u), na.rm=TRUE)),
                0, 0, 0)
  names(prior$S0) <- prior$state.names
  prior$P00 <- diag(c(1, 1, 10, 10, 10))
  colnames(prior$P00) <- rownames(prior$P00) <- prior$state.names

  
  # Initialize ----------------- 
  draw <- list()
  draw$omega <- prior$omega.mean
  names(draw$omega) <- prior$state.names[-4]
  
  draw$S <- cbind(data$y[,1], prior$S0[2], 0, 0, 0)
  colnames(draw$S) <- prior$state.names
  
  
  # Functions for draws --------------------
  build.F <- function(prior, draw) {
    
    F <- matrix(0, nrow = prior$Nobs, ncol = prior$Ns)
    F[1, ] <- c(1, 0, 1, 0, 1)
    
    rownames(F) <- prior$obs.names
    colnames(F) <- prior$state.names
    
    return(F)
  }
  build.G <- function(prior, draw) {
    
    G <- matrix(0, nrow = prior$Ns, ncol = prior$Ns)
    G[1, ] <- c(1, 0, 0, 0, 0)
    G[2, ] <- c(0, 1, 0, 0, 0)
    G[3, ] <- c(0, 0, draw$phi[1], draw$phi[2], 0) 
    G[4, ] <- c(0, 0, 1, 0, 0)
    G[5, ] <- c(0, 0, 0, 0, 0)
    
    colnames(G) <- rownames(G) <- prior$state.names
    
    return(G)
  }
  build.H <- function(prior) {
    
    H <- matrix(0, nrow=prior$Ns, ncol=prior$Ns-1)
    rownames(H) <- prior$state.names 
    colnames(H) <- prior$state.names[-4]
    H[1:3, 1:3] <- diag(3)
    H[5:5, 4:4] <- diag(1)
    
    return(H)
  } 
  
  ar.stable <- function(phi, has.constant=FALSE) {
    
    # dont include constant
    if (has.constant==TRUE) {
      k <- length(phi) - 1
      phi <- phi[-1]
    } else {
      k <- length(phi)
    }
    
    # Put in companion form
    F <- matrix(0, nrow=k, ncol=k)
    F[1,] <- phi
    if (k>1) F[2:k, 1:(k-1)] <- diag(k-1) # Identity matrix
    
    max.e <- max(abs(eigen(F)$values))
    if (max.e<=1) { # Stable
      check <- 1
    } else { # Not stable
      check <- 0
    }
    return(check)
  } # Checks if AR(p) process is stable
  draw.phi <- function(prior, draw) {
    
    y <- embed(as.matrix(draw$S[, 'c']), prior$p + 1)[, 1]
    x <- embed(as.matrix(draw$S[, 'c']), prior$p + 1)[, prior$p:(prior$p + 1)]
    
    omegainv <- 1/draw$omega['c']
    
    P <- solve(solve(diag(prior$phi.V)) + omegainv*crossprod(x))
    p <- P %*% (solve(diag(prior$phi.V)) %*% prior$phi.mu + omegainv * crossprod(x, y))

    check <- 0 # check for stability
    while(check!=1) { 
      phi <- mvrnorm(mu=p, Sigma=P)
      check <- ar.stable(phi, has.constant=FALSE)
    }
    return(phi)
    
  }
  draw.omega <- function(prior, draw) {
    
    omega <- array(NA, prior$Ns-1)
    names(omega) <- prior$state.names[-4]
    
    S <- as.matrix(draw$S)
    
    y <- S[2:prior$T, ]
    x <- S[1:(prior$T-1), ]
    G <- build.G(prior, draw)
    e <- y - x %*% t(G)
    
    j <- 1
    for (i in c(1:3, 5)) {
      nu1 <- prior$omega.nu[j] + prior$T - 1
      S1 <- prior$omega.S[j] + crossprod(e)[i, i]
      omega[j] <- 1 / rgamma(n=1, nu1/2, S1/2)
      j <- j + 1
    }

    return(omega)
  }
  KFKS <- function(Y, prior, model) {
    # Kalman filter for stationary systems with missing data.
    # The model is        y_t = A + F*s_t
    #                     s_t+1 = M + G*s_t + H*u_t; u \sim N(0,Omega_t)
    #
    # Inputs
    #       prior         A list of priors
    #       model         A list of model parameters (as in matrices above)
    #       data          An array of dimension (prior$T x prior$Nobs)
    #
    # Output
    #        KF$st       Predicted state vector  s_t|t-1      (T x Ns)
    #        KF$stt      Filtered  state vector  s_t|t        (T x Ns)
    #        KF$Pt       Predicted covariance of s_t|t-1      (Ns x Ns x T)
    #        KF$Ptt      Filtered  covariance of s_t|t        (Ns x Ns)
    #        KF$ll       Value of likelihood function         (Tx1)
    
    A <- model$A
    M <- model$M
    F <- model$F
    G <- model$G
    H <- model$H
    Omega <- model$omega
    
    # Output structure & dimensions
    KF <- list()
    KF$st <- array(NA, c(prior$T, prior$Ns)) # s_t|t-1
    KF$Pt <- array(NA, c(prior$Ns, prior$Ns, prior$T)) # P_t|t-1
    KF$stt <- array(NA, c(prior$T, prior$Ns))
    KF$Ptt <- array(NA, c(prior$Ns, prior$Ns, prior$T))
    KF$vt <- array(0, c(prior$T, prior$Nobs)) # v_t = y_t - F%*%st|t-1
    KF$Kt <- array(0, c(prior$Ns, prior$Nobs, prior$T))
    KF$Ftinv <- array(0, c(prior$Nobs, prior$Nobs, prior$T))
    KF$ll <- array(NA, c(prior$T))
    
    # Initial values
    stt <- mvrnorm(1, model$S0, model$P0) # s_0|0
    Ptt <- model$P0 # P_0|0
    
    for (t in 1:prior$T) {
      
      # Predict
      st <- M + G%*%stt
      Pt <- G%*%Ptt%*%t(G) + H%*%Omega%*%t(H)
      Pt <- 1/2*(Pt+t(Pt))
      
      # Handle the missing data
      y <- as.vector(Y[t,])
      i <- !is.na(y)
      y.star <- y[i]
      A.star <- A[i]
      F.star <- matrix(F[i,], ncol=prior$Ns)
      
      # Update
      if (sum(i)==0) {
        stt <- st
        Ptt <- Pt
      } else {
        Ftinv <- solve(F.star%*%Pt%*%t(F.star))
        Mt <- G%*%Pt%*%t(F.star)
        Kt <- Mt%*%Ftinv
        vt <- y.star - A.star - F.star%*%st
        stt <- st + Pt%*%t(F.star)%*%Ftinv%*%vt
        Ptt <- Pt - Pt%*%t(F.star)%*%Ftinv%*%F.star%*%Pt
      }
      
      KF$st[t,]<- st
      KF$Pt[,,t] <- Pt
      KF$vt[t,i] <- vt
      KF$stt[t,] <- stt
      KF$Ptt[,,t] <- Ptt
      KF$Kt[,i,t] <- Kt
      KF$Ftinv[i,i,t] <- Ftinv
      KF$ll[t] <- dmvnorm(drop(vt), sigma=solve(Ftinv), log=T)
    }
    
    # Smoothing filter
    KS <- list()
    KS$stT <- array(NA, c(prior$T, prior$Ns))
    KS$PtT <- array(NA, c(prior$Ns, prior$Ns, prior$T))
    KS$Ct <- array(NA, c(prior$Ns, prior$Nobs, prior$T))
    KS$Nt <- array(NA, c(prior$Ns, prior$Ns, prior$T))
    KS$Lt <- array(NA, c(prior$Ns, prior$Ns, prior$T))
    
    r.tm1 <- array(0, prior$Ns)
    N.tm1 <- array(0, c(prior$Ns, prior$Ns))
    for (t in prior$T:1) {
      
      i <- !is.na(Y[t,])
      
      rt <- r.tm1
      Nt <- N.tm1
      Lt <- G - KF$Kt[,,t]%*%F
      
      if (sum(i)==0) {
        r.tm1 <- t(Lt)%*%rt
        N.tm1 <- t(Lt)%*%Nt%*%Lt
        stT <- KF$st[t,] + KF$Pt[,,t]%*%r.tm1
        PtT <- KF$Pt[,,t] - KF$Pt[,,t]%*%N.tm1%*%KF$Pt[,,t]
        Dt <- t(KF$Kt[,,t]) %*% Nt %*% KF$Kt[,,t]
        Ct <- t(F)%*%Dt - t(G)%*%Nt%*%KF$Kt[,,t]
      } else {
        r.tm1 <- t(F)%*%KF$Ftinv[,,t]%*%KF$vt[t,] + t(Lt)%*%rt
        N.tm1 <- t(F)%*%KF$Ftinv[,,t]%*%F + t(Lt)%*%Nt%*%Lt
        stT <- KF$st[t,] + KF$Pt[,,t]%*%r.tm1
        PtT <- KF$Pt[,,t] - KF$Pt[,,t]%*%N.tm1%*%KF$Pt[,,t]
        Dt <- KF$Ftinv[,,t] + t(KF$Kt[,,t]) %*% Nt %*% KF$Kt[,,t]
        Ct <- t(F)%*%Dt - t(G)%*%Nt%*%KF$Kt[,,t]
      }
      
      KS$stT[t,] <- stT
      KS$PtT[,,t] <- PtT
      KS$Ct[,,t] <- Ct
      KS$Nt[,,t] <- Nt
      KS$Lt[,,t] <- Lt
    }
    K <- list()
    K$KF <- KF
    K$KS <- KS
    return(K)
  } # Kalman filter & smoother as in KH 2003
  draw.S.dk <- function(data, prior, draw) {
    
    # This implements Algorithm 2a described in 
    # Jarocinski (2015) and using auxilliary function KFKS
    
    # Write the model following DK 2002:
    # y_t = A + F*s_t
    # s_{t+1} = M + G*s_t + H*eps_t, eps_t ~ N(0,\omega_t)
    
    # Note: varience of f is time-varying!
    A <- matrix(0, prior$Nobs)
    M <- matrix(0, prior$Ns)
    F <- build.F(prior, draw)
    G <- build.G(prior, draw)
    H <- build.H(prior)
    omega <- diag(draw$omega)
    
    # Step 1. Draw s+ (a+) and y+
    a.plus <- array(NA, c(prior$T, prior$Ns))
    Y.plus <- array(NA, c(prior$T, prior$Nobs))
    
    for (t in 1:prior$T) {
      
      w.plus <- mvrnorm(mu=rep(0, prior$Ns), Sigma=H%*%omega%*%t(H))
      
      if (t==1) {
        a.plus[t,] <- mvrnorm(1, prior$S0, prior$P0) # Initialize
      } else {
        a.plus[t,] <- M + G%*%a.plus[(t-1),] + w.plus
      }
      Y.plus[t,] <- A + F%*%a.plus[t,]
    }
    
    # Step 2. 
    Y.star <- data$y - Y.plus
    
    # Now use KFKS function to calculate E(A|Y*)
    # Reset S0, A, M to 0 as in Jarocinski (2015)
    mod <- list()
    mod$S0 <- rep(0, prior$Ns) 
    mod$P0 <- prior$P0
    mod$A <- matrix(0, prior$Nobs)
    mod$M <- matrix(0, prior$Ns)
    mod$F <- F
    mod$G <- G
    mod$H <- H
    mod$omega <- omega
    
    K <- KFKS(Y.star, prior, mod)
    S <- K$KS$stT + a.plus
    colnames(S) <- prior$state.names
    return(S)
    
  } # Draw S from DK (2002) / Jarocinski (2015)
  
  
  # MCMC -----
  # Storage
  res <- list()
  res$phi <- array(NA, c(2, mcmc$keep))
  res$omega <- array(NA, c(prior$Ns-1, mcmc$keep))
  res$S <- array(NA, c(prior$T, prior$Ns, mcmc$keep))

  for (i in 1:mcmc$nit) {
    
#    if ((i/10)%%1 == 0) cat('.')
#    if ((i/mcmc$print)%%1 == 0) cat(' Draw', i, 'of', mcmc$nit, '\n')
    
    draw$phi <- draw.phi(prior, draw)
    draw$omega <- draw.omega(prior, draw)
    draw$S <- draw.S.dk(data, prior, draw)
  
    if (i>mcmc$burn && ((i-mcmc$burn)/mcmc$thin)%%1==0) {
      
      j <- (i - mcmc$burn) / mcmc$thin
      
      res$phi[,j] <- draw$phi
      res$omega[,j] <- draw$omega
      res$S[,,j] <- draw$S
    }
  }
  
  cat('Vintage ', v, ' done')
  
  # Output to save ------
  if (data[['vintage']]=='1965-11-15') {
    return(list(prior=prior, res=res, 
                KFKS=KFKS, build.F=build.F, build.G=build.G, build.H=build.H))
  } else {
    return(list(prior=prior, res=res))
  }
}


# Parameters for mcmc routines ------------------
mcmc <- list()
mcmc$burn <- 2000
mcmc$thin <- 10
mcmc$keep <- 1000
mcmc$nit <- mcmc$burn + mcmc$keep*mcmc$thin 
mcmc$print <- 500

# Common hyperparameters for parallel loops 
V.rt <- length(rt.vintages)
V.qrt <- length(qrt.vintages)

# UC1u ----------------------
model='uc1u'

cat("Starting on real-time vintages: \n")

# RT
type='rt'
vintage.output <- mclapply(1:V.rt, model, type=type, mcmc=mcmc, mc.cores=detectCores()-1)
saveRDS(vintage.output, file=paste0('../Data/', type, '-', model, '.Rds'))

cat("Starting on quasi-real-time vintages: \n")

# QRT
type='qrt'
vintage.output <- mclapply(1:V.qrt, model, type=type, mcmc=mcmc, mc.cores=detectCores()-1)
saveRDS(vintage.output, file=paste0('../Data/', type, '-', model, '.Rds'))

