
rm(list=ls())
local.dir <- '../Code/'
data.dir <- '../Data/'
setwd(local.dir)

require(xts)
require(mFilter)
require(lubridate)
require(readxl)

# Data and hyperparameters -----------------
raw <- data.frame(read_excel('../Data/Vintages-RGDP.xlsx')) # Quarterly RGDP Vintages
gdp <- xts(log(raw[,-1]), order.by=as.yearqtr(seq(from=as.Date('1947-01-01'), 
                                                           to=as.Date('2022-01-01'), by='quarter')))
colnames(gdp) <- paste0(seq(from=as.Date('1965-11-15'), to=as.Date('2022-05-15'), by='quarter'))

gdp.vintages <- gdp['1960/'] # Limit estimation period!!
rm(raw)

T <- dim(gdp.vintages)[1]
V <- dim(gdp.vintages)[2]
rt <- list() # Storage for univariate real-time
qrt <- list() # Storage for quasi real-time

# linear trend ---------------------
rt$line <- xts(array(NA, c(T,V)), order.by=index(gdp.vintages))
qrt$line <- xts(array(NA, c(T,V)), order.by=index(gdp.vintages))
colnames(rt$line) <- colnames(qrt$line) <- colnames(gdp.vintages)

for (v in 1:V) {
  Y <- gdp.vintages[!is.na(gdp.vintages[,v]), v]
  x <- 1:length(Y)
  c <- residuals(lm(Y~x))*100
  rt$line[which(index(rt$line) %in% index(c)), v] <- c

  tm1 <- which(index(gdp.vintages) == as.yearqtr(as.Date(colnames(gdp.vintages)[v]) %m-% months(3))) # index for expanding window
  Y <- gdp.vintages[1:tm1, V]
  x <- 1:length(Y)
  c <- residuals(lm(Y~x))*100
  qrt$line[which(index(qrt$line) %in% index(c)), v] <- c
}


# Quadratic trend ---------------------
rt$quad <- xts(array(NA, c(T,V)), order.by=index(gdp.vintages))
qrt$quad <- xts(array(NA, c(T,V)), order.by=index(gdp.vintages))
colnames(rt$quad) <- colnames(qrt$quad) <- colnames(gdp.vintages)

for (v in 1:V) {
  Y <- gdp.vintages[!is.na(gdp.vintages[,v]), v]
  x <- 1:length(Y)
  x2 <- (1:length(Y))^2
  c <- residuals(lm(Y~x+x2))*100
  rt$quad[which(index(rt$quad) %in% index(c)),v] <- c
  
  tm1 <- which(index(gdp.vintages) == as.yearqtr(as.Date(colnames(gdp.vintages)[v]) %m-% months(3))) # index for expanding window
  Y <- gdp.vintages[1:tm1, V]
  x <- 1:length(Y)
  x2 <- (1:length(Y))^2
  c <- residuals(lm(Y~x+x2))*100
  qrt$quad[which(index(qrt$quad) %in% index(c)), v] <- c
}


# HP filter ---------------------------
rt$hp <- xts(array(NA, c(T,V)), order.by=index(gdp.vintages))
qrt$hp <- xts(array(NA, c(T,V)), order.by=index(gdp.vintages))
colnames(qrt$hp) <- colnames(rt$hp) <- colnames(gdp.vintages)

for (v in 1:V) {
  Y <- gdp.vintages[!is.na(gdp.vintages[,v]), v]
  c <- xts(hpfilter(Y, freq=1600, type='lambda')$cycle*100, order.by=index(Y))
  rt$hp[which(index(rt$hp) %in% index(c)), v] <- c
  
  tm1 <- which(index(gdp.vintages) == as.yearqtr(as.Date(colnames(gdp.vintages)[v]) %m-% months(3))) # index for expanding window
  Y <- gdp.vintages[1:tm1, V]
  c <- xts(hpfilter(Y, freq=1600, type='lambda')$cycle*100, order.by=index(Y))
  qrt$hp[which(index(qrt$hp) %in% index(c)), v] <- c
}


# Hamilton filter ---------------------
ham.trend <- function(Y, P=4, h=8) {

  X <- embed(Y, (P+h-1))
  y <- X[,1]
  x <- X[,(h-1):(P+h-1)]

  y.trend <- predict(lm(y ~ x))
  return(y.trend)
}
rt$ham <- xts(array(NA, c(T,V)), order.by=index(gdp.vintages))
qrt$ham <- xts(array(NA, c(T,V)), order.by=index(gdp.vintages))
colnames(qrt$ham) <- colnames(rt$ham) <- colnames(gdp.vintages)

for (v in 1:V) {
  Y <- gdp.vintages[!is.na(gdp.vintages[,v]), v]
  c <- xts(c(rep(NA, 10), (Y[-c(1:10)] - ham.trend(Y))*100),
           order.by=index(Y))
  rt$ham[which(index(rt$ham) %in% index(c)),v] <- c

  tm1 <- which(index(gdp.vintages) == as.yearqtr(as.Date(colnames(gdp.vintages)[v]) %m-% months(3))) # index for expanding window
  Y <- gdp.vintages[1:tm1, V]
  c <- xts(c(rep(NA, 10), (Y[-c(1:10)] - ham.trend(Y))*100),
           order.by=index(Y))
  qrt$ham[which(index(qrt$ham) %in% index(c)), v] <- c
}


# Mueller-Watson -------------------------
mw.trend <- function(Y, Q=12) {
  y <- Y[!is.na(Y)]
  T <- length(y)
  Phi <- matrix(nrow=T, ncol=Q)
  j <- 1:T
  for (q in 1:Q) {
    Phi[,q] <- sqrt(2)*cos(q*(j-0.5)/T*pi)
  }
  mod <- lm(y ~., data=as.data.frame(Phi), na.action = na.exclude)
  y.trend <- predict(mod)
  return(y.trend)
}
rt$mw <- xts(array(NA, c(T,V)), order.by=index(gdp.vintages))
qrt$mw <- xts(array(NA, c(T,V)), order.by=index(gdp.vintages))
colnames(qrt$mw) <- colnames(rt$mw) <- colnames(gdp.vintages)

for (v in 1:V) {
  Y <- gdp.vintages[!is.na(gdp.vintages[,v]), v]
  c <- (Y - mw.trend(Y))*100
  rt$mw[which(index(rt$mw) %in% index(c)),v] <- c
  
  tm1 <- which(index(gdp.vintages) == as.yearqtr(as.Date(colnames(gdp.vintages)[v]) %m-% months(3))) # index for expanding window
  Y <- gdp.vintages[1:tm1, V]
  c <- (Y - mw.trend(Y))*100
  qrt$mw[which(index(qrt$mw) %in% index(c)),v] <- c
}



# Garrett et al ------------------
# Note: All functions below return the forecast augmented input series.
# The difference is in how they produce their forecasts

# Forecast Augmentation via Univariate AR [Garratt]
# Notes: 
# 1) 'data' is list of vintages for a variable y in xts format;
# 2) 'q' is order of AR; q cannot be larger than number of non na elements in each vintage;  
# 3) 'H' is number of forecasts ahead, the length of the augmentation.
# 4) ar.garratt$fa returns the forecast augmented vintage t log-output

ar.garratt = function(data, q, H) {
  
  # Function return the forecast-augmented input series.
  
  # Forecast Augmentation via univariate AR [Garratt]
  # Notes: 
  # 1) 'data' is matrix (T x 1) for a variable y in xts format;
  # 2) 'q' is order of AR; q cannot be larger than number of non na elements in each vintage;  
  # 3) 'H' is number of forecasts ahead, the length of the augmentation.
  # 4) ar.garratt$fa returns the forecast augmented vintage t log-output
  #    This thing has the same dimension as "data"
  
  data.diff=diff(data)  # Apply diff Operator
  T <- dim(Y[!is.na(Y)])[1]
  last <- last(index(Y)[!is.na(Y)]) # Last quarter of data
  
  # Clean NAs
  y=matrix(na.omit(data.diff))
  
  # Estimate AR(q)
  ar.uni <- ar(y, order.max=q, aic=FALSE)
  
  # Forecast AR(q)
  . <- predict(ar.uni, n.ahead=H)$pred
  yhat.xts <- xts(., order.by=seq(from=as.Date(last) %m+% months(3),
                                  length.out=H,
                                  by='quarter'))
  
  # Storage for augmented
  Y.aug <- xts(array(NA, (T+H)), order.by=seq(as.Date('1960-01-01'),
                                              length.out=(T+H),
                                              by='quarter'))
  
  for (t in 1:(T+H)) {
    if (t<=T) {
      Y.aug[t] <- as.numeric(Y[t])
    } else {
      Y.aug[t] <- as.numeric(Y.aug[t-1]) + as.numeric(yhat.xts[t-T])
    }
  }
  
  # Collect and Return Results
  result=list(fa=Y.aug, q=q, H=H)
  return(result)
}
q=8             # AR(q)
H=40            # Forecast Periods 

rt$hp.exp <- xts(array(NA, c(T,V)), order.by=index(gdp.vintages))
qrt$hp.exp <- xts(array(NA, c(T,V)), order.by=index(gdp.vintages))
colnames(qrt$hp.exp) <- colnames(rt$hp.exp) <- colnames(gdp.vintages)

for (v in 1:V) {
  Y <- gdp.vintages[, v]
  
  tm1 <- as.Date(colnames(gdp.vintages)[v]) %m-% months(3) # index for expanding window

  gdp.aug <- ar.garratt(Y,q,H)$fa
  c <- xts(hpfilter(na.omit(gdp.aug), freq=1600, type='lambda')$cycle*100, 
           order.by=index(na.omit(gdp.aug)))[paste0('/', tm1)]
  rt$hp.exp[1:length(c), v] <- c
  
  Y.qrt <- gdp.vintages[paste0('/', tm1), V]
  
  gdp.aug.qrt <- ar.garratt(Y.qrt, q, H)$fa
  c.qrt <- xts(hpfilter(na.omit(gdp.aug.qrt), freq=1600, type='lambda')$cycle*100, 
           order.by=index(na.omit(gdp.aug.qrt)))[paste0('/', tm1)]
  qrt$hp.exp[1:length(c.qrt), v] <- c.qrt
}


# Kamber Morley Wong -----------------------
max.amp.to.noise <- function(y, p=12, grid=.05) {
  
  # Travis Berge via code of Benjamin Wong
  # March 2022
  
  # The code chooses the signal to noise ratio, \delta, by doing a grid
  # search from 0.01 in increments of 0.01 to find the first local maximum of
  # the amplitude to noise ratio.
  
  # Note: Code has other dependencies
  
  # Inputs
  # y          time series
  # p          lag order
  # grid       step size

  delta <- 0.01  # Initialize amplitude to noise ratio
  
  bn <- KMW.filter(y, p, delta)
  amp.to.noise <- var(bn$c) / mean(bn$aux$resid^2)
  
  diff <- 1  # Initialize the difference to enter loop
  while (diff > 0) {
    delta <- delta + grid
    bn <- KMW.filter(y, p, delta);
    new.amp.to.noise <- var(bn$c) / mean(bn$aux$resid^2)

    diff <- new.amp.to.noise - amp.to.noise
    if (diff > 0) amp.to.noise <- new.amp.to.noise
    }
  delta <- delta - grid
  return(delta)
}
KMW.filter <- function(y, p=12, delta=.25) {
  
  # TJB from code of Benjamin Wong
  # Feb 2022
  
  # Inputs
  #   y -- Time series to implement BN filter in first differences
  #   p -- lag length
  #   delta -- signal-to-noise ratio
  
  # Outputs
  #   c -- Estimated cycle
  
  index <- index(y) # Save for later

  df.var <- function(y, p, constant=FALSE) {

    # TJB from code from Ben Wong Feb 2022
    
    # INPUTS
    #   y  Data (TxN) in first diffs
    #   p  lags
    
    # OUTPUTS
    #   B      VAR Coefficients
    #   sigma  Reduced form variance covariance matrix
    #   u      Time Series of Residuals
    #   X      The data matrix for the DF VAR
    
    T <- dim(y)[1] - p
    N <- dim(y)[2]
    
    y <- matrix(y, ncol=N)
    Y <- matrix(y[(1+p):(T+p), ]) # Lose p lags
    
    if (constant) {
      X <- matrix(1, nrow=T, ncol=1)
    } else {
      X <- array(numeric())
    }
    
    # First lag of Y on LHS
    x <- matrix(y[p:(T+p-1),], ncol=N)
    X <- cbind(X, x)
    
    # Now do second difference
    dy <- diff(y) # dy is (T.raw-1)xN
    Y <- Y[-1,] # (Lose an obs)
    X <- X[-1,] # (Lose an obs)

    # Build X matrix
    for (i in 1:(p-1)) {
      x <- dy[(1+p-i):(dim(dy)[1]-i),]
      X <- cbind(X, x)
    }
    
    # OLS 
    B <- solve(crossprod(X)) %*% crossprod(X,Y)
    e <-  Y - X%*%B # OLS residuals
    S <- as.numeric(crossprod(e) / (T-p))
    
    colnames(X) <- NULL
    
    return(list(B=B, e=e, sig2=S, X=X))
  } 
  ols.var <- function(y, p, constant=FALSE) {
    
    # This estimates an OLS reduced form VAR
    # TJB from code of Benjamin Wong Feb 2022
    
    # INPUTS
    #   y  Data needs to be matrix of (TxN)
    #   p  lags
    #   constant=FALSE by default
    
    T.raw <- dim(y)[1]
    T <- T.raw - p
    N <- dim(y)[2]
    
    Y <- matrix(y[(p+1):dim(y)[1], ],) # Cut Away first p lags
    
    # Put in first lag on LHS
    if (constant) {
      X <- array(1, c(T,1))
    } else {
      X <- array(numeric())
    }
    
    for (i in 1:p) {
      z <- y[(p+1-i):(T.raw-i),]
      X <- cbind(X, z)
    }
    
    B <- solve(crossprod(X)) %*% crossprod(X,Y)
    e <- Y - X %*% B
    S <- as.numeric(crossprod(e) / (T-N))
    
    colnames(X) <- NULL
    
    return(list(B=B, e=e, sig2=S, X=X))
  }

  # Calculate rho from delta 
  rho <- 1 - 1/sqrt(delta) # Eqn 6
  
  mu <- mean(y)
  y <- matrix(y)-mu
  
  var.df <- df.var(y, p) # AR via Dicky-Fuller
  var <- ols.var(y, p) # AR via OLS
  
  # Create Data for transformed model to estimate
  .y <- y[(p+2):dim(y)[1], ] - rho*var.df[['X']][,1]
  Y <- .y
  X <- var.df[['X']][,-1] # From DF above, keep only diff(y)'s
  
  # Set priors
  A.prior <- rep(0, p-1) # (p-1) phi* coefficients 
  V.prior <- diag(rep(.5, p-1) / (1:(p-1))^2) # Uninformative prior on constant centered on zero
  
  V.post <- solve(solve(V.prior) + (1/var[['sig2']])*crossprod(X))
  A.post <- V.post%*%(solve(V.prior) %*% A.prior + 1/var[['sig2']]*crossprod(X,Y))
  A.post <- c(rho, A.post) # Include rho
  
  # Map the parameters in DF form back to AR form 
  AR.params <- array(0, p)
  AR.params[p] <- -A.post[p] # last lag
  
  for (i in ((p-1):2)) {
    AR.params[i] <- -A.post[i] - sum(AR.params[i:p]) # Back to AR(2) term
  }
  
  # Calculate AR(1) term
  AR.params[1] <- rho - sum(AR.params[2:p])
  
  aux <- list()
  aux$AR <- AR.params
  aux$resid = y[(p+1):dim(y)[1]] - var$X%*%AR.params
  
  
  # BN Decomposition
  X <- rbind(var$X[2:dim(var$X)[1],], 
             c(last(y), var$X[dim(var$X)[1], -p])) # Update X
  
  F <- rbind(AR.params,
             diag(1, p-1, p))
  I <- diag(p)
  
  phi <- -F %*% solve(I-F) 
  c <- xts((phi %*% t(X))[1,], order.by=index[(p+1):dim(y)[1]])
  
  return(list(c=c, aux=aux))
}

rt$kmw <- xts(array(NA, c(T,V)), order.by=index(gdp.vintages))
qrt$kmw <- xts(array(NA, c(T,V)), order.by=index(gdp.vintages))
colnames(rt$kmw) <- colnames(qrt$kmw) <- colnames(gdp.vintages)

for (v in 4:V) {
  Y <- gdp.vintages[!is.na(gdp.vintages[,v]), v]
  gdp.rt <- diff(Y)[-1]*100
  
  delta.rt <- max.amp.to.noise(gdp.rt, grid=.01) # Optimize delta
  c.rt <- KMW.filter(gdp.rt, delta=delta.rt)[['c']]
  
  tm1 <- which(index(gdp.vintages) == as.yearqtr(as.Date(colnames(gdp.vintages)[v]) %m-% months(3))) # index for expanding window
  Y.qrt <- gdp.vintages[1:tm1, V]
  gdp.qrt <- diff(Y.qrt)[-1]*100
  
  delta.qrt <- max.amp.to.noise(gdp.qrt, grid=.01)
  c.qrt <- KMW.filter(gdp.qrt, delta=delta.qrt)[['c']]
  
  rt$kmw[which(index(rt$kmw) %in% index(c.rt)), v] <- c.rt
  qrt$kmw[which(index(qrt$kmw) %in% index(c.qrt)), v] <- c.qrt
}


# Save ---------
saveRDS(rt, file=paste0(data.dir, 'uni-rt-gap-vintages.Rds'))
saveRDS(qrt, file=paste0(data.dir, 'uni-qrt-gap-vintages.Rds'))
