# Extract Kalman Weights and Contributions

# ---------- DEFINE FUNCTIONS ----------

# ----- Extract Kalman Matrices Given Model and Data -----
KFKS <- function(data, prior, model) {
  # Kalman filter for stationary systems with missing data.
  # The model is        y_t = F*s_t      
  #                     s_t+1 = G * s_t + H*u_t; u \sim N(0,Omega)     
  # Inputs
  #       prior         A list of priors 
  #       model         A list of model parameters (as in matrices above)
  #       data          An array of dimension (prior$T x prior$Nobs)
  #     
  # Output 
  #        KF$st       Predicted state vector  s_t|t-1      (T x Ns)
  #        KF$stt      Filtered  state vector  s_t|t        (T x Ns)
  #        KF$Pt       Predicted covariance of s_t|t-1      (Ns x Ns x T)
  #        KF$Ptt      Filtered  covariance of s_t|t        (Ns x Ns)
  #        KF$loglik   Value of likelihood function

  F <- model$F
  G <- model$G
  H <- model$H
  Omega <- model$omega
  
  Y <- data
  
  # Output structure & dimensions
  KF <- list()
  KF$st <- array(NA, c(prior$T, prior$Ns)) # s_t|t-1
  KF$Pt <- array(NA, c(prior$Ns, prior$Ns, prior$T)) # P_t|t-1
  KF$stt <- array(NA, c(prior$T, prior$Ns))
  KF$Ptt <- array(NA, c(prior$Ns, prior$Ns, prior$T)) 
  KF$vt <- array(0, c(prior$T, prior$Nobs)) # v_t = y_t - F%*%st|t-1
  KF$Kt <- array(0, c(prior$Ns, prior$Nobs, prior$T))
  KF$Ftinv <- array(0, c(prior$Nobs, prior$Nobs, prior$T))
  
  stt <- prior$S0 # s_0|0
  Ptt <- prior$P0 # P_0|0
  
  for (t in 1:prior$T) {
    
    # Predict
    st <- G%*%stt  
    Pt <- G%*%Ptt%*%t(G) + H%*%Omega%*%t(H)
    Pt <- 1/2*(Pt+t(Pt)) 
    
    # Handle the missing data
    y <- as.vector(Y[t,])
    i <- !is.na(y)
    y.star <- y[i]
    F.star <- matrix(F[i,], ncol=prior$Ns)

    # Update
    if (sum(i)==0) {
      stt <- st
      Ptt <- Pt
    } else {
      Ftinv <- solve(F.star%*%Pt%*%t(F.star))
      Mt <- G%*%Pt%*%t(F.star)
      Kt <- Mt%*%Ftinv
      vt <- y.star - F.star%*%st
      stt <- st + Pt%*%t(F.star)%*%Ftinv%*%vt
      Ptt <- Pt - Pt%*%t(F.star)%*%Ftinv%*%F.star%*%Pt
    }
    
    KF$st[t,]<- st
    KF$Pt[,,t] <- Pt
    KF$vt[t,i] <- vt
    KF$stt[t,] <- stt
    KF$Ptt[,,t] <- Ptt
    KF$Kt[,i,t] <- Kt
    KF$Ftinv[i,i,t] <- Ftinv
  }
  
  # Smoothing filter
  KS <- list()
  KS$stT <- array(NA, c(prior$T, prior$Ns))
  KS$PtT <- array(NA, c(prior$Ns, prior$Ns, prior$T))
  KS$Ct <- array(NA, c(prior$Ns, prior$Nobs, prior$T))
  KS$Nt <- array(NA, c(prior$Ns, prior$Ns, prior$T))
  KS$Lt <- array(NA, c(prior$Ns, prior$Ns, prior$T))
  
  r.tm1 <- array(0, prior$Ns)
  N.tm1 <- array(0, c(prior$Ns, prior$Ns))
  for (t in prior$T:1) {

    i <- !is.na(Y[t,])

    rt <- r.tm1
    Nt <- N.tm1
    Lt <- G - KF$Kt[,,t]%*%F
    
    if (sum(i)==0) {
      r.tm1 <- t(Lt)%*%rt
      N.tm1 <- t(Lt)%*%Nt%*%Lt
      stT <- KF$st[t,] + KF$Pt[,,t]%*%r.tm1
      PtT <- KF$Pt[,,t] - KF$Pt[,,t]%*%N.tm1%*%KF$Pt[,,t]
      Dt <- t(KF$Kt[,,t]) %*% Nt %*% KF$Kt[,,t]
      Ct <- t(F)%*%Dt - t(G)%*%Nt%*%KF$Kt[,,t]
    } else {
      r.tm1 <- t(F)%*%KF$Ftinv[,,t]%*%KF$vt[t,] + t(Lt)%*%rt
      N.tm1 <- t(F)%*%KF$Ftinv[,,t]%*%F + t(Lt)%*%Nt%*%Lt
      stT <- KF$st[t,] + KF$Pt[,,t]%*%r.tm1
      PtT <- KF$Pt[,,t] - KF$Pt[,,t]%*%N.tm1%*%KF$Pt[,,t]
      Dt <- KF$Ftinv[,,t] + t(KF$Kt[,,t]) %*% Nt %*% KF$Kt[,,t]
      Ct <- t(F)%*%Dt - t(G)%*%Nt%*%KF$Kt[,,t] 
    }
    
    KS$stT[t,] <- stT
    KS$PtT[,,t] <- PtT
    KS$Ct[,,t] <- Ct
    KS$Nt[,,t] <- Nt
    KS$Lt[,,t] <- Lt
  }
  K <- list()
  K$KF <- KF
  K$KS <- KS
  return(K)
} # Kalman filter & smoother as in KH 2003


# ----- Extract State Prediction Weights Given Model and Data -----
w.st <- function(data, prior, model, K, tau) {
  # Finds weights for predicted value of observable using st|t-1
  # Equation (9) of Koopman & Harvey (2003)
  # The model is        y_t = F*s_t      
  #                     s_t+1 = G * s_t + H*u_t     
  # Inputs
  #       data          A data frame of dimension (prior$T x prior$Nobs)
  #       prior         A list of priors, model desc.
  #       model         A list of model parameters F, G, H
  #       K             The output from the function KFKS
  #       tau           Time indicator for observation
  #     
  # Output 
  #       w         (NsxNobsxT) array of weights for st|t-1
  #       cont      (NsxNobsxT) array of contributions to st|t-1
  #       stt       (Nsx1) array of st|t-1
  
  F <- model$F
  G <- model$G
  Y <- data
  Y[is.na(data)] <- 0
  
  w <- array(NA, c(prior$Ns, prior$Nobs, tau-1)) # Includes space for initial state
  B.t.jm1 <- diag(prior$Ns)
  for (j in (tau-1):1) {
    Bt.j <- B.t.jm1
    w[,,j] <- Bt.j %*% K$KF$Kt[,,j]
    B.t.jm1 <- Bt.j%*%G - w[,,j]%*%F
  }

  wdoty <- array(NA, c(prior$Ns, prior$Nobs, tau-1)) 
  for (t in 1:(tau-1)) {
    wdoty[,,t] <- w[,,t] * matrix(Y[t,], nrow=prior$Ns, ncol=prior$Nobs, byrow=T)
  }
  wy <- t(apply(wdoty, c(1,3), sum))

  # Calculate influence of initialization
  if (sum(prior$S0)==0) {
    w0 <- array(0, prior$Ns)
  } else {
    Lt <- K$KS$Lt[,,(tau-1)]
    for (t in (tau-2):1) Lt <- Lt %*% K$KS$Lt[,,t]
    w0 <- Lt%*%G%*%prior$S0
  }

  ret <- list()
  ret$w <- w
  ret$w0 <- w0
  ret$cont <- wdoty
  ret$st <- apply(wy, 2, sum) + w0
  return(ret)
}


# ----- Extract State Filtering Weights Given Model and Data -----

w.stt <- function(data, prior, model, K, tau) {
  # Weights for predicted value of observable using st|t
  # The model is        y_t = F*s_t      
  #                     s_t+1 = G * s_t + H*u_t     
  # Inputs
  #       data          A data.frame of dimension (prior$T x prior$Nobs)
  #       prior         A list of priors, model desc.
  #       model         A list of model parameters F, G, H
  #       K             The output from the function KF.2003
  #       tau           Time indicator for observation
  #     
  # Output 
  #       w         (NsxNobsxT) array of weights for st|t
  #       cont      (NsxNobsxT) array of contributions to st|t
  #       stt        (Nsx1) array of st|t
  F <- model$F
  G <- model$G
  Y <- data
  Y[is.na(data)] <- 0
  
  w <- array(NA, c(prior$Ns, prior$Nobs, tau))
  w[,,tau] <- K$KF$Pt[,,tau]%*%t(F)%*%K$KF$Ftinv[,,tau]
  
  B.tm1 <- (diag(prior$Ns) - K$KF$Pt[,,tau] %*%t(F)%*%K$KF$Ftinv[,,tau]%*%F)
  for (t in (tau-1):1) {
    Bt <- B.tm1
    w[,,t] <- Bt %*% K$KF$Kt[,,t]
    B.tm1 <- Bt%*%G - w[,,t]%*%F
  }
  
  wdoty <- array(NA, c(prior$Ns, prior$Nobs, tau))
  for (t in 1:tau) {
    wdoty[,,t] <- w[,,t] * matrix(Y[t,], nrow=prior$Ns, ncol=prior$Nobs, byrow=T)
  }
  wy <- t(apply(wdoty, c(1,3), sum))
  
  # Calculate influence of initialization
  if (sum(prior$S0)==0) {
    w0 <- array(0, prior$Ns)
  } else {
    Lt <- K$KS$Lt[,,(tau-1)]
    for (t in (tau-2):1) Lt <- Lt %*% K$KS$Lt[,,t]
    w0 <- (diag(prior$Ns) - K$KF$Pt[,,tau]%*%t(F)%*%K$KF$Ftinv[,,tau]%*%F)%*%Lt%*%G%*%prior$S0
  }
  
  ret <- list()
  ret$w <- w
  ret$w0 <- w0
  ret$cont <- wdoty
  ret$stt <- apply(wy, 2, sum) + w0
  return(ret)  
}

# ----- Extract State Smoothing Weights Given Model and Data -----
w.stT <- function(data, prior, model, K, tau) {
  # Weights for predicted value of observable using st|T
  # The model is        y_t = F*s_t      
  #                     s_t+1 = G * s_t + H*u_t     
  # Inputs
  #       data          A data.frame of dimension (prior$T x prior$Nobs)
  #       prior         A list of priors, model desc.
  #       model         A list of model parameters F, G, H
  #       K             The output from the function KF.2003
  #       tau           Time indicator for observation
  #     
  # Output 
  #       w         (NsxNobsxT) array of weights for s_{t|T}
  #       cont      (NsxNobsxT) array of contributions to s_{t|T}
  #       stT       (Nsx1) array of s_{t|T}
  
  F <- model$F
  G <- model$G
  Y <- data
  Y[is.na(data)] <- 0
  
  w <- array(NA, c(prior$Ns, prior$Nobs, prior$T))
  
  # Backwards recursion (9)
  B.tm1 <- (diag(prior$Ns) - K$KF$Pt[,,tau]%*%K$KS$Nt[,,tau-1])
  for (t in (tau-1):1) {
    Bt <- B.tm1
    w[,,t] <- Bt %*% K$KF$Kt[,,t]
    B.tm1 <- Bt%*%G - w[,,t]%*%F
  }
  
  # Forward recursion (eqn 15)
  Btt <- K$KF$Pt[,,tau]
  for (t in tau:prior$T) {
    Bt <- Btt
    w[,,t] <- Bt%*%K$KS$Ct[,,t]
    Btt <- Bt%*%t(K$KS$Lt[,,t])
  }
  
  wdoty <- array(NA, c(prior$Ns, prior$Nobs, prior$T))
  for (t in 1:prior$T) {
    wdoty[,,t] <- w[,,t] * matrix(Y[t,], nrow=prior$Ns, ncol=prior$Nobs, byrow=T)
  }
  wy <- t(apply(wdoty, c(1,3), sum))
  
  # Calculate influence of initialization
  if (sum(prior$S0)==0) {
    w0 <- array(0, prior$Ns)
  } else {
    Lt <- K$KS$Lt[,,(tau-1)]
    for (t in (tau-2):1) Lt <- Lt %*% K$KS$Lt[,,t]
    w0 <- (diag(prior$Ns) - K$KF$Pt[,,tau] %*% K$KS$Nt[,,tau-1]) %*% Lt%*%G%*%prior$S0
  }
  
  ret <- list()
  ret$w <- w
  ret$w0 <- w0
  ret$cont <- wdoty
  ret$stT <- apply(wy, 2, sum) + w0
  return(ret)
}

# ----- Extract Signal Prediction Weights Given Model and Data -----
w.yt <- function(data, prior, model, K, tau) {
  # Finds weights for predicted value of observable using st|t-1
  # Equation (9) of Koopman & Harvey (2003)
  # The model is        y_t = F*s_t      
  #                     s_t+1 = G * s_t + H*u_t     
  # Inputs
  #       data          A data.frame of dimension (prior$T x prior$Nobs)
  #       prior         A list of priors, model desc.
  #       model         A list of model parameters F, G, H
  #       K             The output from the function KF.2003
  #       tau           Time indicator for observation
  #     
  # Output 
  #       w         Weights for E(y_t|t) -- (NobsxNobsxT)
  #       wy        Contributions to E(y_t|t) -- (NobsxNobsxT)
  #       yt        E(y_t|t-1) -- (1x1)
  
  F <- model$F
  G <- model$G
  Y <- data
  Y[is.na(data)] <- 0
  
  Fw <- array(NA, c(prior$Nobs, prior$Nobs, tau-1))
  B.tm1 <- F%*%diag(prior$Ns)
  for (t in (tau-1):1) {
    Bt <- B.tm1
    Fw[,,t] <- Bt %*% K$KF$Kt[,,t]
    B.tm1 <- Bt%*%G - Fw[,,t]%*%F
  }
  
  wdoty <- array(NA, c(prior$Nobs, prior$Nobs, tau-1))
  for (t in 1:(tau-1)) {
    wdoty[,,t] <- Fw[,,t] * matrix(Y[t,], nrow=prior$Nobs, ncol=prior$Nobs, byrow=T)
  }

  
    
  ret <- list()
  ret$w <- Fw
  ret$cont <- wdoty
  wy <- t(apply(wdoty, c(1,3), sum)) #Sums across w*y indicators each t
  ret$yt <- apply(wy, 2, sum) # sums across t
  return(ret)
}

# ----- Extract Signal Filtering Weights Given Model and Data -----
w.ytt <- function(data, prior, model, K, tau) {
  # Weights for predicted value of observable using st|t
  # The model is        y_t = F*s_t      
  #                     s_t+1 = G * s_t + H*u_t     
  # Inputs
  #       prior         A list of priors, model desc.
  #       model         A list of model parameters F, G, H
  #       data          A data.frame of dimension (prior$T x prior$Nobs)
  #       KFKS          The output from the function KF.2003
  #       tau           Time indicator for observation
  #     
  # Output 
  #       w         Weights for E(y_t|t) -- (NobsxNobsxT)
  #       wy        Contributions to E(y_t|t) -- (NobsxNobsxT)
  #       ytt        E(y_t|T) -- (1x1)
  
  F <- model$F
  G <- model$G
  Y <- data
  Y[is.na(data)] <- 0
  
  Fw <- array(NA, c(prior$Nobs, prior$Nobs, tau))
  Fw[,,tau] <- F%*%K$KF$Pt[,,tau]%*%t(F)%*%K$KF$Ftinv[,,tau]
  
  B.tm1 <- F%*%(diag(prior$Ns) - K$KF$Pt[,,tau]%*%t(F)%*%K$KF$Ftinv[,,tau]%*%F)
  for (t in (tau-1):1) {
    Bt <- B.tm1
    Fw[,,t] <- Bt %*% K$KF$Kt[,,t]
    B.tm1 <- Bt%*%G - Fw[,,t]%*%F
  }
  
  wdoty <- array(NA, c(prior$Nobs, prior$Nobs, tau))
  for (t in 1:tau) {
    wdoty[,,t] <- Fw[,,t] * matrix(Y[t,], nrow=prior$Nobs, ncol=prior$Nobs, byrow=T)
  }
  
  
  ret <- list()
  ret$w <- Fw
  ret$cont <- wdoty
  wy <- t(apply(wdoty, c(1,3), sum)) #Sums across w*y indicators each t
  ret$ytt <- apply(wy, 2, sum) # sums across t
  return(ret)
}

# ----- Extract Signal Smoothing Weights Given Model and Data -----
w.ytT <- function(data, prior, model, K, tau) {
  # Weights for predicted value of observable using st|T
  # The model is        y_t = F*s_t      
  #                     s_t+1 = G * s_t + H*u_t     
  # Inputs
  #       prior         A list of priors, model desc.
  #       model         A list of model parameters F, G, H
  #       data          A data.frame of dimension (prior$T x prior$Nobs)
  #       KFKS          The output from the function KF.2003
  #       tau           Time indicator for observation
  #     
  # Output 
  #       w         Weights for E(y_t|T) -- (Nobs x Nobs x T)
  #       wy        Contributions to E(y_t|T) -- (Nobs x Nobs x T)
  #       ytT        E(y_t|T) -- (1x1)
  
  F <- model$F
  G <- model$G
  Y <- data
  Y[is.na(data)] <- 0
  
  Fw <- array(NA, c(prior$Nobs, prior$Nobs, prior$T))
  
  # Backwards recursion (9)
  B.tm1 <- F%*%(diag(prior$Ns) - K$KF$Pt[,,tau]%*%K$KS$Nt[,,tau-1])
  for (t in (tau-1):1) {
    Bt <- B.tm1
    Fw[,,t] <- Bt %*% K$KF$Kt[,,t]
    B.tm1 <- Bt%*%G - Fw[,,t]%*%F
  }
  
  # Forward recursion (15)
  Btt <- F%*%K$KF$Pt[,,tau]
  for (t in tau:prior$T) {
    Bt <- Btt
    Fw[,,t] <- Bt%*%K$KS$Ct[,,t]
    Btt <- Bt%*%t(K$KS$Lt[,,t])
  }
  
  wdoty <- array(NA, c(prior$Nobs, prior$Nobs, prior$T))
  for (t in 1:prior$T) {
    wdoty[,,t] <- Fw[,,t] * matrix(Y[t,], nrow=prior$Nobs, ncol=prior$Nobs, byrow=T)
  }
  
  ret <- list()
  ret$w <- Fw
  ret$cont <- wdoty
  wy <- t(apply(wdoty, c(1,3), sum)) # sums across i each t
  ret$ytT <- apply(wy, 2, sum) # sums across t
  return(ret)
}


